 /*******************************************************************************
 *
 *                ctran
 *
 *******************************************************************************
 *
 *  File         : Sy_AvsCtran.cc
 *
 *  Type         : Implementation, Interface
 * 
 *  Purpose      : 
 *
 *******************************************************************************
 *                 
 *  Author       : Markus Naef
 *  started      :  3-May-94
 *  last change  :  4-May-94
 *
 *******************************************************************************/
#include <stdio.h>

#include <avs/avs.h>
#include <avs/field.h>
#include <avs/geom.h>
#include <avs/colormap.h>
#include <sys/types.h>

#include "Sy_Skeleton.h"

AVS_STATIC(int Sy_AvsCtran_compute,(AVSfield_char *in,char **out,char *filename,int verbose,int ascii, int timer));

int Sy_AvsCtran_desc()
/******************************************************************************
 *
 * 
 * 
 *
 *****************************************************************************/
{
   int param;
   AVSset_module_name("crack translator",MODULE_MAPPER);
   AVScreate_input_port("image","field 2D scalar uniform byte",REQUIRED);
   AVScreate_output_port("filename of output","string");

   param=AVSadd_parameter("output file","string","dummy.ipo","",".ipo");
   AVSconnect_widget(param,"browser");
   param=AVSadd_parameter("verbose","boolean",0,0,1);
   param=AVSadd_parameter("ascii output","boolean",0,0,1);
   param=AVSadd_parameter("timer","boolean",0,0,1);

   AVSset_compute_proc(Sy_AvsCtran_compute);
   return(1);
}

static int
Sy_AvsCtran_compute(
/* ******************************************************************************
 *
 * AVS computation function
 *
 * 
 *
 *
 ****************************************************************************** */
AVSfield_char	*in,
char	 **out,		
char *filename,
int verbose,
int ascii,
int timer)  
{
   int dims[2];
   SyIpoDAC       ipd;
   SyCratranControl cntl;
   SyImage_t      *ima;
   SyIOhandle*    iohandle;
   char           charbuf[SY_CHARBUF_SIZE];
   int            is_verbose=verbose, is_ascii=ascii, is_timer=timer,i;

   
   AVSfield_get_dimensions((AVSfield *)in,dims);

   if (*out) AVSdata_free("string",*out);
   (*out)= (char*)malloc((strlen(filename)+1)*sizeof(char));;
   strcpy(*out,filename);

   Sy_SetVerbose (is_verbose);

   ima = Sy_Malloc(sizeof(SyImage_t));

   ima->title     = NULL;
   ima->rgb.red   = NULL;
   ima->rgb.blue  = NULL;
   ima->rgb.green = NULL;

   ima->width=dims[0];
   ima->height=dims[1];
   ima->depth=8;
   ima->linebytes=dims[0];
   ima->data=(char*)Sy_Malloc(dims[0]*dims[1]*sizeof(char));
   for (i=0;i<dims[0]*dims[1];i++) {
      ima->data[i]=in->data[i];
   };

   if (is_timer) Sy_InitTimer();

   /* run crack extraction */
   cntl.reverse_bits = 0;
   cntl.invert_image = SY_FALSE;

   Sy_Cratran (ima, &ipd, &cntl);

   Sy_FreeImage (ima); 

   /* build initiating contours */
   Sy_Message ("\n%s ", "@ building SyIcns");

   Sy_CratranBuildIcn (&ipd);

   Sy_Message ("%s\n", "done");

   if (is_timer) Sy_ShowTimer();

   sprintf (charbuf, "@ initiating points: %d\n", ipd.ipodim);
   Sy_Message ("%s", charbuf);

   iohandle = Sy_CreateIOhandle ();
   Sy_SetIOproperty (iohandle, FORMAT, (is_ascii) ? SY_ASCII:SY_BINARY);
   Sy_SetIOproperty (iohandle, TYPE, SY_IPD_STRIP);
   Sy_SetIOproperty (iohandle, DATA, (char*) &ipd);
   Sy_SetIOproperty (iohandle, IO_MODE, SY_FREE_AFTER_WRITE);
   
   Sy_PutIOhandle (iohandle);
   Sy_ManageOutput (*out);
   
   sprintf (charbuf, "@ writing SyIpos to %s \n", *out);
   Sy_Message ("%s", charbuf);
      
   return 1;
}   



