 /*******************************************************************************
 *
 *                prune
 *
 *******************************************************************************
 *
 *  File         : prune_avs.cc
 *
 *  Type         : Implementation, Interface
 * 
 *  Purpose      : 
 *
 *******************************************************************************
 *                 
 *  Author       : Markus Naef
 *  started      :  6-May-94
 *  last change  :  6-May-94
 *
 *******************************************************************************/

#include <stdio.h>

#include <avs/avs.h>
#include <avs/field.h>
#include <avs/geom.h>
#include <avs/colormap.h>

#include <sys/types.h>

#include "Sy_Skeleton.h"
#include "Sy_Graphics.h"

AVS_STATIC(int Sy_AvsPrune_compute,(char *in_file,char **out,char *out_file,char *reschoice, int bdtrans,int border,int pyramid,int thorough,int trigger,int cluster,float *th_p,float *pyr_th_p,float *pyr_delta_p,int rank,float *sigma_p,float *sigma_hi_p,float *sigma_lo_p,int max_iter,int verbose,int ascii,int timer));

int
Sy_AvsPrune_desc()
/******************************************************************************
 *
 * 
 * 
 *
 *****************************************************************************/
{
   int param;
   AVSset_module_name("prune skeleton",MODULE_MAPPER);
   AVScreate_input_port("filename of input","string",REQUIRED);
   AVScreate_output_port("filename of output","string");

   param=AVSadd_parameter("output file","string","dummy.ske","",".ske");
   AVSconnect_widget(param,"browser");

   param = AVSadd_parameter("residual", "choice", "smooth chord", 
			    "chord|circularity|bicircularity|potential|smooth potential|smooth chord|automatic", "|");
   AVSconnect_widget(param, "radio_buttons");

   param=AVSadd_parameter("bdtrans","boolean",0,0,1);
   param=AVSadd_parameter("add border","boolean",0,0,1);

   param=AVSadd_parameter("skeleton pyramid","boolean",0,0,1);
   param=AVSadd_parameter("thorough","boolean",0,0,1);
   param=AVSadd_parameter("trigger","boolean",0,0,1);
   param=AVSadd_parameter("cluster","boolean",0,0,1);

   param=AVSadd_float_parameter("th",3.0,0.0,100.0);
   AVSconnect_widget(param,"slider");

   param=AVSadd_float_parameter("pyr th",2.0,0.0,100.0);
   AVSconnect_widget(param,"slider");

   param=AVSadd_float_parameter("pyr delta",10.0,0.0,100.0);
   AVSconnect_widget(param,"slider");

   param=AVSadd_parameter("rank","integer",1,0,10);
   AVSconnect_widget(param,"islider");

   param=AVSadd_float_parameter("sigma",3.0,0.0,100.0);
   AVSconnect_widget(param,"slider");

   param=AVSadd_float_parameter("sigma hi",50.0,0.0,200.0);
   AVSconnect_widget(param,"slider");

   param=AVSadd_float_parameter("sigma lo",2.0,0.0,200.0);
   AVSconnect_widget(param,"slider");

   param=AVSadd_parameter("max iterations","integer",0,0,10);
   AVSconnect_widget(param,"islider");

   param=AVSadd_parameter("verbose","boolean",0,0,1);
   param=AVSadd_parameter("ascii output","boolean",0,0,1);
   param=AVSadd_parameter("timer","boolean",0,0,1);

   AVSset_compute_proc(Sy_AvsPrune_compute);
   return(1);
}

static int
Sy_AvsPrune_compute(
/* ******************************************************************************
 *
 * AVS computation function
 *
 * 
 *
 *
 ****************************************************************************** */
char *in_file,
char **out,
char *out_file,
char *reschoice,
int bdtrans,
int border,
int pyramid,
int thorough,
int trigger,
int cluster,
float *th_p,
float *pyr_th_p,
float *pyr_delta_p,
int rank,
float *sigma_p,
float *sigma_hi_p,
float *sigma_lo_p,
int max_iter,
int verbose,
int ascii,
int timer)  
{
   int dims[2];
   enum PruneOp {PRUNE_WITH_IPD, PRUNE_WITH_TRIAN, PRUNE_WITH_IPD_PYR, 
		   NO_PRUNEOP};

   SyVoroDAC   *vd; 
   SyIpoDAC    *ipd;
   SyTriDAC    *tri;
   SySkelDAC   vsk;
   SyMScaleSkelDAC mvsk;
   SyTriPot    *pot; 

   SyIOhandle  *iohandle, *iohandle_ipo, *iohandle_tri;

   int          k;
   int          is_verbose=verbose, is_ascii=ascii, 
                is_timer=timer, is_eresid=SY_FALSE, is_border=border, 
                is_output=SY_FALSE, is_automatic=SY_FALSE,
                is_infix=SY_FALSE, is_maxfix=SY_FALSE,
                is_circ=SY_FALSE, is_bicirc=SY_FALSE, is_chord=SY_FALSE,
                is_poti=SY_FALSE, is_smoothpoti=SY_FALSE, 
                is_smoothcho=SY_FALSE, is_thorough=thorough,
                is_pyramid=pyramid, is_reconstruct=SY_FALSE;
   
   int          is_bdtrans = bdtrans;
   
   char         charbuf[SY_CHARBUF_SIZE],
                th_str[SY_SMALL_STRING_LEN];
   
   float        base_reg_thres;
   float        th=*th_p, pyr_th=*pyr_th_p,
                pyr_delta= *pyr_delta_p;

   float        sigma=*sigma_p, sigma_hi=*sigma_hi_p, sigma_lo=*sigma_lo_p;
   
   float        sigma_step;
   int          io_result;
   int          no_ipo_found = 1, no_voro_found = 1, no_tri_found = 1;
   enum PruneOp oper = NO_PRUNEOP;

   SyDisplayControl control;
   SyRegulControl   rcntrl;

   switch (AVSchoice_number("residual", reschoice)){
   case 1: is_chord      = SY_TRUE; break;
   case 2: is_circ       = SY_TRUE; break;
   case 3: is_bicirc     = SY_TRUE; break;
   case 4: is_poti       = SY_TRUE; break;
   case 5: is_smoothpoti = SY_TRUE; break;   
   case 6: is_smoothcho  = SY_TRUE; break;
   case 7: is_automatic  = SY_TRUE; break;
   default: is_smoothcho = SY_TRUE;
   }

   if (*out) AVSdata_free("string",*out);
   (*out)= (char*)malloc(strlen(out_file)*sizeof(char));;
   strcpy(*out,out_file);

   control.display_mode = SKE_PLAIN;
   sigma_step = 2.0;

   if (max_iter > 0){
      if (max_iter == 1) max_iter = 2;
      sigma_step = (sigma_hi - sigma_lo)/(max_iter - 1);
   }
 
   rcntrl.sigma_lo       = sigma_lo;
   rcntrl.sigma_hi       = sigma_hi;
   rcntrl.sigma_step     = sigma_step;
   rcntrl.base_threshold = 1.0;
   rcntrl.is_edge_selection     = SY_FALSE;
   rcntrl.is_pyr_trigger_hint   = SY_FALSE;
   rcntrl.is_pyr_cluster_hint   = SY_FALSE;
   rcntrl.pyr_cluster_hint      = pyr_delta;

   rcntrl.gaussian_smoothing = SY_TRUE;
   rcntrl.oliensis_smoothing = SY_FALSE;

   Sy_SetVerbose (is_verbose);

   sprintf (syStrBuf, "loading Voronoi diagram from %s...", in_file);
   Sy_Message (SY_FMTNL, syStrBuf);

   Sy_ManageInput (in_file);
   while ((iohandle = Sy_NextIOhandle()) != NULL) {
      switch (Sy_GetIOproperty (iohandle, TYPE).type){
       case SY_IPD_STRIP:
	 ipd = (SyIpoDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	 iohandle_ipo = iohandle;
	 oper = PRUNE_WITH_IPD;
	 break;
       case SY_VORO_STRIP:
	 vd = (SyVoroDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	 no_voro_found = 0;
	 break;
       case SY_TRIAN_STRIP:
	 tri = (SyTriDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	 oper = PRUNE_WITH_TRIAN;
	 iohandle_tri = iohandle;
	 break;
      };
   };

   if (no_voro_found){
      Sy_SimpleError ("No SyVoroDAC structure found in %s\n", in_file);
   }

   sprintf (syStrBuf, "points:%d edges:%d vertices:%d\n",
	    vd->pdim, vd->edim, vd->vdim);
    Sy_Message (SY_FMTNL, syStrBuf); 
   
   ipd->info  = NULL;
   ipd->contourdim = 0;
   ipd->maxcontourdim = 0;
   vsk.pdim   = vd->pdim;
   vsk.edim   = vd->edim;
   vsk.vdim   = vd->vdim;
   vsk.point  = vd->point;
   vsk.edge   = vd->edge;
   vsk.vertex = vd->vertex;
   vsk.regul  = (SyRegul*) NULL;
   vsk.is_nfurcation_info = SY_FALSE;
   vsk.is_smcir           = SY_FALSE;

   if (is_eresid) Sy_SetResidualCompMode (RES_FULL);
   if (is_border) Sy_SetResidualBorderMode (RES_FULL);
   if (is_infix)  Sy_SetInfix (is_infix);
   if (is_maxfix) Sy_SetMaxfix (is_maxfix);

   if (is_timer) Sy_InitTimer();

   switch (oper){
    case NO_PRUNEOP:
      Sy_SimpleError ("No IPD/TRIAN structure found in %s\n", in_file);
      break;
    case PRUNE_WITH_IPD:
      if (is_automatic){
	vsk.prune_method = SKE_SMCHO;
	rcntrl.smcir_term_thres      = is_thorough ? 3.0 : 1.0;
	rcntrl.smcir_search_thres    = SY_SMCIR_SEARCH_THRES;
	Sy_Message (SY_FMTNL, "Automatic scale selection:");
	Sy_SkelSpaceRegularize (&mvsk, &vsk, ipd, &rcntrl);
      }
      else  
	if (is_smoothpoti){
	   vsk.prune_method = SKE_SMPOT;
	   Sy_RegSmooth (&vsk, ipd, Sy_SmoothRegularize, is_verbose, sigma, 
			 is_bdtrans, &rcntrl);
	   if (is_output) 
	     Sy_OutputBirthSetHistogramNew (&vsk, "dummy.out", th, 
					    sigma, 1);
	}
	else
	  if (is_smoothcho){
	     vsk.prune_method = SKE_SMCHO;
	     if (max_iter == 0){
		Sy_RegSmooth (&vsk, ipd, Sy_SmoothRegularize, is_verbose, sigma,
			      is_bdtrans, &rcntrl);
		if (is_output) 
		  Sy_OutputBirthSetHistogramm (&vsk, "dummy.out", th, 
					       sigma, 1);
	     }
	     else{
		for (k = 0; k < max_iter; k++){
		   if (vsk.regul != (SyRegul*)NULL) free(vsk.regul);
		   sigma = sigma_lo + k*sigma_step;
		   Sy_RegSmooth (&vsk, ipd, Sy_SmoothRegularize, is_verbose, sigma, is_bdtrans, &rcntrl);
		   Sy_AddMScaleInfo (&mvsk, &vsk, sigma, max_iter, k);
		}
		if (is_output)
		  Sy_OutputSkelSpaceLog (&mvsk, "dummy.out", th);
	     }/* else : max_iter > 0 */
	  } else 
	    if (is_circ) 
	      Sy_Regularize (&vsk, ipd, Sy_CirResidual, is_verbose);
	    else 
	      if (is_poti) 
		Sy_Regularize (&vsk, ipd, Sy_PotResidual, is_verbose);
	      else
		if (SY_FALSE) 
		  Sy_Regularize (&vsk, ipd, Sy_KleinResidual, is_verbose);
		else
		  if (is_bicirc) 
		    Sy_Regularize (&vsk, ipd, Sy_BicirResidual, is_verbose);
		  else
		    Sy_Regularize (&vsk, ipd, Sy_ChordResidual, is_verbose);
      break;
    case PRUNE_WITH_TRIAN:
      if (is_circ) 
	Sy_ComplexRegularize (&vsk, tri, Sy_ComplexCirResidual);
      else
	if (is_poti) 
	  Sy_ComplexRegularize (&vsk, tri, Sy_ComplexPotResidual);
	else
	  if (SY_FALSE) 
	    Sy_ComplexRegularize (&vsk, tri, Sy_ComplexKleinResidual);
	  else
	    Sy_ComplexRegularize (&vsk, tri, Sy_ComplexChordResidual);
      break;
   }

   if (is_timer) Sy_ShowTimer();
   
   vsk.is_hierarchy = SY_FALSE;
   vsk.hier         = NULL;

   if (is_pyramid){
      base_reg_thres               = 1.0;
      rcntrl.term_threshold        = 0.5;
      rcntrl.base_threshold        = 1.0;
      rcntrl.search_threshold      = 1.0;

      rcntrl.smcir_term_thres      = is_thorough ? 3.0 : SY_SMCIR_TERM_THRES;
      rcntrl.smcir_search_thres    = SY_SMCIR_SEARCH_THRES;
      rcntrl.pyr_trigger_threshold = pyr_th;
      rcntrl.pyr_cluster_delta     = pyr_delta;
      rcntrl.is_search_threshold   = SY_TRUE;
      rcntrl.is_term_threshold     = SY_TRUE;
      
      if (trigger) 
	rcntrl.is_pyr_trigger_hint = SY_FALSE;
      if (cluster) 
	rcntrl.is_pyr_cluster_hint = SY_FALSE;

      if (is_timer) Sy_InitTimer();
    
      Sy_Pyramid (&vsk, ipd, &rcntrl);
    
      if (is_timer) Sy_ShowTimer();    

      oper = PRUNE_WITH_IPD_PYR;
   }

   iohandle = Sy_CreateIOhandle ();
   Sy_SetIOproperty (iohandle, FORMAT, (is_ascii) ? SY_ASCII:SY_BINARY);
   Sy_SetIOproperty (iohandle, TYPE, SY_SKEL_STRIP);
   Sy_SetIOproperty (iohandle, DATA, (char*) &vsk);
   Sy_SetIOproperty (iohandle, IO_MODE, SY_FREE_AFTER_WRITE);

   switch (oper){
    case PRUNE_WITH_IPD:
    case PRUNE_WITH_IPD_PYR:
      Sy_SetIOproperty (iohandle_ipo, FORMAT, (is_ascii)?SY_ASCII:SY_BINARY);
      Sy_PutIOhandle (iohandle_ipo);
      break;
    case PRUNE_WITH_TRIAN:
      Sy_SetIOproperty (iohandle_tri, FORMAT, (is_ascii)?SY_ASCII:SY_BINARY);
      Sy_PutIOhandle (iohandle_tri);
      break;
   }

   Sy_PutIOhandle (iohandle);
   Sy_ManageOutput (out_file);
     
   sprintf (syStrBuf, "writing Voronoi skeleton to %s", out_file);
   Sy_Message (SY_FMTNL, syStrBuf);
   
   return 1;
}   




