 /*******************************************************************************
 *
 *                voro
 *
 *******************************************************************************
 *
 *  File         : voro_avs.cc
 *
 *  Type         : Implementation, Interface
 * 
 *  Purpose      : 
 *
 *******************************************************************************
 *                 
 *  Author       : Markus Naef
 *  started      :  3-May-94
 *  last change  :  5-May-94
 *
 *******************************************************************************/

#include <stdio.h>

#include <avs/avs.h>
#include <avs/field.h>
#include <avs/geom.h>
#include <avs/colormap.h>
#include <sys/types.h>

#include "Sy_Skeleton.h"
#include "Sy_Graphics.h"

AVS_STATIC(int Sy_AvsVoro_compute,(char *in_file,char **out,char *out_file,int verbose,int ascii, int timer));

int Sy_AvsVoro_desc()
/******************************************************************************
 *
 * 
 * 
 *
 *****************************************************************************/
{
   int param;
   AVSset_module_name("Voronoi",MODULE_MAPPER);
   AVScreate_input_port("filename of input","string",REQUIRED);
   AVScreate_output_port("filename of output","string");

   param=AVSadd_parameter("output file","string","dummy.vor","",".vor");
   AVSconnect_widget(param,"browser");

   param=AVSadd_parameter("verbose","boolean",0,0,1);
   param=AVSadd_parameter("ascii output","boolean",0,0,1);
   param=AVSadd_parameter("timer","boolean",0,0,1);

   AVSset_compute_proc(Sy_AvsVoro_compute);
   return(1);
}

static int Sy_AvsVoro_compute(
/* ******************************************************************************
 *
 * AVS computation function
 *
 * 
 *
 *
 ****************************************************************************** */
char *in_file,
char **out,
char *out_file,
int verbose,
int ascii,
int timer)  
{
   int dims[2];
   SyVoroDAC     vd;
   SyIpoDAC*     ipd;
   SyTriDAC*     tri;
   
   int           e_mem, v_mem;
  
   char          charbuf[SY_CHARBUF_SIZE];

   int           no_ipo_found = SY_TRUE, no_trian_found = SY_FALSE;
   SyIOhandle    *iohandle, *iohandle_point;
   SyRusdelta    delta;

   enum SyStripType  data_type;

   int            is_verbose=verbose, is_ascii=ascii, is_timer=timer;

   
   if (*out) AVSdata_free("string",*out);
   (*out)= (char*)malloc(strlen(out_file)*sizeof(char));;
   strcpy(*out,out_file);

   Sy_SetVerbose (is_verbose);

   if (is_verbose)
     fprintf (stderr, "loading SyIpos from %s...\n", in_file);

   Sy_ManageInput (in_file);
   while (((iohandle = Sy_NextIOhandle()) != NULL) && no_ipo_found){
      switch (data_type = Sy_GetIOproperty (iohandle, TYPE).type){
       case SY_IPD_STRIP:
	 no_ipo_found = 0;
	 ipd = (SyIpoDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	 iohandle_point = iohandle;
	 break;
       case SY_TRIAN_STRIP:
	 no_trian_found = 0;
	 tri = (SyTriDAC*) (Sy_GetIOproperty(iohandle, DATA).data);
	 iohandle_point = iohandle;
	 break;
      };
   };
   if (no_ipo_found && no_trian_found) {
      Sy_SimpleError ("No SyIpo/TRIAN file found in %s\n", in_file);
   };
   
   if (is_verbose){ 
      switch(data_type){
       case SY_IPD_STRIP:
	 fprintf (stderr, "SyIpos: %d  SyIcns: %d\n", 
		  ipd->ipodim, ipd->icndim);
	 break;
       case SY_TRIAN_STRIP:
	 fprintf (stderr, "points: %d  edges: %d\n", 
		  tri->pdim, tri->edim);
	 break;
      };
   };
   
   switch(data_type){
    case SY_IPD_STRIP:
      vd.pdim  = ipd->ipodim;  
      vd.point = (SyPoint*) ipd->ipo;
      break;
    case SY_TRIAN_STRIP:
      vd.pdim  = tri->pdim;
      vd.point = tri->point;
      break;
   }
   
   vd.edge   = Sy_AllocEdge  (3*vd.pdim);
   vd.vertex = Sy_AllocVertex(2*vd.pdim);
   
   if (is_timer) Sy_InitTimer();
   
   Sy_Voronoi (&vd, is_verbose);
   
   if (is_timer){ 
      delta = Sy_GetTimer(1);
      fprintf (stderr, "Processing speed: %6.0f generators/sec\n", 
	       vd.pdim/delta.utime);
   }
   if (is_verbose) 
     fprintf (stderr, "\n points: %d  edges: %d  vertices: %d\n",
	      vd.pdim, vd.edim, vd.vdim);

   iohandle = Sy_CreateIOhandle ();
   Sy_SetIOproperty (iohandle, FORMAT, (is_ascii) ? SY_ASCII:SY_BINARY);
   Sy_SetIOproperty (iohandle, TYPE, SY_VORO_STRIP);
   Sy_SetIOproperty (iohandle, DATA, (char*) &vd);
   Sy_SetIOproperty (iohandle, IO_MODE, SY_FREE_AFTER_WRITE);
   
   Sy_SetIOproperty (iohandle_point, FORMAT, 
		     (is_ascii) ? SY_ASCII:SY_BINARY);
   
   Sy_PutIOhandle (iohandle_point);
   Sy_PutIOhandle (iohandle);
   Sy_ManageOutput (out_file);
   
   sprintf (charbuf, "@ writing Voronoi Diagram to %s\n", out_file);
   Sy_Message ("%s", charbuf);
   
   return 1;
}   



