/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#ifndef _SY_GRAPHICS_H
#define _SY_GRAPHICS_H 1

#if defined(ultrix)
#define NeedFunctionPrototypes 0 
#endif

#ifndef NOX
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif

#include "Sy_Stddef.h"

#ifdef  __cplusplus
extern "C" {
#endif

#define SY_DEFX    64
#define SY_DEFY    64
#define SY_DEFXSTR "64"
#define SY_DEFYSTR "64"
#define SY_DISPLAY_STRING_LEN  512
#define SY_HOSTNAME_LEN        512
#define SY_TITLE_LEN           512
#define SY_DEF_BORDERWIDTH     2
#define SY_ARCWIDTH            16

#define SY_XBLOCK                 4096

#define SY_DEF_WINDIM (512)
#define SY_MAX_WINDIM (800)
#define SY_MIN_WINDIM (300)
#define SY_MIN_BORDER (16)
#define SY_DEF_MINY   (0)

#define SY_VALRAD    (4096.0)
#define SY_PLANESIZE (SY_VALRAD/SY_SQRT2)

enum SyDisplayMode   {SKE_FOREGROUND, SKE_BACKGROUND, SKE_PLAIN};
enum SyDisplaySwitch {SKE_ON, SKE_OFF};

typedef struct _SyDisplayControl {
  int                   window_dimX, 
                        window_dimY;
  enum SyDisplayMode    display_mode;
  enum SyDisplaySwitch  show_boundary, 
                        raw_mode,
                        set_window;
  int                   line_width;
  float                 threshold;
  int                   rank;
  float                 scale;
} SyDisplayControl;

int Sy_XIpoDisp (SyIpoDAC *ipd, char *windowName, char *iconName, int xPos, int yPos, 
		 int xZen, int yZen, float uscale, char *userDisplay);

int Sy_XArcDisp (SyIpoDAC *ipd, char *windowName, char *iconName, 
		 int xPos, int yPos, int xZen, int yZen, float uscale, 
		 char *userDisplay);

int Sy_XVoroDisp (SyVoroDAC *vd, char *windowName, char *iconName, int xPos, int yPos, 
		  int xZen, int yZen, float uscale, char *userDisplay, 
		  SyDisplayControl control);

int Sy_XSkelDisp (SySkelDAC *vsk, SyIpoDAC *ipd, float th, char *windowName, 
		  char *iconName, int xPos, int yPos, int xZen, int yZen, 
		  float uscale, char *userDisplay, SyDisplayControl control);

int Sy_XReconstruct (SySkelDAC *vsk, char *fginfo, float thres, char *windowName, 
		     char *iconName, int xPos, int yPos, int xZen, int yZen,
		     float uscale, char *userDisplay, SyDisplayControl control);

int Sy_XPyramidDisp (SySkelDAC *vsk, SyIpoDAC *ipd, int rank, char *windowName, 
		     char *iconName, int xPos, int yPos, int xZen, int yZen, 
		     float uscale, char *userDisplay, SyDisplayControl control, 
		     int *pyrlab);

int Sy_XPyrReconstruct (SySkelDAC *vsk, SyIpoDAC *ipd, int *pyrlab, int rank, 
			float thres, char *windowName, char *iconName, 
			int xPos, int yPos, int xZen, int yZen, float uscale, 
			char *userDisplay, SyDisplayControl control);

int Sy_XDisplayHipoContainer(SyHipoContainer *container, char *windowName, 
			     char *iconName, int xPos, int yPos, 
			     int xZen, int yZen, float uscale, 
			     char *userDisplay, SyIpoMappingControl *mcontrol);

int Sy_XSkelDispWithTrian (SySkelDAC *vsk, SyTriDAC *tri, float th, 
			   char *windowName, char *iconName, int xPos, int yPos, 
			   int xZen, int yZen, float uscale, char *userDisplay, 
			   SyDisplayControl control);

int Sy_XSkeletonSetDisp (SyVSkeletonSet *vss, char *windowName, char *iconName, 
			 int xPos, int yPos, int xZen, int yZen, 
			 float uscale, char *userDisplay, SyDisplayControl control);

/* interface to idraw program */

void Sy_Ipo2Idraw(SyIpoDAC *ipd, FILE *fpout, SyBoolean internal_call);

void Sy_Voro2Idraw(SyVoroDAC *vd, FILE *fpout, int xZen, int yZen, 
		   float uscale);

void Sy_Skel2Idraw(SySkelDAC *vsk, SyIpoDAC *ipd, float thres, 
		   enum SyDisplayMode display_mode, FILE *fpout);

void Sy_SkelPyr2Idraw(SySkelDAC *vsk, SyIpoDAC *ipd, int rank, 
		       enum SyDisplayMode display_mode, 
		       int *pyrlab, FILE *fpout);

void Sy_SkeletonSet2Idraw(SyVSkeletonSet *vss, float uscale, 
			  SyDisplayControl *ctrl, FILE *fpout);

#ifdef  __cplusplus
}
#endif
#endif /* _SY_GRAPHICS_H */

