/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Division of Applied Sciences,
 *          Harvard University, Cambridge, Massachusetts (rlo@hrl.harvard.edu)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1995 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#ifndef _SY_LANG_H
#define _SY_LANG_H 1

#include <stdio.h>
#include "Sy_AvlTree.h"

#if defined(SYSV) || defined(SVR4) || defined(sparc)
#include <string.h>
#else
#include <strings.h>
#endif

#define SY_SYNC_COOKIE_BEGIN  "#2718sync314beg#"
#define SY_SYNC_COOKIE_END    "#2718sync314end#"

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct _Sy_ParseControl {
  int is_memory, 
      is_file;
  FILE *fpin;
  char *strbuf;
} Sy_ParseControl;

void Sy_ParseStrip ();

typedef enum _SyIoDataType {SY_INT_T, SY_FLOAT_T, SY_DOUBLE_T, SY_STR_T, 
			    SY_PTR_T, SY_LOADFUNC_T} SyIoDataType ;
union SyGenType {
  int    i_val;
  float  f_val;
  double d_val;
  char*  s_val;
  char*  p_val;
  char*  func_val;
};

typedef struct _SyDataObject {
  int id;
  char *name;
  SyIoDataType tag;
  union SyGenType t;
} SyDataObject;

typedef struct _SyIoObject {
  int id;
  char *name;
  SyAvlTree *tree;
} SyIoObject;

typedef struct _SyIoToken {
  SyAvlTree *yacc_tree,
            *func_tree,
            *item_tree;
  void      *genptr;
  FILE      *in, *out;
  char      *infilename, *outfilename;
  void      *memptr;
  void      *shmemptr;
  void      *extra;
  char      *funcname;
} SyIoToken;


SyIoObject *Sy_CreateIoObject(char *str, int in_id);

SyDataObject *Sy_CreateDataObject(char *str, int in_id, 
				  SyIoDataType ty, void *val);

int  Sy_IoObjCompare (void* l, void* r);
int  Sy_DataObjCompare (void* l, void* r);
int  Sy_DataObjDelete (void* d);
int  Sy_IoObjDelete (void* l);
int  Sy_IoObjPrint (void* l);

int Sy_AddIoObject (SyAvlTree **objtree, char *objname);

int Sy_AddDataObject (SyAvlTree **objtree, char *objname, 
		      char *dataname, SyIoDataType ty, void *val);

int Sy_AddDataSameObject (SyAvlTree **objtree, char *dataname, 
			  SyIoDataType ty, void *val);

SyIoObject* Sy_GetIoObject (SyAvlTree **objtree, char *objname);

SyDataObject* Sy_GetDataObject (SyAvlTree **objtree, char *objname, 
		      char *dataname);

SyDataObject* Sy_GetDataSameObject (SyAvlTree **objtree, char *dataname);

int  Sy_DataObjectInfo (void *din);
int  Sy_IoObjInfo (void* l);
void Sy_IoObjectTreeInfo (SyAvlTree **objtree);

int Sy_DataObjPurge (void *data, void *key);
int Sy_IoObjPurge (void *obj, void *key);

void Sy_DeleteIoObjectTree(SyAvlTree **objtree);

int Sy_SyncWithStream(FILE *stream, char* cookie);

void SY_Wrbegload(FILE *out, char* callee, int no_cookie, int is_ascii);
void SY_Wrendload(FILE *out, int no_cookie);

#ifndef SY_MSG
#ifdef SY_DEBUG
#define	SY_MSG(msg) printf("DEBUG: '%s'\n", msg)
#else
#define	SY_MSG(msg)
#endif
#endif

#ifdef  __cplusplus
}
#endif

#endif /* _SY_LANG_H */

/* loader template 
int functionname (SyIoToken *tok)
{
  int is_sync;

#ifdef SY_DEBUG
  printf ("Entering %s...\n", tok->funcname);
#endif

  is_sync = Sy_SyncWithStream(tok->in, SY_SYNC_COOKIE_BEGIN);

  if (is_sync){

  write here your code to load user data...
  in case of an error return a zero;

  }
  else{ return is_sync; }

  is_sync = Sy_SyncWithStream(tok->in, SY_SYNC_COOKIE_END);

#ifdef SY_DEBUG
  printf ("Exiting %s...\n", tok->funcname);
#endif
  return is_sync;  
}
*/
