/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#ifndef _SY_STDDEF_H
#define _SY_STDDEF_H 1

#ifdef  __cplusplus
extern "C" {
#endif

#ifndef max
#define max(x,y) ((x) > (y) ? (x) :  (y))
#endif

#ifndef min
#define min(x,y) ((x) < (y) ? (x) :  (y))
#endif

#ifndef sq
#define sq(x) ((x)*(x))
#endif

#ifndef SY_NORM
#define SY_NORM(x,y) (sqrt((x)*(x)+(y)*(y)))
#endif

#ifndef SY_NORM2
#define SY_NORM2(x,y) ((x)*(x)+(y)*(y))
#endif

#ifndef SY_SQUARE
#define SY_SQUARE(x) ((x)*(x))
#endif

#ifndef SY_ALV
#define SY_ALV(x)   ((x) >= 0  ? (x) : -(x))
#endif

#ifndef SY_DET
#define SY_DET(x1,y1,x2,y2) ((x1)*(y2)-(x2)*(y1)) 
#endif

#ifndef SY_SGN
#define SY_SGN(x)    (((x) == 0)? 0:(((x) > 0)? 1:-1))
#endif

#define SY_TRUE       (1)
#define SY_FALSE      (0)
#define SY_FOREGRND   (1)
#define SY_BACKGRND   (-1)
#define SY_DELETEMARK (1)
#define SY_TMPMARK    (-1)
#define SY_BORDERMARK (-1)
#define SY_VOIDLABEL  (-1)

#define  SY_SQRT2         1.41421356237309504880
#define  SY_PI            3.14159265358979323846
#define  SY_SQRTE         1.648721271
#define  SY_PI_2       1.57079632679489661923
#define  SY_FRAC_2_PI     0.63661977

#define SY_RES_EPS        0.05
#define SY_LOC_EPS        1.0
#define SY_SMALL_OFFSET   0.1

#define  SY_NEG_INFTY      (-1.0e10)
#define  SY_POS_INFTY      (1.0e10)
#define  SY_NEG_INT_INFTY  (-100000000)
#define  SY_POS_INT_INFTY  (100000000)
#define  SY_FIX_THRES      (0.5e10)

#define SY_INIT_RES_TAB_SIZE  (3)
#define SY_RES_TAB_INCR       (2)
#define SY_INIT_SKEL_TAB_SIZE (256)
#define SY_SKEL_TAB_INCR      (256)
#define SY_TPYR_MIN           (10.0)

#define SY_MICRO              (1000000.0)
#define SY_ROTORSPEED          1000
#define SY_RODIM               4

#define SY_SMCIR_SEARCH_THRES  (3.0)
#define SY_SMCIR_TERM_THRES    (2.0)

#define SY_SMALL_STRING_LEN   (32)
#define SY_FILENAME_LEN       (256)

#define SY_INCH2IPO(x)       ((x)*72.0)

#define SY_IPO_STRIP_INITSIZE (4000)

#define SY_NBUFF              (10)

#define SY_UINT  unsigned

#ifdef __STDC__
#define SY_VOID void
#define SY_CONST_VOID const void
#else
#define SY_VOID char
#define SY_CONST_VOID char
#endif

#define SY_UCHAR unsigned char
#define UCHAR unsigned char
#define COORD    short
#define SyIndex  int

#define SY_FMTRAW "%s"
#define SY_FMT    "%s"
#define SY_FMTNL  "%s\n"
#define SY_GLOBALSTRBUFLEN 1024
#define SY_CHARBUF_SIZE    1024

typedef unsigned char Boolean;
typedef unsigned char SyBoolean;

/**********************************************************************
 * SyByteVector
 *********************************************************************/
typedef struct
{
   char x, y;
} SyByteVector;

typedef struct 
{
   float x, y;
} SyFloatVector;

/**********************************************************************
 * SyVector
 *********************************************************************/
typedef 
struct{
	double dirx, diry;
} SyVector;

/**********************************************************************
 * SyIntVector
 *********************************************************************/
typedef 
struct{
	int dirx, diry;
} SyIntVector;

typedef struct _SyIntLocvec {
  int dirx, diry;
  int xoff, yoff;
} SyIntLocvec;

typedef struct _SyLocvec {
  float dirx, diry;
  float xoff, yoff;
} SyLocvec;

typedef struct _SyDblLocvec {
  double dirx, diry;
  double xoff, yoff;
} SyDblLocvec;

/**********************************************************************
 * SyFraction
 *********************************************************************/
typedef 
struct{
	double den, num;     /* denominator, numerator */
} SyFraction;


typedef struct _SyFloatPoint{
  float x, y;
} SyFloatPoint;

typedef struct _SyDblPoint{
  double x, y;
} SyDblPoint;

void Sy_NormVector (SyVector *v);

#ifdef  __cplusplus
}
#endif
#endif /* _SY_STDDEF_H */





