/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * Handling of Skeleton Boundary
 * 
 */

static char sy_BorderId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

typedef struct _SyRotInfo{
  int last_edge;
} SyRotInfo;

static int Sy_Intersection(SyEdge *e1, SyEdge *e2, SyVertex *v);
static int Sy_FullMember(SyEdge *chain, SyEdge *e, SyVertex *vertex);
static int Sy_Right(SyEdge *chain, SyEdge *e, SyVertex *vertex, int *vright);
static void Sy_SimpleRotate(SyEdge *edge, int ref_edge, int ref_vertex, 
SyRotInfo *info);

#ifdef stardent
#pragma OPT_LEVEL 0
#endif
void Sy_SkelBorder (SySkelDAC *vsk, SyIpoDAC *ipd, int is_verbose)
{

  int    i, corrcount = 0; 
  SyPoint  lpoint, rpoint; 
  SyEdge   new_edge, pedge; 
  SyVertex tvertex, hvertex, new_vertex, vtmp; 
  SyIntVector pvec;
  SyVector  evec;
  SyRotInfo rotinfo;
  int    left_point, right_point; 
  int    tail_vertex, head_vertex, vindex, vindex2;
  int    tail_edge, head_edge, eindex;
  int    save_vertex, save_edge, ref_vertex, ref_edge,
         next_edge, last_edge;
  int    conn_flag, conn[4];
  int    edim, vdim, pdim;
  int    vright;

  Sy_Message (SY_FMTNL, "Boundary Vertex Completion ");

  vdim = vsk->vdim;
  edim = vsk->edim;
  pdim = vsk->pdim;
 
  vsk->edge   = Sy_ReallocEdge   (vsk->edge, edim + pdim);
  vsk->vertex = Sy_ReallocVertex (vsk->vertex, vdim + 2*pdim);
  vsk->is_border = 1;

  for (i = 1; i <= vsk->edim; i++){

    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;

    conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));
    conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));
    conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].head_icn != 0));
    conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].head_icn != 0));
    
    conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);

    if (conn_flag == 0) continue;

    head_vertex = vsk->edge[i].head_vertex;
    tail_vertex = vsk->edge[i].tail_vertex;
    head_edge   = vsk->edge[i].head_edge;
    tail_edge   = vsk->edge[i].tail_edge;

    lpoint  = vsk->point[left_point];
    rpoint  = vsk->point[right_point];
    tvertex = vsk->vertex[tail_vertex];
    hvertex = vsk->vertex[head_vertex];

    if (vsk->edge[i].rank > 0){
      evec.dirx = hvertex.xden/hvertex.num - tvertex.xden/tvertex.num;
      evec.diry = hvertex.yden/hvertex.num - tvertex.yden/tvertex.num;
      pvec.dirx = lpoint.x - rpoint.x;
      pvec.diry = lpoint.y - rpoint.y;
      if ((evec.dirx * pvec.diry - evec.diry * pvec.dirx) > 0){
	pedge.dirx = pvec.dirx;
	pedge.diry = pvec.diry;
	pedge.xoff = rpoint.x;
	pedge.yoff = rpoint.y;
      }
      else{	
	pedge.dirx = - pvec.dirx;
	pedge.diry = - pvec.diry;
	pedge.xoff = lpoint.x;
	pedge.yoff = lpoint.y;
      }
    }
    
    switch(vsk->edge[i].rank){
    case 2:
      
      if (Sy_FullMember (&pedge, &(vsk->edge[i]), vsk->vertex)){
      
	vindex  = ++vdim;
	vindex2 = ++vdim;
	eindex  = ++edim;

	save_vertex = head_vertex;
	save_edge   = head_edge;
	ref_edge    = i;
	ref_vertex  = head_vertex;
	
	vsk->vertex[vindex].xden = lpoint.x + rpoint.x;
	vsk->vertex[vindex].yden = lpoint.y + rpoint.y;
	vsk->vertex[vindex].num  = 2.0;

	vsk->vertex[vindex2] = vsk->vertex[vindex];

	Sy_SimpleRotate (vsk->edge, ref_edge, ref_vertex, &rotinfo);

	if (vsk->edge[rotinfo.last_edge].tail_vertex == ref_vertex){
	  vsk->edge[rotinfo.last_edge].tail_edge = eindex;
	}
	else{
	  vsk->edge[rotinfo.last_edge].head_edge = eindex;
	}
	
	vsk->edge[i].head_vertex = vindex;
	vsk->edge[i].head_edge   = 0;
        
	vsk->edge[eindex] = vsk->edge[i];
	
	vsk->edge[eindex].tail_edge   = save_edge;
	vsk->edge[eindex].head_edge   = 0;
	vsk->edge[eindex].tail_vertex = save_vertex;
	vsk->edge[eindex].head_vertex = vindex2;
	vsk->edge[eindex].left_point  = right_point;
	vsk->edge[eindex].right_point = left_point;
      }
      else{
	if (Sy_Right(&pedge, &(vsk->edge[i]), vsk->vertex, &vright)){
	  
	  ref_edge    = i;
	  ref_vertex  = tail_vertex;
	  vindex      = ++vdim;

	  vsk->vertex[vindex] = vsk->vertex[ref_vertex];

	  Sy_SimpleRotate (vsk->edge, ref_edge, ref_vertex, &rotinfo);

	  if (vsk->edge[rotinfo.last_edge].tail_vertex == ref_vertex){
	    vsk->edge[rotinfo.last_edge].tail_edge = tail_edge;
	  }
	  else{
	    vsk->edge[rotinfo.last_edge].head_edge = tail_edge;
	  }
	  vsk->edge[i].tail_edge   = 0;
	  vsk->edge[i].tail_vertex = vindex;
	}
	else{
	  ref_edge    = i;
	  ref_vertex  = head_vertex;
	  vindex      = ++vdim;

	  vsk->vertex[vindex] = vsk->vertex[ref_vertex];

	  Sy_SimpleRotate (vsk->edge, ref_edge, ref_vertex, &rotinfo);

	  if (vsk->edge[rotinfo.last_edge].tail_vertex == ref_vertex){
	    vsk->edge[rotinfo.last_edge].tail_edge = head_edge;
	  }
	  else{
	    vsk->edge[rotinfo.last_edge].head_edge = head_edge;
	  }
	  vsk->edge[i].head_edge   = 0;
	  vsk->edge[i].head_vertex = vindex;
	}
      }
      break;
    case 1:

      if (Sy_FullMember (&pedge, &(vsk->edge[i]), vsk->vertex)){

	vindex  = ++vdim;
	vindex2 = ++vdim;
	eindex  = ++edim;

	save_vertex = head_vertex;
	
	vsk->vertex[vindex].xden = lpoint.x + rpoint.x;
	vsk->vertex[vindex].yden = lpoint.y + rpoint.y;
	vsk->vertex[vindex].num  = 2.0;

	vsk->vertex[vindex2] = vsk->vertex[vindex];

	vsk->edge[i].head_vertex = vindex;
	vsk->edge[i].head_edge   = 0;
	vsk->edge[i].rank        = 2;

	vsk->edge[eindex] = vsk->edge[i];
	
	vsk->edge[eindex].tail_edge = 0;
	vsk->edge[eindex].head_edge = 0;
	vsk->edge[eindex].tail_vertex = vindex2;
	vsk->edge[eindex].head_vertex = save_vertex;
	vsk->edge[eindex].rank = 1;
	
	vtmp = vsk->vertex[vindex2];
	vtmp.xden = vsk->edge[eindex].dirx * vtmp.num + vtmp.xden;
	vtmp.yden = vsk->edge[eindex].diry * vtmp.num + vtmp.yden;
	vsk->vertex[save_vertex] = vtmp;
      }
      else{
	ref_edge    = i;
	ref_vertex  = tail_vertex;
	vindex      = ++vdim;

	vsk->vertex[vindex] = vsk->vertex[ref_vertex];
	
	Sy_SimpleRotate (vsk->edge, ref_edge, ref_vertex, &rotinfo);

	if (vsk->edge[rotinfo.last_edge].tail_vertex == ref_vertex){
	  vsk->edge[rotinfo.last_edge].tail_edge = tail_edge;
	}
	else{
	  vsk->edge[rotinfo.last_edge].head_edge = tail_edge;
	}
	vsk->edge[i].tail_edge   = 0;
	vsk->edge[i].tail_vertex = vindex;
      }
      break;
    case 0:
      vindex  = ++vdim;
      vindex2 = ++vdim;
      eindex  = ++edim;

      vsk->vertex[vindex].xden = lpoint.x + rpoint.x;
      vsk->vertex[vindex].yden = lpoint.y + rpoint.y;
      vsk->vertex[vindex].num  = 2.0;

      vsk->vertex[vindex2] = vsk->vertex[vindex];

      vsk->edge[i].head_vertex = tail_vertex;
      vsk->edge[i].tail_vertex = vindex;
      vsk->edge[i].tail_edge   = eindex;
      vsk->edge[i].rank        = 1;

      vtmp.xden = vsk->edge[i].dirx + vsk->vertex[vindex].xden;
      vtmp.yden = vsk->edge[i].diry + vsk->vertex[vindex].yden;
      vtmp.num  = 2.0;
      vsk->vertex[head_vertex] = vtmp;

      vsk->edge[eindex] = vsk->edge[i];

      vsk->edge[eindex].tail_edge = i;
      vsk->edge[eindex].head_edge = 0;
      vsk->edge[eindex].tail_vertex = vindex2;
      vsk->edge[eindex].head_vertex = head_vertex;
    
      vtmp.xden = -1.0*vsk->edge[eindex].dirx + vsk->vertex[vindex2].xden;
      vtmp.yden = -1.0*vsk->edge[eindex].diry + vsk->vertex[vindex2].yden;
      vsk->vertex[head_vertex] = vtmp;
      break;     
    }
  }
  sprintf (syStrBuf, "added %d edges, %d vertices ", 
	   edim - vsk->edim, vdim - vsk->vdim); 
  Sy_Message (SY_FMTNL, syStrBuf);

  vsk->vdim = vdim;
  vsk->edim = edim;
}


/**********************************************************************
 * function int Sy_Intersection (e1, e2, v)
 *
 * purpose: computes the Sy_Intersection of two lines, returns 0 in
 *          the case of parallel lines
 *
 * param:   e1, e2 - lines (infinite edges)
 *          v      - resulting vertex
 *********************************************************************/
static int Sy_Intersection (SyEdge *e1, SyEdge *e2, SyVertex *v)
{
        double          tden;

        (*v).num = (*e2).dirx * (*e1).diry  - (*e2).diry * (*e1).dirx;
        if ( (*v).num == 0 ) return (0);

        if ( (*v).num < 0 ) {
                (*v).num = - (*v).num;
                tden =  (*e2).diry * ( (*e2).xoff - (*e1).xoff )
                      + (*e2).dirx * ( (*e1).yoff - (*e2).yoff );
        }
        else
                tden =  (*e2).diry * ( (*e1).xoff - (*e2).xoff )
                      + (*e2).dirx * ( (*e2).yoff - (*e1).yoff );

        (*v).xden =  (*e1).dirx * tden + (*e1).xoff * (*v).num;
        (*v).yden =  (*e1).diry * tden + (*e1).yoff * (*v).num;

        return (1);
}


/**********************************************************************
 * function int Sy_FullMember (chain, e, vertex)
 *
 * purpose: tests whether the vertex resulting from the Sy_Intersection
 *          of the dividing chain with an edge is a (true) member of that 
 *          edge, returns 1 if positive          
 *
 * param:   chain  - dividing chain
 *          e      - edge
 *          vertex - vertex array
 *********************************************************************/
static int Sy_FullMember (SyEdge *chain, SyEdge *e, SyVertex *vertex)
{
        double tmp1, tmp2;
        SyVertex  vt, vh;

        switch ((*e).rank) {
        case 2:
                vt   = vertex[(*e).tail_vertex];
                vh   = vertex[(*e).head_vertex];
                tmp1 = ((*chain).dirx * ( vt.yden - (*chain).yoff * vt.num ))
                     - ((*chain).diry * ( vt.xden - (*chain).xoff * vt.num ));
                if (tmp1 == 0.0) return (0);
                tmp2 = ((*chain).dirx * ( vh.yden - (*chain).yoff * vh.num ))
                     - ((*chain).diry * ( vh.xden - (*chain).xoff * vh.num ));
                if (tmp2 == 0.0) return (0);
                return ((tmp1 < 0.0) != (tmp2 < 0.0));
                break;
        case 1:
                vt   = vertex[(*e).tail_vertex];
                tmp1 = ((*chain).dirx * ( vt.yden - (*chain).yoff * vt.num ))
                     - ((*chain).diry * ( vt.xden - (*chain).xoff * vt.num ));
                if (tmp1 == 0.0) return (0);
                tmp2 = ((*chain).diry * (*e).dirx)
                     - ((*chain).dirx * (*e).diry);
                if (tmp2 == 0.0) return (0);
                return ((tmp1 < 0.0) == (tmp2 < 0.0));
                break;
        case 0:
                return (1); break;
        }
}


/**********************************************************************
 * function int Sy_Right (chain, e, vertex, vright)
 *
 * purpose: tests whether edge `e` is on the right side with respect
 *          to `chain`, an element of the dividing chain
 *          returns 1 if positive
 *
 * input:   chain  - element of the dividing chain
 *          e      - edge to be tested
 *          vertex - vertex array
 *          vright - vertex located on the right side of `chain`
 *********************************************************************/
static int Sy_Right (SyEdge *chain, SyEdge *e, SyVertex *vertex, int *vright)
{
        int    res1, res2;
        SyVertex  vt, vh;

        switch ((*e).rank) {
        case 2:
                vt = vertex[(*e).tail_vertex];
                vh = vertex[(*e).head_vertex];
                res1 = ((*chain).dirx * ( vt.yden - (*chain).yoff * vt.num ))
                    <= ((*chain).diry * ( vt.xden - (*chain).xoff * vt.num ));
                res2 = ((*chain).dirx * ( vh.yden - (*chain).yoff * vh.num ))
                    <= ((*chain).diry * ( vh.xden - (*chain).xoff * vh.num ));
                if (res1) *vright = (*e).tail_vertex;
                if (res2) *vright = (*e).head_vertex;
                return (res1 && res2);
                break;
        case 1:
                vt = vertex[(*e).tail_vertex];
                res1 = ((*chain).dirx * ( vt.yden - (*chain).yoff * vt.num ))
                    <= ((*chain).diry * ( vt.xden - (*chain).xoff * vt.num ));
                *vright = (*e).tail_vertex;
                return (res1);
                break;
        case 0:
                *vright = 0; return (0); break;
        }
}

static void Sy_SimpleRotate (SyEdge *edge, int ref_edge, int ref_vertex, SyRotInfo *info)
{
  int next_edge, last_edge;

  next_edge   = (edge[ref_edge].tail_vertex == ref_vertex) ? edge[ref_edge].tail_edge:
                                                             edge[ref_edge].head_edge; 

  if (next_edge == 0){
    info->last_edge = ref_edge; 
    return;
  }

  while (next_edge != ref_edge){
    last_edge = next_edge;
    if (ref_vertex == edge[next_edge].tail_vertex){
      next_edge = edge[next_edge].tail_edge;	      
    }
    else{
      next_edge = edge[next_edge].head_edge;
    }
  }
  
  info->last_edge = last_edge;
}
