/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Graphics.h"
#include "Sy_Idraw.h"

void Sy_Ipo2IdrawSimple (SyIpoDAC *ipd, FILE *fpout)
{
  int      x1, x2, y1, y2;

  float    scaleX, scaleY, scaleXY;
  float    xOff,   yOff, 
           xShift, yShift;

  int      maxX, minX, maxY, minY;
  int      dimX, dimY; 
  int      effDimX, effDimY;

  int      tail_ipo, head_ipo;
  int      i;

  maxX = minX = ipd->ipo[1].x;
  minY = ipd->ipo[1].y;
  maxY = ipd->ipo[ipd->ipodim].y;

  for (i = 1; i <= ipd->ipodim; i++){
     maxX = max(maxX, ipd->ipo[i].x);
     minX = min(minX, ipd->ipo[i].x);
  }

  dimX = maxX - minX;
  dimY = maxY - minY;

  effDimX = max(min(dimX, SY_IDRAW_DEF_XBBOX), SY_IDRAW_MIN_XBBOX);
  effDimY = max(min(dimY, SY_IDRAW_DEF_YBBOX), SY_IDRAW_MIN_YBBOX);
  scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
  scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
  scaleXY = min(scaleX, scaleY);
  xOff    = dimX/2 + minX;
  yOff    = dimY/2 + minY;

  xShift  = effDimX/2;
  yShift  = effDimY/2;

  Sy_IdrawSetCreator("Function Sy_Ipo2Idraw");
  Sy_IdrawSetBoundingBox(0,0,effDimX,effDimY);
  Sy_IdrawPrologue(fpout);
  
  Sy_IdrawSetXYScale(1.0, 1.0);
  Sy_IdrawSetXYTranslate(0,0);
  Sy_IdrawSetCurrentFGColor("Blue", 0,0,1);

  Sy_IdrawBeginGroup();

  for (i = 1; i <= ipd->ipodim; i++){
    x1 = scaleXY * ((float)ipd->ipo[i].x - xOff) + xShift;
    y1 = scaleXY * ((float)ipd->ipo[i].y - yOff) + yShift;
    Sy_IdrawEllipse(x1, y1, 1, 1);
  }

  Sy_IdrawEndGroup();

  Sy_IdrawSetXYScale(scaleXY, scaleXY);
  Sy_IdrawSetXYTranslate(xShift-scaleXY*xOff, yShift-scaleXY*yOff);
  Sy_IdrawSetCurrentFGColor("Black", 0,0,0);

  Sy_IdrawBeginGroup();
  for (i = 1; i <= ipd->icndim; i++) {
    if ((tail_ipo = ipd->icn[i].tail_ipo) && 
	(head_ipo = ipd->icn[i].head_ipo)){
      x1 = ipd->ipo[tail_ipo].x;
      y1 = ipd->ipo[tail_ipo].y; 
      x2 = ipd->ipo[head_ipo].x;
      y2 = ipd->ipo[head_ipo].y;
      Sy_IdrawLine(x1, y1, x2, y2); 
    }
  }
  Sy_IdrawEndGroup(); 
  Sy_IdrawEpilogue();
}

#define SY_IPOBUFFDIM 1024
static SyWalktask wt;
static int contourdim = 0, ipomaxdim = 0, ipocount = 0;
static SyIdrawPoint *idraw_point = NULL;

static void Sy_Ipo2IdrawOutputWalk(void)
{
  if (ipocount >= ipomaxdim){
    ipomaxdim += SY_IPOBUFFDIM;
    idraw_point = Sy_ReallocIdrawPoint (idraw_point, ipomaxdim);
  }
  idraw_point[ipocount].x = wi.ipd->ipo[wi.current.head_ipo].x;
  idraw_point[ipocount].y = wi.ipd->ipo[wi.current.head_ipo].y;
  ipocount++;
}

static void Sy_Ipo2IdrawOutputTrav(void)
{
  ipocount = 0; contourdim++;
  if (idraw_point == NULL){ 
    ipomaxdim = SY_IPOBUFFDIM;
    idraw_point = Sy_AllocIdrawPoint(ipomaxdim);
  }
}

static void Sy_Ipo2IdrawOutputTerm(void)
{
  if (ipocount >= ipomaxdim){
    ipomaxdim += SY_IPOBUFFDIM;
    idraw_point = Sy_ReallocIdrawPoint (idraw_point, ipomaxdim);
  }
  idraw_point[ipocount++] = idraw_point[0];
  Sy_IdrawMLine(idraw_point, ipocount);
}

void Sy_Ipo2Idraw (SyIpoDAC *ipd, FILE *fpout, SyBoolean internal_call)
{
  FILE     *farr[1];

  int      x1, x2, y1, y2;

  float    scaleX, scaleY, scaleXY;
  float    xOff,   yOff, 
           xShift, yShift;

  int      maxX, minX, maxY, minY;
  int      dimX, dimY; 
  int      effDimX, effDimY;

  int      tail_ipo, head_ipo;
  int      i;

  wt.user_file    = farr;
  wt.user_file[0] = fpout;
  wt.fildim       = 1;
  wt.user_function[WALK] = (void(*)())Sy_Ipo2IdrawOutputWalk;
  wt.user_function[TRAV] = (void(*)())Sy_Ipo2IdrawOutputTrav;
  wt.user_function[TERM] = (void(*)())Sy_Ipo2IdrawOutputTerm;

  maxX = minX = ipd->ipo[1].x;
  minY = ipd->ipo[1].y;
  maxY = ipd->ipo[ipd->ipodim].y;

  for (i = 1; i <= ipd->ipodim; i++){
     maxX = max(maxX, ipd->ipo[i].x);
     minX = min(minX, ipd->ipo[i].x);
  }

  dimX = maxX - minX;
  dimY = maxY - minY;

  effDimX = max(min(dimX, SY_IDRAW_DEF_XBBOX), SY_IDRAW_MIN_XBBOX);
  effDimY = max(min(dimY, SY_IDRAW_DEF_YBBOX), SY_IDRAW_MIN_YBBOX);
  scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
  scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
  scaleXY = min(scaleX, scaleY);
  xOff    = dimX/2 + minX;
  yOff    = dimY/2 + minY;

  xShift  = effDimX/2;
  yShift  = effDimY/2;

  if (!internal_call){
    Sy_IdrawSetCreator("Function Sy_Ipo2Idraw");
    Sy_IdrawSetBoundingBox(0,0,effDimX,effDimY);
    Sy_IdrawPrologue(fpout);
  }

  Sy_IdrawSetXYScale(scaleXY, scaleXY);
  Sy_IdrawSetXYTranslate(xShift-scaleXY*xOff, yShift-scaleXY*yOff);
  Sy_IdrawSetCurrentFGColor("Blue", 0,0,1);
  
  Sy_InitIpoTraverse (ipd, &wt);
  Sy_IpoTraverse (ipd, &wt);  

  if (idraw_point) free(idraw_point);
  
  if (!internal_call) Sy_IdrawEpilogue();
}


void Sy_Voro2Idraw(SyVoroDAC *vd, FILE *fpout, int xZen, int yZen, float uscale)
{

  SyVertex vhead, vtail;
  float    x1, y1, x2, y2;

  float    scaleX, scaleY, scaleXY;
  float    xOff,   yOff, 
           xShift, yShift;

  int      maxX, minX, maxY, minY;
  int      dimX, dimY; 
  int      effDimX, effDimY;

  int      tail_ipo, head_ipo;
  int      i;

  maxX = minX = vd->point[1].x;
  minY = vd->point[1].y;
  maxY = vd->point[vd->pdim].y;

  for (i = 1; i <= vd->pdim; i++){
     maxX = max(maxX, vd->point[i].x);
     minX = min(minX, vd->point[i].x);
  }

  dimX = maxX - minX;
  dimY = maxY - minY;

  if (uscale > 0){
    effDimX = max(min(dimX * uscale, SY_IDRAW_DEF_XBBOX), SY_IDRAW_MIN_XBBOX);
    effDimY = max(min(dimY * uscale, SY_IDRAW_DEF_YBBOX), SY_IDRAW_MIN_YBBOX);
    xOff    = (xZen > 0) ? xZen : dimX/2 + minX;
    yOff    = (yZen > 0) ? yZen : dimY/2 + minY;
    scaleXY = uscale;
  }
  else{
    effDimX = max(min(dimX, SY_IDRAW_DEF_XBBOX), SY_IDRAW_MIN_XBBOX);
    effDimY = max(min(dimY, SY_IDRAW_DEF_YBBOX), SY_IDRAW_MIN_YBBOX);
    scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
    scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
    scaleXY = min(scaleX, scaleY);
    xOff    = dimX/2 + minX;
    yOff    = dimY/2 + minY;
  }

  xShift  = effDimX/2;
  yShift  = effDimY/2;

  Sy_IdrawSetCreator("Function Sy_Voro2Idraw");
  Sy_IdrawSetBoundingBox(0,0,effDimX,effDimY);
  Sy_IdrawPrologue(fpout);
  
  Sy_SetSimpleViewport(effDimX, effDimY);
  Sy_SetVectorProc(Sy_IdrawLine);

  Sy_IdrawSetXYScale(1.0, 1.0);
  Sy_IdrawSetXYTranslate(0,0);

  Sy_IdrawSetCurrentFGColor("Red", 1,0,0);
  Sy_IdrawBeginGroup();
  for (i = 1; i <= vd->pdim; i++){
    x1 = scaleXY * ((float)vd->point[i].x - xOff) + xShift;
    y1 = scaleXY * ((float)vd->point[i].y - yOff) + yShift;
    Sy_IdrawEllipse(x1, y1, 1, 1);
  }
  Sy_IdrawEndGroup();

  Sy_IdrawSetCurrentFGColor("Blue", 0,0,1);
  Sy_IdrawBeginGroup();

  for (i = 1; i <= vd->vdim; i++) {
    x1 = scaleXY * ((vd->vertex[i].xden/vd->vertex[i].num) - xOff) + xShift;
    y1 = scaleXY * ((vd->vertex[i].yden/vd->vertex[i].num) - yOff) + yShift;
    if (Sy_IdrawIsInBoundingBox(x1, y1)){
      Sy_IdrawEllipse(x1, y1, 1, 1);
    }
  }
  Sy_IdrawEndGroup();

  Sy_IdrawSetCurrentFGColor("Black", 0,0,0);

  Sy_IdrawBeginGroup();

  for (i = 1; i <= vd->edim; i++) {
    vtail = vd->vertex[vd->edge[i].tail_vertex];
    vhead = vd->vertex[vd->edge[i].head_vertex];

    x1 = scaleXY * ((vtail.xden / vtail.num) - xOff) + xShift;
    y1 = scaleXY * ((vtail.yden / vtail.num) - yOff) + yShift;

    x2 = scaleXY * ((vhead.xden / vhead.num) - xOff) + xShift;
    y2 = scaleXY * ((vhead.yden / vhead.num) - yOff) + yShift;

    switch (vd->edge[i].rank) {
    case 0:
      Sy_DrawClippedLine(x1, y1, vd->edge[i].dirx, 
			 vd->edge[i].diry); break;
    case 1: 
      Sy_DrawClippedRay(x1, y1, vd->edge[i].dirx, 
			vd->edge[i].diry); break;
    default:
      Sy_DrawClippedVector (x1, y1, x2, y2);
      break;
    }
  }
  Sy_IdrawEndGroup(); 
  Sy_IdrawEpilogue();
}


void Sy_Skel2Idraw(SySkelDAC *vsk, SyIpoDAC *ipd, float thres, 
		   enum SyDisplayMode display_mode, FILE *fpout)
{
  SyVertex   vhead, vtail;

  float    x1, y1, x2, y2;
  int      conn_flag;
  int      conn[4];
  int      tail_vertex, head_vertex, 
           left_point, right_point,
           tail_ipo, head_ipo;
  int      tail_flag, head_flag;
  int      is_residual_cut, is_skel, is_foreground, is_render;

  float    scaleX, scaleY, scaleXY;
  float    xOff,   yOff, 
           xShift, yShift;

  int      maxX, minX, maxY, minY;
  int      dimX, dimY; 
  int      effDimX, effDimY;

  int      i;

  maxX = minX = vsk->point[1].x;
  minY = vsk->point[1].y;
  maxY = vsk->point[vsk->pdim].y;

  for (i = 1; i <= vsk->pdim; i++){
     maxX = max(maxX, vsk->point[i].x);
     minX = min(minX, vsk->point[i].x);
  }

  dimX = maxX - minX;
  dimY = maxY - minY;

  effDimX = max(min(dimX, SY_IDRAW_DEF_XBBOX), SY_IDRAW_MIN_XBBOX);
  effDimY = max(min(dimY, SY_IDRAW_DEF_YBBOX), SY_IDRAW_MIN_YBBOX);
  scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
  scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
  scaleXY = min(scaleX, scaleY);
  xOff    = dimX/2 + minX;
  yOff    = dimY/2 + minY;

  xShift  = effDimX/2;
  yShift  = effDimY/2;

  Sy_IdrawSetCreator("Function Sy_Skel2Idraw");
  Sy_IdrawSetBoundingBox(0,0,effDimX+1,effDimY+1);
  Sy_IdrawPrologue(fpout);

  Sy_Ipo2Idraw (ipd, fpout, SY_TRUE);

  Sy_IdrawSetXYScale(1.0, 1.0);
  Sy_IdrawSetXYTranslate(0,0);

  Sy_IdrawSetCurrentFGColor("Red", 0,0,1);
  Sy_IdrawBeginGroup();

  Sy_SetSimpleViewport(effDimX+5, effDimY+5);
  Sy_SetVectorProc(Sy_IdrawLine);

  for (i = 1; i <= vsk->edim; i++) {
    
    tail_vertex = vsk->edge[i].tail_vertex;
    head_vertex = vsk->edge[i].head_vertex;
    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;

    is_foreground  = Sy_SkeletonType (vsk, ipd, tail_vertex, i);

    if ((is_foreground && (display_mode == SKE_BACKGROUND)) ||
        (!is_foreground && (display_mode == SKE_FOREGROUND))) continue;

    tail_flag   = vsk->regul[tail_vertex].birth[2] > thres;
    head_flag   = vsk->regul[head_vertex].birth[2] > thres;

    conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);

    
    is_residual_cut = vsk->is_eresid && 
                      ((vsk->eresid[i].tail_residual > thres) || 
                       (vsk->eresid[i].head_residual > thres));

    is_skel         = tail_flag && head_flag;

    is_render = (is_residual_cut || is_skel) && (!conn_flag || vsk->is_border);

    if (is_render){

      if (!(tail_flag && head_flag)){
	Sy_SimpleResidualCut (vsk, i, thres, &vtail, &vhead);
      }
      else{
	vtail = vsk->vertex[vsk->edge[i].tail_vertex];
	vhead = vsk->vertex[vsk->edge[i].head_vertex];
      }

      x1 = scaleXY * ((vtail.xden / vtail.num) - xOff) + xShift;
      y1 = scaleXY * ((vtail.yden / vtail.num) - yOff) + yShift;
      
      x2 = scaleXY * ((vhead.xden / vhead.num) - xOff) + xShift;
      y2 = scaleXY * ((vhead.yden / vhead.num) - yOff) + yShift;

      switch (vsk->edge[i].rank) {
      case 0:
	Sy_DrawClippedLine(x1, y1, vsk->edge[i].dirx, 
			   vsk->edge[i].diry); break;
      case 1: 
	Sy_DrawClippedRay(x1, y1, vsk->edge[i].dirx, 
			  vsk->edge[i].diry); break;
      default:
	Sy_DrawClippedVector (x1, y1, x2, y2);
	break;
      }
    }
  }
  Sy_IdrawEndGroup(); 
  Sy_IdrawEpilogue();
}


void Sy_SkelPyr2Idraw(SySkelDAC *vsk, SyIpoDAC *ipd, int rank, 
		       enum SyDisplayMode display_mode, 
		       int *pyrlab, FILE *fpout)
{
  SyVertex   vhead, vtail;

  static 
  int      line_width[5] = {1,8,4,2,1};
  int      lw_dim  =  4;

  float    x1, y1, x2, y2;
  int      conn_flag, cur_rank;
  int      conn[4];
  int      tail_vertex, head_vertex, 
           left_point, right_point,
           tail_ipo, head_ipo;
  int      tail_flag, head_flag;
  int      is_residual_cut, is_skel, is_foreground, is_render;

  float    scaleX, scaleY, scaleXY;
  float    xOff,   yOff, 
           xShift, yShift;

  int      maxX, minX, maxY, minY;
  int      dimX, dimY; 
  int      effDimX, effDimY;

  int      i, nhit = 0;

  maxX = minX = vsk->point[1].x;
  minY = vsk->point[1].y;
  maxY = vsk->point[vsk->pdim].y;

  for (i = 1; i <= vsk->pdim; i++){
     maxX = max(maxX, vsk->point[i].x);
     minX = min(minX, vsk->point[i].x);
  }

  dimX = maxX - minX;
  dimY = maxY - minY;

  effDimX = max(min(dimX, SY_IDRAW_DEF_XBBOX), SY_IDRAW_MIN_XBBOX);
  effDimY = max(min(dimY, SY_IDRAW_DEF_YBBOX), SY_IDRAW_MIN_YBBOX);
  scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
  scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
  scaleXY = min(scaleX, scaleY);
  xOff    = dimX/2 + minX;
  yOff    = dimY/2 + minY;

  xShift  = effDimX/2;
  yShift  = effDimY/2;

  Sy_IdrawSetCreator("Function Sy_Pyr2Idraw");
  Sy_IdrawSetBoundingBox(0,0,effDimX,effDimY);
  Sy_IdrawPrologue(fpout);
  
  Sy_Ipo2Idraw (ipd, fpout, SY_TRUE);

  Sy_IdrawSetXYScale(scaleXY, scaleXY);
  Sy_IdrawSetXYTranslate(xShift-scaleXY*xOff, yShift-scaleXY*yOff);
  Sy_IdrawSetCurrentFGColor("Red", 0,0,1);

  Sy_SetSimpleViewport(effDimX, effDimY);
  Sy_SetVectorProc(Sy_IdrawLine);

  cur_rank = 1;
  while (cur_rank <= rank){ 
    Sy_IdrawBeginGroup();
    Sy_IdrawSetLineWidth (cur_rank <= 4 ? line_width[cur_rank] : 1);
    nhit = 0;

    for (i = 1; i <= vsk->edim; i++) {
    
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      
      is_foreground  = Sy_SkeletonType (vsk, ipd, tail_vertex, i);
      
      if ((is_foreground && (display_mode == SKE_BACKGROUND)) ||
	  (!is_foreground && (display_mode == SKE_FOREGROUND))) continue;
      
      conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		  (ipd->ipl[left_point].tail_icn != 0));
      
      conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		  (ipd->ipl[left_point].tail_icn != 0));

      conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		  (ipd->ipl[left_point].head_icn != 0));
      
      conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		  (ipd->ipl[left_point].head_icn != 0));
      
      conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);
      
      is_render = (pyrlab[i] > 0) && (cur_rank == pyrlab[i]) && 
	(!conn_flag || vsk->is_border);

      if (is_render){

	nhit++;
	vtail = vsk->vertex[vsk->edge[i].tail_vertex];
	vhead = vsk->vertex[vsk->edge[i].head_vertex];

	x1 = vtail.xden / vtail.num + SY_XCORR;
	y1 = vtail.yden / vtail.num + SY_YCORR;
	x2 = vhead.xden / vhead.num + SY_XCORR;
	y2 = vhead.yden / vhead.num + SY_YCORR;
	
	switch (vsk->edge[i].rank) {
	case 0:
	  Sy_DrawClippedLine(x1, y1, vsk->edge[i].dirx, 
			     vsk->edge[i].diry); break;
	case 1: 
	  Sy_DrawClippedRay(x1, y1, vsk->edge[i].dirx, 
			    vsk->edge[i].diry); break;
      default:
/*	if ((x1 < 0) || (x1 > ncols-1) || (x2 < 0) || (x2 > ncols-1) ||
	(y1 < 0) || (y1 > nrows-1) || (y2 < 0) || (y2 > nrows-1)) break;
	digline (x1, y1, x2, y2, xvdot); */      
	Sy_DrawClippedVector (x1, y1, x2, y2);
	break;
	}
      }
    }
    Sy_IdrawEndGroup(); 
    if (nhit == 0) break;
    cur_rank++;
  } /* while */
  Sy_IdrawEpilogue();
}


void Sy_SkeletonSet2Idraw(SyVSkeletonSet *vss, float uscale, 
			  SyDisplayControl *ctrl, FILE *fpout)
{
  int	    i, k, v;
  float     x1, y1, x2, y2, x01, y01, x02, y02, mx, my, tx, ty;
  float     scaleX, scaleY, scaleXY;
  float     xOff, yOff, 
            xShift, yShift;
  float     dirX, dirY;
  int       maxX, maxY, minX, minY;
  int       dimX, dimY, effDimX, effDimY;
  int       max_branch_elementdim = 0;
  float     min_strength, max_strength, strength, width_factor;

  int       xZen = 0, yZen = 0;

  SyBoolean is_foreground;

  SyIdrawPoint *idraw_point = NULL;

  union SyGenericPoint *axis;

  min_strength = SY_POS_INFTY;
  max_strength = SY_NEG_INFTY;

  maxX = maxY = SY_NEG_INT_INFTY; 
  minX = minY = SY_POS_INT_INFTY;

  for (i = 1; i <= vss->dim; i++){

    if ((vss->skel[i].is_foreground && ctrl->display_mode == SKE_BACKGROUND) ||
        (!vss->skel[i].is_foreground && ctrl->display_mode == SKE_FOREGROUND)) 
      continue;

    for (k = 1; k <= vss->skel[i].branchdim; k++){

      axis = vss->skel[i].branch[k].axis;
      if (max_branch_elementdim < vss->skel[i].branch[k].elementdim){
	max_branch_elementdim = vss->skel[i].branch[k].elementdim;
      }

      if (vss->skel[i].branch[k].hierarchy_level != 1){
	strength = vss->skel[i].branch[k].strength;
	if (strength > max_strength) max_strength = strength;
	if (strength < min_strength) min_strength = strength;
      }

      for (v = 1; v <= vss->skel[i].branch[k].elementdim; v++){       
	if (maxX < axis[v].rp.x){ maxX = axis[v].rp.x; }
	if (maxY < axis[v].rp.y){ maxY = axis[v].rp.y; }
	if (minX > axis[v].rp.x){ minX = axis[v].rp.x; }
	if (minY > axis[v].rp.y){ minY = axis[v].rp.y; }
      }

      for (v = 2; v <= vss->skel[i].branch[k].elementdim; v++){  
	mx = 0.5*(axis[v].rp.x + axis[v-1].rp.x);
	my = 0.5*(axis[v].rp.y + axis[v-1].rp.y);
	x1 = mx + vss->skel[i].branch[k].left_bndry[v-1].xoff;
	y1 = my + vss->skel[i].branch[k].left_bndry[v-1].yoff;
	x2 = mx + vss->skel[i].branch[k].right_bndry[v-1].xoff;
	y2 = my + vss->skel[i].branch[k].right_bndry[v-1].yoff;

	if (maxX < x1) maxX = x1;
	if (maxX < x2) maxX = x2;
	if (maxY < y1) maxY = y1;
	if (maxY < y2) maxY = y2;
 
	if (minX > x1) minX = x1;
	if (minX > x2) minX = x2;
	if (minY > y1) minY = y1;
	if (minY > y2) minY = y2;
      }
    }
  }

  dimX = maxX-minX;
  dimY = maxY-minY;

  if (uscale > 0){
    effDimX = max(min(dimX * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
    effDimY = max(min(dimY * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
    xOff    = (xZen > 0) ? xZen : dimX/2 + minX;
    yOff    = (yZen > 0) ? yZen : dimY/2 + minY;
    scaleXY = uscale;
  }
  else{
    effDimX = max(min(dimX, SY_DEF_WINDIM), SY_MIN_WINDIM);
    effDimY = max(min(dimY, SY_DEF_WINDIM), SY_MIN_WINDIM);
    scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
    scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
    scaleXY = min(scaleX, scaleY);
    xOff    = dimX/2 + minX;
    yOff    = dimY/2 + minY;
  }

  xShift  = effDimX/2 + SY_MIN_BORDER;
  yShift  = effDimY/2 + SY_MIN_BORDER;

  idraw_point = Sy_AllocIdrawPoint(max_branch_elementdim+1);

  Sy_IdrawSetCreator("Function Sy_SkeletonSet2Idraw");
  Sy_IdrawSetBoundingBox(0,0,effDimX,effDimY);
  Sy_IdrawPrologue(fpout);
  
  Sy_IdrawSetXYScale(1.0, 1.0);
  Sy_IdrawSetXYTranslate(0,0);

  Sy_IdrawBeginGroup();

  for (i = 1; i <= vss->dim; i++){

    if ((vss->skel[i].is_foreground && ctrl->display_mode == SKE_BACKGROUND) ||
        (!vss->skel[i].is_foreground && ctrl->display_mode == SKE_FOREGROUND)) 
      continue;

    Sy_IdrawSetCurrentFGColor("Red", 1,0,0);
    Sy_IdrawBeginGroup();
    
    /* draw nodes */
    for (k = 1; k <= vss->skel[i].nodedim; k++){

      if (ctrl->rank > 0){
	if (vss->skel[i].node[k].hierarchy_level <= 0 ||
	    vss->skel[i].node[k].hierarchy_level > ctrl->rank) continue;
      }

      x1 = vss->skel[i].node[k].loc.rp.x;
      y1 = vss->skel[i].node[k].loc.rp.y;
      x1 = scaleXY * (x1- xOff) + xShift;
      y1 = scaleXY * (y1- yOff) + yShift;

      Sy_IdrawEllipse(x1, y1, 3, 3);
    }
    Sy_IdrawEndGroup();
    
    Sy_IdrawSetCurrentFGColor("Blue", 0,0,1);
    Sy_IdrawBeginGroup();

    /* draw axes */
    width_factor = 15.0/(max_strength-min_strength);
    fprintf (stderr, "strength dynamic: %f\n", max_strength-min_strength);

    for (k = 1; k <= vss->skel[i].branchdim; k++){

      if (ctrl->rank > 0){
	if (vss->skel[i].branch[k].hierarchy_level <= 0 ||
	    vss->skel[i].branch[k].hierarchy_level > ctrl->rank) continue;
      }

      if (vss->skel[i].branch[k].hierarchy_level == 1){
	Sy_IdrawSetCurrentFGColor("Blue", 0,0,1);
	Sy_IdrawSetLineWidth (20);
      }
      else{
	Sy_IdrawSetCurrentFGColor("Green", 0,1,0);
	strength = vss->skel[i].branch[k].strength;
	Sy_IdrawSetLineWidth (floor((strength-min_strength)*width_factor+1.5));
      }

      axis = vss->skel[i].branch[k].axis;
      for (v = 1; v <= vss->skel[i].branch[k].elementdim; v++){
	x1 = axis[v].rp.x;
	y1 = axis[v].rp.y;
	x1 = scaleXY * (x1 - xOff) + xShift;
	y1 = scaleXY * (y1 - yOff) + yShift; 
	idraw_point[v-1].x = floor(x1+0.5);
	idraw_point[v-1].y = floor(y1+0.5);
      }

      Sy_IdrawMLine(idraw_point, vss->skel[i].branch[k].elementdim);
      
    }
    Sy_IdrawEndGroup();
  } /* for all skeletons */

  Sy_IdrawEndGroup();

  Sy_IdrawEpilogue();
  if (idraw_point) Sy_Free(idraw_point);
}



