/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * crack extraction routine
 *
 */ 

static char sy_cratranId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

#define SY_INIT_NR_IPO 10000
#define SY_ADD_NR_IPO  10000
 
#define SY_EXCHANGE_PTR(a,b) {int *h; h = a; a = b; b = h;}

void Sy_SetIpo(SyIpoDAC *crack, int nr, int x, int y, 
	       int tail_icn, int head_icn, SyFloatVector grad);
 
void Sy_ExpandLine1Bit(unsigned char *line, SyCratranBuff *buff, 
		       SyCratranUtil *util, SyImage_t *ima);

void Sy_ExpandLine8Bit(unsigned char *line, SyCratranBuff *buff, 
		       SyCratranUtil *util, SyImage_t *ima);

void Sy_ExpandLine24Bit(unsigned char (*line)[3], SyCratranBuff *buff, 
			SyCratranUtil *util, SyImage_t *ima);

int  Sy_Con8To4(int *lineA, int *lineB, int width); 
void Sy_DisplayLine(int *line, int width, char *text);

void Sy_FilterLine8Bit (unsigned char *line, SyCratranUtil *util, 
			SyImage_t *ima);
void Sy_FilterLine24Bit (unsigned char (*line)[3], SyCratranUtil *util, 
			 SyImage_t *ima);

void Sy_Cratran (SyImage_t *ima, SyIpoDAC *crack, SyCratranControl *cntl)
{
  SyCratranStack stack;
  SyCratranTable table;
  SyCratranUtil  util;
  SyCratranBuff  buff;
  SyCratranInfo  info;

  SyFloatVector   grad;

  SY_UCHAR*      imline;

  int           arc_ptr, arc_ptr1, arc_ptr2,
                arc_nr,  arc_nr1,  arc_nr2;
     
  int           val_xy, val_x1y, val_xy1, val_x1y1,
                arc_count = 0, table_count = 0,
                ipo_count = 0;

  int           step_down, step_right;

  int           x, y, yy, i, kk;

  int           oper, last_line_blank, this_line_blank = 1,
                con8_count = 0, memo = 0, single_step, xincr;

  int           ipo_mem, ima_blank_val;

  ima->filter_line = (void (*)())NULL;

  if (ima->pixlen < 0){
    switch (ima->depth){
    case  1: ima->expand_line = Sy_ExpandLine1Bit;  
#ifndef HAS_GEN_LOADER
             cntl->invert_image = cntl->invert_image ? SY_FALSE : SY_TRUE;
#endif
             ima_blank_val = cntl->invert_image ? 255 : 0; break;
    case  8: ima->expand_line = Sy_ExpandLine8Bit;  
             if (ima->is_pixval) ima->filter_line = Sy_FilterLine8Bit;
             ima_blank_val = cntl->invert_image ? 255 : 0; break;
    case 24: ima->expand_line = Sy_ExpandLine24Bit; 
             if (ima->is_pixval) ima->filter_line = Sy_FilterLine24Bit;
             ima_blank_val = cntl->invert_image ? 255 : 0; break;
    default:  
      sprintf (syStrBuf, 
	       "Crack Translator: unsupported image depth (%d)",
	       ima->depth);
      Sy_SimpleError (SY_FMTNL, syStrBuf); 
      exit(2);
    }/* switch */  
  } /* if */
  else {
    switch (ima->pixlen){
    case  1: ima->expand_line = Sy_ExpandLine8Bit;
             if (ima->is_pixval) ima->filter_line = Sy_FilterLine8Bit;
             ima_blank_val = cntl->invert_image ? 255 : 0; break;
    case  3: ima->expand_line = Sy_ExpandLine24Bit; 
             if (ima->is_pixval) ima->filter_line = Sy_FilterLine24Bit;
             ima_blank_val = cntl->invert_image ? 255 : 0; break;
    default:  
      sprintf (syStrBuf, 
	       "Crack Translator: unsupported image pixlen (%d)",
	       ima->pixlen);
      Sy_SimpleError (SY_FMTNL, syStrBuf); 
      exit(2);
    }/* switch */  
  } /* else */

  sprintf (syStrBuf, "Crack Translator %s ", sy_cratranId);
  Sy_Message (SY_FMTNL, syStrBuf);
  
  /* memory allocation */

  Sy_AllocCratranEnvironment (ima, &buff, &table, &stack, 
			      &util, &info, cntl);
  crack->ipo   = Sy_AllocIpo (SY_INIT_NR_IPO);
  crack->ipl   = Sy_AllocIpl (SY_INIT_NR_IPO);
  info.nr_ipos = SY_INIT_NR_IPO;

  switch (ima->depth){
  case  1: /* creating masks */
           for (i = 0; i < ima->width; i++){
             util.offset[i] = i   >> 3;
	     util.mask[i]   = cntl->reverse_bits ? 128 >> (7-i%8) : 
	                                           128 >> (i%8);
	   }
           break;
  case  8: ;
  case 24: ;
  default: ;
  }

  /* crack to vector conversion */

  (void) memset((SY_UCHAR*)(ima->data + ima->linebytes * 
                (ima->height-1)), ima_blank_val, ima->linebytes);

  for (y = 0, imline = (SY_UCHAR*)ima->data + ima->linebytes; 
       imline < (SY_UCHAR*) ima->data + ima->linebytes * ima->height;
       imline += ima->linebytes, y++){

    if (syIsVerbose){ 
      sprintf (syStrBuf, "Y-coord. : %d  SyIpos: %d  8-conn.: %d\r", 
	       y, ipo_count, con8_count);
      Sy_Message (SY_FMT, syStrBuf);
    }

    if (ima->filter_line) ima->filter_line (imline, &util, ima);

    if (ipo_count + ima->width >= info.nr_ipos) {
      info.nr_ipos += SY_ADD_NR_IPO;
      crack->ipo = Sy_ReallocIpo (crack->ipo, info.nr_ipos);
      crack->ipl = Sy_ReallocIpl (crack->ipl, info.nr_ipos);
    }
    SY_EXCHANGE_PTR (buff.line[0], buff.line[2]);
    SY_EXCHANGE_PTR (buff.line[1], buff.line[3]);

    last_line_blank = this_line_blank;    
    this_line_blank = 
      memcmp ((SY_UCHAR*)util.blank_line, imline, ima->linebytes) == 0;
    
    if (!this_line_blank){
      if (last_line_blank){
        (void) memset ((SY_UCHAR*)buff.line[0], 0, 2*ima->width * sizeof(int));
	(void) memset ((SY_UCHAR*)buff.line[1], 0, 2*ima->width * sizeof(int));
      }
    }

    if (!(last_line_blank && this_line_blank)){
      if (this_line_blank){
        (void) memset ((SY_UCHAR*)buff.line[2], 0, 2*ima->width * sizeof(int));
	(void) memset ((SY_UCHAR*)buff.line[3], 0, 2*ima->width * sizeof(int));
      }
      else{
	ima->expand_line (imline, &buff, &util, ima); 
	con8_count += single_step = 
	  Sy_Con8To4 (buff.line[1], buff.line[2], 2*buff.width);
	if (single_step) memo = 4;
      }
     
#ifdef DEBUG        
      Sy_DisplayLine(buff.line[2], buff.width, "line 2");
#endif
      for (yy = memo?2*y:2*y+1, kk = memo?0:1; kk <= 1; kk++, yy++){
        xincr = memo?1:2;
	val_xy = val_xy1 = 0;
	for (x = memo?0:1; x < 2*ima->width-1; x+=xincr) {

	  val_x1y  = buff.line[kk][x+1];
	  val_x1y1 = buff.line[kk+1][x+1];

	  if ((val_xy != val_x1y)||(val_xy != val_x1y1)||(val_xy != val_xy1)){

	    step_down  = val_xy ^ val_x1y;
	    step_right = val_xy ^ val_xy1;
	
	    if      (!(step_down || step_right)) oper = 1;
	    else if (step_right && step_down)    oper = 2;
	    else if (step_down)                  oper = 3;
	    else if (step_right)                 oper = 4;
	  
	    switch (oper) {
	    case 1: /* new segment */
	      arc_ptr1 = 
              stack.arc_ptr ? stack.arc_[--stack.arc_ptr] : ++table_count; 
	      arc_nr1  = 
              stack.ind_ptr ? stack.index_[--stack.ind_ptr] : ++arc_count;
	      arc_ptr2 = 
	      stack.arc_ptr ? stack.arc_[--stack.arc_ptr] : ++table_count; 
	      
	      table.x_[x]                   = arc_ptr1;
	      stack.x_[stack.x_ptr].x       = x;
	      stack.x_[stack.x_ptr].arc_ptr = arc_ptr2;
	      stack.x_ptr++;
	      
              if ((x%2) & (yy%2)){
		arc_nr2  = 
		stack.ind_ptr ? stack.index_[--stack.ind_ptr] : ++arc_count;

		grad.x = grad.y = 2*(buff.line[kk+1][x+1] != 0) - 1;
		Sy_SetIpo (crack, ++ipo_count, x, yy, arc_nr2, arc_nr1, grad);
	   
		table.arc_[arc_ptr1].arc_id = arc_nr1;
		table.arc_[arc_ptr1].ipo_id = ipo_count;
		table.arc_[arc_ptr2].arc_id = arc_nr2;
		table.arc_[arc_ptr2].ipo_id = ipo_count;
	      }
	      else {
		table.arc_[arc_ptr1].arc_id = arc_nr1;
		table.arc_[arc_ptr1].ipo_id = 0;
		table.arc_[arc_ptr2].arc_id = arc_nr1;
		table.arc_[arc_ptr2].ipo_id = 0;
	      }
	      break;
	      
	    case 4: /* step right */
	      arc_ptr = table.x_[x-xincr];
	      if (arc_ptr == 0){
		Sy_SimpleError (SY_FMTNL, 
				"Crack Translator error: chain lost");
	      }
	      if ((x%2) & (yy%2)){
		grad.x = 0; 
	        grad.y = 2*(buff.line[kk+1][x] != 0) - 1;
		arc_nr = 
		stack.ind_ptr ? stack.index_[--stack.ind_ptr] : ++arc_count;
		Sy_SetIpo (crack, ++ipo_count, x, yy, 
			 table.arc_[arc_ptr].arc_id, arc_nr, grad);
		table.arc_[arc_ptr].arc_id = arc_nr;
		table.arc_[arc_ptr].ipo_id = ipo_count;
	      }
	      table.x_[x-xincr] = 0;
	      table.x_[x]   = arc_ptr;
	      break;
	      
	    case 3: /* step Sy_DcelDown */
	      arc_ptr = table.x_[x];
	      if (arc_ptr == 0){
		Sy_SimpleError (SY_FMTNL, 
				"Crack Translator error: chain lost");
	      }
	      if ((x%2) & (yy%2)){
		grad.x = 2*(buff.line[kk][x+1] != 0) - 1;
		grad.y = 0.0;
		arc_nr  = 
		stack.ind_ptr ? stack.index_[--stack.ind_ptr] : ++arc_count;
		
		Sy_SetIpo (crack, ++ipo_count, x, yy,
			  table.arc_[arc_ptr].arc_id, arc_nr, grad);
		table.arc_[arc_ptr].arc_id = arc_nr;
		table.arc_[arc_ptr].ipo_id = ipo_count;
	      }
	      break;
	      
	    case 2: /* step right & Sy_DcelDown */
	      arc_ptr1 = table.x_[x-xincr];
	      arc_ptr2 = table.x_[x];

	      if (arc_ptr1 == 0){
		Sy_SimpleError (SY_FMTNL, 
				"Crack Translator error: chain lost");
	      }
	      if (arc_ptr2 == 0){
		Sy_SimpleError (SY_FMTNL, 
				"Crack Translator error: chain lost");
	      }
	      if ((x%2) & (yy%2)){
		grad.x = 2*(buff.line[kk][x+1] != 0) - 1;
		grad.y = 2*(buff.line[kk+1][x] != 0) - 1;
		Sy_SetIpo (crack, ++ipo_count, x, yy, 
			 table.arc_[arc_ptr1].arc_id,
			 table.arc_[arc_ptr2].arc_id, grad);
              }
	      else{

		if (table.arc_[arc_ptr1].ipo_id){

		  if (crack->ipl[table.arc_[arc_ptr1].ipo_id].head_icn == 
		      table.arc_[arc_ptr1].arc_id){ 
		    crack->ipl[table.arc_[arc_ptr1].ipo_id].head_icn =
		      table.arc_[arc_ptr2].arc_id;
		  }
		  else{
		    crack->ipl[table.arc_[arc_ptr1].ipo_id].tail_icn =
		      table.arc_[arc_ptr2].arc_id;
		  }
		}
		else{
		  arc_ptr = stack.x_[stack.x_ptr-1].arc_ptr;
		  table.arc_[arc_ptr].arc_id = table.arc_[arc_ptr2].arc_id;
		}
		  stack.index_[stack.ind_ptr++]  = 
		  table.arc_[arc_ptr1].arc_id;

	      }
              stack.arc_[stack.arc_ptr++] = arc_ptr1;
	      stack.arc_[stack.arc_ptr++] = arc_ptr2;
	      table.x_[x-xincr] = table.x_[x] = 0;
	      break;
	    }
	  }
	  val_xy  = val_x1y; 
	  val_xy1 = val_x1y1;	 
	}
	while (stack.x_ptr){
	  stack.x_ptr--; 
	  table.x_[stack.x_[stack.x_ptr].x] = 
	    stack.x_[stack.x_ptr].arc_ptr;
	}      
	if (memo) memo--;
      } /* for (yy ...) */
    } /* if (!(last_line_blank ...) */
  }
  crack->ipodim = ipo_count;

  Sy_FreeCratranEnvironment (ima, &buff, &table, &stack, 
			     &util, &info, cntl);
}

void Sy_ExpandLine1Bit (unsigned char *line, SyCratranBuff *buff, 
			SyCratranUtil *util, SyImage_t *ima)
{
  int x, i, width;
  SY_UCHAR object, zero;

  object = util->invert_image ? 0 : 255;
  zero   = util->invert_image ? 255 : 0;

  width = ima->width;
  
  (void) memset ((char *)buff->line[2], 0, (int)(2*width * sizeof(int)));
  (void) memset ((char *)buff->line[3], 0, (int)(2*width * sizeof(int)));
  
  for (x = 0; x < width; x += 8){
    if (line[util->offset[x]] == zero);
    else{ 
      if (line[util->offset[x]] == object){
 	(void) memcpy((char*)&(buff->line[2][2*x]),(char*)util->ones16, 64);
	(void) memcpy((char*)&(buff->line[3][2*x]),(char*)util->ones16, 64); 
      }
      else{
	for (i = x; (i < x+8) && (i < width); i++){
	  buff->line[2][2*i] = buff->line[2][2*i+1] =
	  buff->line[3][2*i] = buff->line[3][2*i+1] = 
	    util->invert_image ? 
	    ((~line[util->offset[i]] & util->mask[i]) != 0): 
	    ((line[util->offset[i]] & util->mask[i]) != 0);
	}
      }
    }
  }
  buff->line[2][0]         = buff->line[2][1]         = 
  buff->line[2][2*width-2] = buff->line[2][2*width-1] =
  buff->line[3][0]         = buff->line[3][1]         = 
  buff->line[3][2*width-2] = buff->line[3][2*width-1] = 0;
  return;
}

void Sy_ExpandLine8Bit (unsigned char *line, SyCratranBuff *buff, 
			SyCratranUtil *util, SyImage_t *ima)
{
  int x, width, pixel;

  width = ima->width;

  (void) memset ((char *)buff->line[2], 0, (int)(2*width * sizeof(int)));
  (void) memset ((char *)buff->line[3], 0, (int)(2*width * sizeof(int)));

  for (x = 1; x < width-1; x++){
    pixel = util->invert_image ? line[x] < 128 : line[x] >= 128; 
    if (pixel){
      buff->line[2][2*x] = buff->line[2][2*x+1] =
      buff->line[3][2*x] = buff->line[3][2*x+1] = 1;
    }
  }

  buff->line[2][0]         = buff->line[2][1]         = 
  buff->line[2][2*width-2] = buff->line[2][2*width-1] =
  buff->line[3][0]         = buff->line[3][1]         = 
  buff->line[3][2*width-2] = buff->line[3][2*width-1] = 0;
  return;
}

void Sy_ExpandLine24Bit (unsigned char (*line)[3], SyCratranBuff *buff, 
			 SyCratranUtil *util, SyImage_t *ima)
{
  int x, i, width;
  width = ima->width;
  
  (void) memset ((SY_UCHAR *)buff->line[2], 0, (int)(2*width * sizeof(int)));
  (void) memset ((SY_UCHAR *)buff->line[3], 0, (int)(2*width * sizeof(int)));
  
  for (x = 1; x < width-1; x++){
    buff->line[2][2*x] = buff->line[2][2*x+1] =
    buff->line[3][2*x] = buff->line[3][2*x+1] = util->invert_image ?
    ((line[x][0] | line[x][1] | line[x][2]) == 0) :
    ((line[x][0] | line[x][1] | line[x][2]) != 0);
  }
  return;
}

int Sy_Con8To4 (int *lineA, int *lineB, int width)
{
  int x, con8_count = 0;

  for (x = 1; x < width-1; x+=2){
    if (lineA[x] != lineA[x+1]){
      if (lineA[x]  & lineB[x+1]  & !lineA[x+1] & !lineB[x]){
	lineA[x+1] = lineB[x] = 1;
	lineA[x+2] = lineA[x+3] & lineB[x+3];
	lineB[x-1] = lineA[x-2] & lineB[x-2];
	con8_count++;
      }
      else{
	if (!lineA[x] & !lineB[x+1] & lineA[x+1]  & lineB[x]){
	  lineA[x] = lineB[x+1] = 1;
	  lineB[x+2] = lineA[x+3] & lineB[x+3];
	  lineA[x-1] = lineA[x-2] & lineB[x-2];
	  con8_count++;
	}
      }
    }
  }
  return(con8_count);
}

void Sy_SetIpo (SyIpoDAC *crack, int nr, int x, int y, int tail_icn, 
		int head_icn, SyFloatVector grad)
{
    crack->ipo[nr].x         = x/2;
    crack->ipo[nr].y         = y/2;
    crack->ipl[nr].tail_icn  = tail_icn;
    crack->ipl[nr].head_icn  = head_icn;
    crack->ipl[nr].grad      = grad;
}

void Sy_CratranBuildIcn (SyIpoDAC *crack)
{
  SyIpl* ipl = crack->ipl;
  SyIcn* icn;
  int  tail_icn, head_icn, i;

  crack->icn    = Sy_AllocIcn (crack->ipodim);
  icn           = crack->icn;
  crack->icndim = crack->ipodim;

  for (i = 1; i <= crack->ipodim; i++){

    tail_icn = ipl[i].tail_icn;
    head_icn = ipl[i].head_icn;

    if (icn[tail_icn].tail_ipo) icn[tail_icn].head_ipo = i;
    else                        icn[tail_icn].tail_ipo = i; 

    if (icn[head_icn].tail_ipo) icn[head_icn].head_ipo = i;
    else                        icn[head_icn].tail_ipo = i; 
  }
}

int Sy_CratranIpoCheck (SyIpoDAC *crack)
{
  SyIpl* ipl = crack->ipl;
  int* hit;
  int  tail_icn, head_icn, i, err_count = 0;

  (void) Sy_AllocCratranHittab (&hit, crack->icndim);

  for (i = 1; i <= crack->ipodim; i++){

    tail_icn = ipl[i].tail_icn;
    head_icn = ipl[i].head_icn;

    if ((tail_icn > crack->icndim) || (head_icn > crack->icndim)){
      if (++err_count == 1){ 
	sprintf(syStrBuf, " %%%%CRATRAN ERROR LISTING ");
	Sy_Message (SY_FMTNL, syStrBuf);
      }
      Sy_Message(SY_FMTNL, "overflow  error: ");
      sprintf(syStrBuf, "SyIcns: %6d %6d dim: %6d ", 
	      tail_icn, head_icn, crack->icndim);
      Sy_Message (SY_FMTNL, syStrBuf);
    }
    else{
      hit[tail_icn]++;
      hit[head_icn]++;
    }
  }
  if (hit[0] != 0){
    if (++err_count == 1){ 
      sprintf(syStrBuf, " %%%%CRATRAN ERROR LISTING ");
      Sy_Message (SY_FMTNL, syStrBuf);
    }
    sprintf (syStrBuf, "reference error: missing references: %3d ", hit[0]); 
	     Sy_Message (SY_FMTNL, syStrBuf);
  }
  for (i = 1; i <= crack->icndim; i++){
    if (hit[i] != 2){
      if (++err_count == 1){ 
	sprintf(syStrBuf, " %%%%CRATRAN ERROR LISTING ");
	Sy_Message (SY_FMTNL, syStrBuf);
      }
      sprintf (syStrBuf, "reference error: SyIcn: %6d ref.: %3d ", 
	       i, hit[i]);
      Sy_Message (SY_FMTNL, syStrBuf);

    }
  }

  Sy_FreeCratranHittab (&hit);
  return(err_count);
}


void Sy_DisplayLine (int *line, int width, char *text)
{
  int i;

  fprintf (stderr, "displaying %s:  ", text);
  for (i = 0; i < width; i++){
    fprintf (stderr, "%d", line[2*i]); 
  }
  fprintf (stderr, "\n");
}

void Sy_FilterLine8Bit (unsigned char *line, SyCratranUtil *util, 
			SyImage_t *ima)
{
  int x, width;
  width = ima->width;

  for (x = 0; x < width; x++){
    if (line[x] != ima->greyval) line[x] = 0; else line[x] = 255;
  }
  return;
}

void Sy_FilterLine24Bit (unsigned char (*line)[3], SyCratranUtil *util, 
			 SyImage_t *ima)
{
  int x, width;
  width = ima->width;
  
  for (x = 0; x < width; x++){
    if (line[x][0] != ima->r_pix ||
	line[x][1] != ima->g_pix ||
	line[x][2] != ima->b_pix  ) line[x][0] = line[x][1] = line[x][2] = 0;
    else line[x][0] = line[x][1] = line[x][2] = 255;
  }
  return;
}

void Sy_CheckPixvalInImage (SyImage_t *ima)
{
  unsigned char *imline;
  int x, y, i;
  int *acc;

  acc = (int*) malloc (sizeof(int)*256);
  memset (acc, 0, sizeof(int)*256);
  
  for (y = 0, imline = (SY_UCHAR*)ima->data + ima->linebytes; 
       imline < (SY_UCHAR*) ima->data + ima->linebytes * ima->height;
       imline += ima->linebytes, y++){

    for (x = 0; x < ima->width; x++){
      if (imline[x] < 256) acc[imline[x]]++;
    }
  }
  
  for (i = 0; i < 256; i++){
    if (acc[i] > 0){
      fprintf (stderr, "%7d pixel(s) with value %d\n", acc[i], i);
    }
  }
  free (acc);
}
