/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char sy_DrawPromId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Graphics.h"
#include "inipoints.xbm"

#ifndef NOX
#include <X11/Xatom.h>
static Atom proto_atom = None, delete_atom = None;

typedef struct{
   Window  winId;
   char    *tag;
   void   (*eventHandler)();
} SyLtEventAction;

static XEvent  LtXevent;

static GC      gc;
static Pixmap  icon;

static XArc     *arcPtr;
static XPoint   *pointPtr;
static XSegment *segmentPtr;
static int      pointDim, segmentDim;

static Display *display;
static char    *displayName = NULL;
static int     screenNr;

static void Sy_CatchEvent(XEvent *event, Window winId, unsigned int winNr, char *tag);
static void Sy_ExposeEvent(XExposeEvent *event, Window winId, unsigned int winNr);
static void Sy_Hipo2X(SyHipoContainer *container, XPoint *pointPtr, int *pointDim, XSegment *segmentPtr, int *segmentDim, int *winWidth, int *winHeight, int xZen, int yZen, float uscale, SyIpoMappingControl *mcontrol);

static int  halt = 0;
#endif

int Sy_XDisplayHipoContainer (SyHipoContainer *container, char *windowName, char *iconName, int xPos, int yPos, int xZen, int yZen, float uscale, char *userDisplay, SyIpoMappingControl *mcontrol)
{
#ifndef NOX
  SyLtEventAction  eventBuff;
  int            winWidth, winHeight;
  int            borderWidth = SY_DEF_BORDERWIDTH;
  int            point_mem, segment_mem, page;
  
  page         = getpagesize();
  point_mem    = (((container->total_iposize+1)*sizeof(XPoint))/page+1)*page;
  segment_mem  = (((container->total_iposize+3)*sizeof(XSegment))/page+1)*page;

  pointPtr   = (XPoint*)   Sy_Malloc ((unsigned)point_mem);
  arcPtr     = (XArc*)     Sy_Malloc ((unsigned)point_mem);
  segmentPtr = (XSegment*) Sy_Malloc ((unsigned)segment_mem);

  Sy_Hipo2X (container, pointPtr, &pointDim, segmentPtr, &segmentDim, 
            &winWidth, &winHeight, xZen, yZen, uscale, mcontrol); 

  displayName  = userDisplay != NULL ? userDisplay : NULL;
  if ((display = XOpenDisplay (displayName)) == NULL) return (-1);  
  screenNr     = XDefaultScreen (display);

  eventBuff.winId = XCreateSimpleWindow (display, 
					 RootWindow (display, screenNr), 
					 xPos, yPos, winWidth, 
					 winHeight, borderWidth,
					 BlackPixel (display, screenNr),
					 WhitePixel (display, screenNr));
  
  gc   = XCreateGC (display, eventBuff.winId, 0, 0);
  icon = XCreateBitmapFromData (display, eventBuff.winId, ipo_icon_bits, 
                                ipo_icon_width, ipo_icon_height);
  XSetForeground (display, gc, BlackPixel (display, screenNr));
  XSetBackground (display, gc, WhitePixel (display, screenNr));
  XSetStandardProperties (display, eventBuff.winId, windowName, 
			  iconName , icon, NULL, 0, NULL);

  XSelectInput (display, eventBuff.winId, 
		ExposureMask | StructureNotifyMask);

  proto_atom = XInternAtom(display, "WM_PROTOCOLS", False);
  delete_atom = XInternAtom(display, "WM_DELETE_WINDOW", False);
  if ((proto_atom != None) && (delete_atom != None))
    XChangeProperty(display, eventBuff.winId, proto_atom, XA_ATOM, 32,
		    PropModeReplace, (unsigned char*) &delete_atom, 1);

  eventBuff.eventHandler = Sy_CatchEvent;
   
  XMapWindow (display, eventBuff.winId);
  while (LtXevent.type != Expose && LtXevent.xany.window != eventBuff.winId) 
    XNextEvent(display, &LtXevent);

  while (1){
    if (LtXevent.xany.window == eventBuff.winId)
      eventBuff.eventHandler (&LtXevent, LtXevent.xany.window, 1, NULL);
    XNextEvent (display, &LtXevent);
  }
#else
  Sy_Message("%s\n", "no X11 graphics support available\n");
#endif
}

#ifndef NOX
static void Sy_CatchEvent (XEvent *event, Window winId, 
			   unsigned int winNr, char *tag)
{
  XClientMessageEvent *me;
  switch (event->type){
  case Expose :
    Sy_ExposeEvent ((XExposeEvent*)event, winId, winNr); 
    break;
  case DestroyNotify:
    exit(2); break;
  case ClientMessage:
    me = (XClientMessageEvent*)event;
    if ((me->window == winId) && (me->data.l[0] == delete_atom)){
      exit(2);
    }
    break;
  default : break;
  }
  return;
}

static void Sy_ExposeEvent (XExposeEvent *event, Window winId, 
			    unsigned int winNr)
{
   XArc     *cur_arc_block;
   XPoint   *cur_point_block;
   XSegment *cur_segment_block;
   int      count;
/*
   for (cur_point_block = pointPtr, count = 0; count < pointDim/SY_XBLOCK;
        cur_point_block += SY_XBLOCK, count++){
     XDrawPoints  (display, winId, gc, cur_point_block, SY_XBLOCK, 
                   CoordModeOrigin);
   };
   if (pointDim%SY_XBLOCK != 0)
     XDrawPoints  (display, winId, gc, cur_point_block, pointDim%SY_XBLOCK, 
                   CoordModeOrigin); */

   for (cur_arc_block = arcPtr, count = 0; count < pointDim/SY_XBLOCK;
        cur_arc_block += SY_XBLOCK, count++){
     XFillArcs  (display, winId, gc, cur_arc_block, SY_XBLOCK);
   };
   if (pointDim%SY_XBLOCK != 0)
     XFillArcs  (display, winId, gc, cur_arc_block, pointDim%SY_XBLOCK);

   for (cur_segment_block = segmentPtr, count = 0; count < segmentDim/SY_XBLOCK;
        cur_segment_block += SY_XBLOCK, count++){
     XDrawSegments (display, winId, gc, cur_segment_block, SY_XBLOCK);
   };
   if (segmentDim%SY_XBLOCK != 0)
     XDrawSegments (display, winId, gc, cur_segment_block, segmentDim%SY_XBLOCK);

   return;
}

static void Sy_Hipo2X (SyHipoContainer *container, XPoint *pointPtr, int *pointDim, 
XSegment *segmentPtr, int *segmentDim, int *winWidth, int *winHeight, 
int xZen, int yZen, float uscale, SyIpoMappingControl *mcontrol)
{
  float    scaleX, scaleY, scaleXY;
  float    xOff,   yOff, 
           xShift, yShift;

  int      maxX, minX, maxY, minY;
  int      dimX, dimY; 
  int      effDimX, effDimY;

  int      tail_ipo, head_ipo;
  int      sDim, pDim, i, cptr;

  SyIpo      *ipo;
  SyIpoFloat *fipo;
  SyHipoData *hipodata;

  
  if (mcontrol->is_float_contour){
    
    maxX = minX = container->fipo[0][1].x;
    maxY = minY = container->fipo[0][1].y;
    
    for (cptr = 0; cptr < container->size; cptr++){
      fipo = container->fipo[cptr];

      for (i = 1; i <= container->iposize[cptr]; i++){
	maxX = max(maxX, fipo[i].x);
	minX = min(minX, fipo[i].x);
	maxY = max(maxY, fipo[i].y);
	minY = min(minY, fipo[i].y);
      }
    }
    
    dimX = maxX - minX;
    dimY = maxY - minY;
    
    if (uscale > 0){
      effDimX = max(min(dimX * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
      effDimY = max(min(dimY * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
      xOff    = (xZen > 0) ? xZen : dimX/2 + minX;
      yOff    = (yZen > 0) ? yZen : dimY/2 + minY;
      scaleXY = uscale;
    }
    else{
      effDimX = max(min(dimX, SY_DEF_WINDIM), SY_MIN_WINDIM);
      effDimY = max(min(dimY, SY_DEF_WINDIM), SY_MIN_WINDIM);
      scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
      scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
      scaleXY = min(scaleX, scaleY);
      xOff    = dimX/2 + minX;
      yOff    = dimY/2 + minY;
    }
    
    xShift  = effDimX/2 + SY_MIN_BORDER;
    yShift  = effDimY/2 + SY_MIN_BORDER;
    
    
    for (cptr = 0, sDim = pDim = 0; cptr < container->size; cptr++){
      fipo = container->fipo[cptr];

      hipodata = container->hipodata[cptr];      
      for (i = 1; i < container->iposize[cptr]; i++){
	if (hipodata[i].is_anchor && hipodata[i].same_boundary &&
	    (hipodata[i].min_hier <= mcontrol->lo_hier) &&
	    (hipodata[i].max_hier >= mcontrol->hi_hier)){

	  arcPtr[pDim].x = scaleXY * ((float)fipo[i].x - xOff) + xShift - SY_ARCWIDTH/2;
	  arcPtr[pDim].y = scaleXY * ((float)fipo[i].y - yOff) + yShift - SY_ARCWIDTH/2; 
	  arcPtr[pDim].width  = arcPtr[pDim].height = SY_ARCWIDTH;
	  arcPtr[pDim].angle1 = 0; 
	  arcPtr[pDim].angle2 = 360*64;
	  pDim++;
	}
      }

      for (i = 1; i < container->iposize[cptr]; i++){
	
	segmentPtr[sDim].x1 = 
	  scaleXY * ((float)fipo[i].x - xOff) + xShift;
	segmentPtr[sDim].y1 = 
	  scaleXY * ((float)fipo[i].y - yOff) + yShift; 
	segmentPtr[sDim].x2 = 
	  scaleXY * ((float)fipo[i+1].x - xOff) + xShift;
	segmentPtr[sDim].y2 = 
	  scaleXY * ((float)fipo[i+1].y - yOff) + yShift;
	sDim++;      
      }
      
      segmentPtr[sDim].x1 = 
	scaleXY * ((float)fipo[container->iposize[cptr]].x - xOff) + xShift;
      segmentPtr[sDim].y1 = 
	scaleXY * ((float)fipo[container->iposize[cptr]].y - yOff) + yShift; 
      segmentPtr[sDim].x2 = 
	scaleXY * ((float)fipo[1].x - xOff) + xShift;
      segmentPtr[sDim].y2 = 
	scaleXY * ((float)fipo[1].y - yOff) + yShift;
      sDim++;      
    }
  }
  
  else{
    
    maxX = minX = container->ipo[0][1].x;
    maxY = minY = container->ipo[0][1].y;
    
    for (cptr = 0; cptr < container->size; cptr++){
      ipo = container->ipo[cptr];
      
      for (i = 1; i <= container->iposize[cptr]; i++){
	maxX = max(maxX, ipo[i].x);
	minX = min(minX, ipo[i].x);
	maxY = max(maxY, ipo[i].y);
	minY = min(minY, ipo[i].y);
      }
    }
    
    dimX = maxX - minX;
    dimY = maxY - minY;
    
    if (uscale > 0){
      effDimX = max(min(dimX * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
      effDimY = max(min(dimY * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
      xOff    = (xZen > 0) ? xZen : dimX/2 + minX;
      yOff    = (yZen > 0) ? yZen : dimY/2 + minY;
      scaleXY = uscale;
    }
    else{
      effDimX = max(min(dimX, SY_DEF_WINDIM), SY_MIN_WINDIM);
      effDimY = max(min(dimY, SY_DEF_WINDIM), SY_MIN_WINDIM);
      scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
      scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
      scaleXY = min(scaleX, scaleY);
      xOff    = dimX/2 + minX;
      yOff    = dimY/2 + minY;
    }
    
    xShift  = effDimX/2 + SY_MIN_BORDER;
    yShift  = effDimY/2 + SY_MIN_BORDER;
    
    
    for (cptr = 0, sDim = pDim = 0; cptr < container->size; cptr++){
      ipo = container->ipo[cptr];

      hipodata = container->hipodata[cptr];      
      for (i = 1; i < container->iposize[cptr]; i++){
	if (hipodata[i].is_anchor && hipodata[i].same_boundary &&
	    (hipodata[i].min_hier <= mcontrol->lo_hier) &&
	    (hipodata[i].max_hier >= mcontrol->hi_hier)){

	  arcPtr[pDim].x = scaleXY * ((float)ipo[i].x - xOff) + xShift - SY_ARCWIDTH/2;
	  arcPtr[pDim].y = scaleXY * ((float)ipo[i].y - yOff) + yShift - SY_ARCWIDTH/2; 
	  arcPtr[pDim].width  = arcPtr[pDim].height = SY_ARCWIDTH;
	  arcPtr[pDim].angle1 = 0; 
	  arcPtr[pDim].angle2 = 360*64;
	  pDim++;
	}
      }
      
      for (i = 1; i < container->iposize[cptr]; i++){
	
	segmentPtr[sDim].x1 = 
	  scaleXY * ((float)ipo[i].x - xOff) + xShift;
	segmentPtr[sDim].y1 = 
	  scaleXY * ((float)ipo[i].y - yOff) + yShift; 
	segmentPtr[sDim].x2 = 
	  scaleXY * ((float)ipo[i+1].x - xOff) + xShift;
	segmentPtr[sDim].y2 = 
	  scaleXY * ((float)ipo[i+1].y - yOff) + yShift;
	sDim++;      
      }
      
      segmentPtr[sDim].x1 = 
	scaleXY * ((float)ipo[container->iposize[cptr]].x - xOff) + xShift;
      segmentPtr[sDim].y1 = 
	scaleXY * ((float)ipo[container->iposize[cptr]].y - yOff) + yShift; 
      segmentPtr[sDim].x2 = 
	scaleXY * ((float)ipo[1].x - xOff) + xShift;
      segmentPtr[sDim].y2 = 
	scaleXY * ((float)ipo[1].y - yOff) + yShift;
      sDim++;      
    }    
  }  

  *winWidth   = effDimX + 2*SY_MIN_BORDER;
  *winHeight  = effDimY + 2*SY_MIN_BORDER;
  *segmentDim = sDim;
  *pointDim   = pDim;
}
#endif
