/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char sy_DrawPyramidId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Graphics.h"
#include "skeleton.xbm"

#ifndef NOX
#include <X11/Xatom.h>
static Atom proto_atom = None, delete_atom = None;

typedef struct{
   Window  winId;
   char    *tag;
   void   (*eventHandler)();
} SyLtEventAction;

/* static line_width[] = {0,6,3,1,1,1,1,1,1,1,1,1}; 
   static line_width[] = {0,6,2,2,2,2,2,2,2,2,2,2}; 
   static line_width[] = {0,2,2,2,2,2,2,2,2,2,2,2}; */

static line_width[]  = {1,1,1,1,1,1,1,1,1,1,1,1};
static n_line_types  = 12;

static XEvent  LtXevent;

static GC      gc;
static Pixmap  icon;

static XPoint   *pointPtr;
static XSegment *segmentPtr;
static int      pointDim, segmentDim;

static Display *display;
static char    *displayName = NULL;
static int     screenNr;

static void  Sy_CatchEvent(XEvent *event, Window winId, unsigned int winNr, 
			   char *tag);

static void  Sy_ExposeEvent(XExposeEvent *event, Window winId, 
			    unsigned int winNr);

static void  Sy_Pyramid2X(SySkelDAC *vsk, SyIpoDAC *ipd, int rank, 
			  XPoint *pointPtr, int *pointDim, XSegment *segmentPtr, 
			  int *segmentDim, int *winWidth, int *winHeight, int xZen, 
			  int yZen, float uscale, enum SyDisplayMode display_mode, 
			  enum SyDisplaySwitch boundary_toggle, int *pyrlab);

static void  Sy_Pyramid2XRaw(SySkelDAC *vsk, SyIpoDAC *ipd, int rank, 
			     XPoint *pointPtr, int *pointDim, XSegment *segmentPtr, 
			     int *segmentDim, enum SyDisplayMode display_mode, 
			     enum SyDisplaySwitch boundary_toggle, int *pyrlab);
#endif

int Sy_XPyramidDisp (SySkelDAC *vsk, SyIpoDAC *ipd, int rank, char *windowName, 
		     char *iconName, int xPos, int yPos, int xZen, int yZen, 
		     float uscale, char *userDisplay, SyDisplayControl control, 
		     int *pyrlab)
{
#ifndef NOX
  SyLtEventAction  eventBuff;
  int              winWidth, winHeight;
  int              borderWidth = SY_DEF_BORDERWIDTH;
  int              i;

  displayName  = userDisplay != NULL ? userDisplay : NULL;
  if ((display = XOpenDisplay (displayName)) == NULL) return (-1);  
  screenNr     = XDefaultScreen (display);

  pointPtr   = (XPoint*)   Sy_Malloc ((unsigned)((vsk->pdim +1) * sizeof(XPoint)));
  segmentPtr = (XSegment*) Sy_Malloc ((unsigned)((vsk->edim + ipd->icndim +2) * 
				   sizeof(XSegment)));

  if (control.set_window == SKE_ON){
    Sy_Pyramid2XRaw (vsk, ipd, rank, pointPtr, &pointDim, segmentPtr, &segmentDim, 
           control.display_mode, control.show_boundary, pyrlab); 
    winWidth  = control.window_dimX;
    winHeight = control.window_dimY;
  }
  else{
    Sy_Pyramid2X (vsk, ipd, rank, pointPtr, &pointDim, segmentPtr, &segmentDim, 
	      &winWidth, &winHeight, xZen, yZen, uscale, control.display_mode,
              control.show_boundary, pyrlab); 
  }

  Sy_DestroySkeleton (vsk);
  Sy_DestroyIpo (ipd);

  eventBuff.winId = XCreateSimpleWindow (display, 
					 RootWindow (display, screenNr), 
					 xPos, yPos, winWidth, 
					 winHeight, borderWidth,
					 BlackPixel (display, screenNr),
					 WhitePixel (display, screenNr));
  
  gc   = XCreateGC (display, eventBuff.winId, 0, 0);
  icon = XCreateBitmapFromData (display, eventBuff.winId, vskel_icon_bits, 
                                vskel_icon_width, vskel_icon_height);
  XSetForeground (display, gc, BlackPixel (display, screenNr));
  XSetBackground (display, gc, WhitePixel (display, screenNr));
  XSetStandardProperties (display, eventBuff.winId, windowName, 
			  iconName , icon, NULL, 0, NULL);

  XSelectInput (display, eventBuff.winId, 
		ExposureMask | StructureNotifyMask);

  proto_atom = XInternAtom(display, "WM_PROTOCOLS", False);
  delete_atom = XInternAtom(display, "WM_DELETE_WINDOW", False);
  if ((proto_atom != None) && (delete_atom != None))
    XChangeProperty(display, eventBuff.winId, proto_atom, XA_ATOM, 32,
		    PropModeReplace, (unsigned char*) &delete_atom, 1);

  eventBuff.eventHandler = Sy_CatchEvent;
   
  XMapWindow (display, eventBuff.winId);

  if (control.line_width > 0){
    for (i = 0; i < n_line_types; i++) line_width[i] *= control.line_width;
  }

  XSetLineAttributes(display, gc, line_width[rank], 
		     LineSolid, CapRound, JoinRound);

  while (LtXevent.type != Expose && LtXevent.xany.window != eventBuff.winId) 
    XNextEvent(display, &LtXevent);
  while (1){
    if (LtXevent.xany.window == eventBuff.winId)
      eventBuff.eventHandler (&LtXevent, LtXevent.xany.window, 1, NULL);
    XNextEvent (display, &LtXevent);
  }
#else
  Sy_Message("%s", "no X11 graphics support available\n");
#endif
}

#ifndef NOX
static void Sy_CatchEvent (XEvent *event, Window winId, 
			   unsigned int winNr, char *tag)
{
  XClientMessageEvent *me;
  switch (event->type){
  case Expose :
    Sy_ExposeEvent ((XExposeEvent*)event, winId, winNr); 
    break;
  case DestroyNotify:
    exit(2); break;
  case ClientMessage:
    me = (XClientMessageEvent*)event;
    if ((me->window == winId) && (me->data.l[0] == delete_atom)){
      exit(2);
    }
    break;
  default : break;
  }
  return;
}

static void Sy_ExposeEvent (XExposeEvent *event, Window winId, 
			    unsigned int winNr)
{
   XPoint   *cur_point_block;
   XSegment *cur_segment_block;
   int      count;

   for (cur_point_block = pointPtr, count = 0; count < pointDim/SY_XBLOCK;
        cur_point_block += SY_XBLOCK, count++){
     XDrawPoints  (display, winId, gc, cur_point_block, SY_XBLOCK, 
		   CoordModeOrigin);
   };
   if (pointDim%SY_XBLOCK != 0)
     XDrawPoints  (display, winId, gc, cur_point_block, pointDim%SY_XBLOCK, 
		   CoordModeOrigin);

   for (cur_segment_block = segmentPtr, count = 0; count < segmentDim/SY_XBLOCK;
        cur_segment_block += SY_XBLOCK, count++){
     XDrawSegments (display, winId, gc, cur_segment_block, SY_XBLOCK);
   };
   if (segmentDim%SY_XBLOCK != 0)
     XDrawSegments (display, winId, gc, cur_segment_block, segmentDim%SY_XBLOCK);

   return;
}

static void Sy_Pyramid2X (SySkelDAC *vsk, SyIpoDAC *ipd, int rank, XPoint *pointPtr, 
			  int *pointDim, XSegment *segmentPtr, int *segmentDim, 
			  int *winWidth, int *winHeight, int xZen, int yZen, 
			  float uscale, enum SyDisplayMode display_mode, 
			  enum SyDisplaySwitch boundary_toggle, int *pyrlab)
{
  SyVertex   vhead, vtail;
  int	   i;
  float    x1, y1, x2, y2;
  float    scaleX, scaleY, scaleXY;
  float    xOff, yOff, 
           xShift, yShift;
  float    rubber;
  float    dirX, dirY;
  int      maxX, minX, minY;
  int      dimX, dimY, sDim; 
  int      effDimX, effDimY;
  int      add_segment, conn_flag;
  int      conn[4];
  int      tail_vertex, head_vertex, 
           left_point, right_point,
           tail_ipo, head_ipo;
  int      is_render, is_foreground;

  maxX = minX = vsk->point[1].x;
  minY = max(vsk->point[1].y,SY_DEF_MINY);

  for (i = 1; i <= vsk->pdim; i++){
     maxX = max(maxX, vsk->point[i].x);
     minX = min(minX, vsk->point[i].x);
  }

  dimX = maxX-minX;
  dimY = vsk->point[vsk->pdim].y - minY;

  if (uscale > 0){
    effDimX = max(min(dimX * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
    effDimY = max(min(dimY * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
    xOff    = (xZen > 0) ? xZen : dimX/2 + minX;
    yOff    = (yZen > 0) ? yZen : dimY/2 + minY;
    scaleXY = uscale;
  }
  else{
    effDimX = max(min(dimX, SY_DEF_WINDIM), SY_MIN_WINDIM);
    effDimY = max(min(dimY, SY_DEF_WINDIM), SY_MIN_WINDIM);
    scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
    scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
    scaleXY = min(scaleX, scaleY);
    xOff    = dimX/2 + minX;
    yOff    = dimY/2 + minY;
  }

  xShift  = effDimX/2 + SY_MIN_BORDER;
  yShift  = effDimY/2 + SY_MIN_BORDER;

  if (boundary_toggle == SKE_ON){
    for (i = 1; i <= vsk->pdim; i++){
      pointPtr[i-1].x = scaleXY * (vsk->point[i].x - xOff) + xShift;
      pointPtr[i-1].y = scaleXY * (vsk->point[i].y - yOff) + yShift;
    }
  }

  for (i = 1, sDim = 0; i <= vsk->edim; i++) {
    
    tail_vertex = vsk->edge[i].tail_vertex;
    head_vertex = vsk->edge[i].head_vertex;
    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;

    conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);

    is_render = (pyrlab[i] > 0) && (rank >= pyrlab[i]) && 
      (!conn_flag || vsk->is_border);

    if (is_render){

      vtail = vsk->vertex[vsk->edge[i].tail_vertex];
      vhead = vsk->vertex[vsk->edge[i].head_vertex];

      is_foreground  = Sy_SkeletonType (vsk, ipd, tail_vertex, i);

      if ((is_foreground && (display_mode == SKE_BACKGROUND)) ||
	  (!is_foreground && (display_mode == SKE_FOREGROUND))) continue;

      x1 = scaleXY * (vtail.xden / vtail.num - xOff) + xShift;
      y1 = scaleXY * (vtail.yden / vtail.num - yOff) + yShift;
      x2 = scaleXY * (vhead.xden / vhead.num - xOff) + xShift;
      y2 = scaleXY * (vhead.yden / vhead.num - yOff) + yShift;

      switch (vsk->edge[i].rank) {
      case 0:
	dirX = x2 - x1;
	dirY = y2 - y1;
	rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY));
	x2 += rubber * dirX;
	y2 += rubber * dirY;
	x1 -= rubber * dirX;
	y1 -= rubber * dirY;
	add_segment = 1;
	break;
      case 1:
	if (!(add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && 
			     (SY_ALV(y1) < SY_PLANESIZE)))) break;
	dirX = x2 - x1;
	dirY = y2 - y1;
	rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY));
	x2 = x1 + rubber * dirX;
	y2 = y1 + rubber * dirY;
	break;
      default:
	if (add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE)
			   && (SY_ALV(x2) < SY_PLANESIZE) && 
			   (SY_ALV(y2) < SY_PLANESIZE))) break;
	if (add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE))) {
	  dirX = x2 - x1;
	  dirY = y2 - y1;
	  rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY)); 
	  x2 = x1 + rubber * dirX;
	  y2 = y1 + rubber * dirY;
	  break;
	}
	if (add_segment = ((SY_ALV(x2) < SY_PLANESIZE) && (SY_ALV(y2) < SY_PLANESIZE))) {
	  dirX = x1 - x2;
	  dirY = y1 - y2;
	  rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY)); 
	  x1 = x2 + rubber * dirX;
	  y1 = y2 + rubber * dirY;
	  break;
	}
      }
      if (add_segment) {
	segmentPtr[sDim].x1 = x1;
	segmentPtr[sDim].y1 = y1;
	segmentPtr[sDim].x2 = x2;
	segmentPtr[sDim].y2 = y2;
        sDim++;
      }
    }
  }

  if (boundary_toggle == SKE_ON){
    for (i = 1; i <= ipd->icndim; i++) {
      if ((tail_ipo = ipd->icn[i].tail_ipo) && (head_ipo = ipd->icn[i].head_ipo)){
	segmentPtr[sDim].x1 = scaleXY * ((float)ipd->ipo[tail_ipo].x - xOff) + xShift;
	segmentPtr[sDim].y1 = scaleXY * ((float)ipd->ipo[tail_ipo].y - yOff) + yShift;
	segmentPtr[sDim].x2 = scaleXY * ((float)ipd->ipo[head_ipo].x - xOff) + xShift;
	segmentPtr[sDim].y2 = scaleXY * ((float)ipd->ipo[head_ipo].y - yOff) + yShift;
	sDim++;
      }
    }
  }

  *winWidth   = effDimX + 2*SY_MIN_BORDER;
  *winHeight  = effDimY + 2*SY_MIN_BORDER;
  *segmentDim = sDim;
  *pointDim   = vsk->pdim;
}

static void Sy_Pyramid2XRaw (SySkelDAC *vsk, SyIpoDAC *ipd, int rank, 
			     XPoint *pointPtr, int *pointDim, XSegment *segmentPtr, 
			     int *segmentDim, enum SyDisplayMode display_mode, 
			     enum SyDisplaySwitch boundary_toggle, int *pyrlab)
{
  SyVertex   vhead, vtail;
  int	   i;
  float    x1, y1, x2, y2;
  float    rubber;
  float    dirX, dirY;
  int      sDim; 
  int      add_segment, conn_flag;
  int      conn[4];
  int      tail_vertex, head_vertex, 
           left_point, right_point,
           tail_ipo, head_ipo;
  int      tail_flag, head_flag;
  int      is_foreground, is_render;


  if (boundary_toggle == SKE_ON){
    for (i = 1; i <= vsk->pdim; i++){
      pointPtr[i-1].x = vsk->point[i].x;
      pointPtr[i-1].y = vsk->point[i].y;
    }
  }

  for (i = 1, sDim = 0; i <= vsk->edim; i++) {
    
    tail_vertex = vsk->edge[i].tail_vertex;
    head_vertex = vsk->edge[i].head_vertex;
    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;

    is_foreground  = Sy_SkeletonType (vsk, ipd, tail_vertex, i);

    if ((is_foreground && (display_mode == SKE_BACKGROUND)) ||
        (!is_foreground && (display_mode == SKE_FOREGROUND))) continue;

    conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].tail_icn != 0));

    conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		(ipd->ipl[left_point].head_icn != 0));

    conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);

    is_render = (pyrlab[i] > 0) && (rank >= pyrlab[i]) && 
      (!conn_flag || vsk->is_border);

    if (is_render){

      vtail = vsk->vertex[vsk->edge[i].tail_vertex];
      vhead = vsk->vertex[vsk->edge[i].head_vertex];

      x1 = vtail.xden / vtail.num;
      y1 = vtail.yden / vtail.num;
      x2 = vhead.xden / vhead.num;
      y2 = vhead.yden / vhead.num;

      switch (vsk->edge[i].rank) {
      case 0:
	dirX = x2 - x1;
	dirY = y2 - y1;
	rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY));
	x2 += rubber * dirX;
	y2 += rubber * dirY;
	x1 -= rubber * dirX;
	y1 -= rubber * dirY;
	add_segment = 1;
	break;
      case 1:
	if (!(add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && 
			     (SY_ALV(y1) < SY_PLANESIZE)))) break;
	dirX = x2 - x1;
	dirY = y2 - y1;
	rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY));
	x2 = x1 + rubber * dirX;
	y2 = y1 + rubber * dirY;
	break;
      default:
	if (add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE)
			   && (SY_ALV(x2) < SY_PLANESIZE) && 
			   (SY_ALV(y2) < SY_PLANESIZE))) break;
	if (add_segment = ((SY_ALV(x1) < SY_PLANESIZE) && (SY_ALV(y1) < SY_PLANESIZE))) {
	  dirX = x2 - x1;
	  dirY = y2 - y1;
	  rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY)); 
	  x2 = x1 + rubber * dirX;
	  y2 = y1 + rubber * dirY;
	  break;
	}
	if (add_segment = ((SY_ALV(x2) < SY_PLANESIZE) && (SY_ALV(y2) < SY_PLANESIZE))) {
	  dirX = x1 - x2;
	  dirY = y1 - y2;
	  rubber = SY_VALRAD/max(SY_ALV(dirX), SY_ALV(dirY)); 
	  x1 = x2 + rubber * dirX;
	  y1 = y2 + rubber * dirY;
	  break;
	}
      }
      if (add_segment) {
	segmentPtr[sDim].x1 = x1;
	segmentPtr[sDim].y1 = y1;
	segmentPtr[sDim].x2 = x2;
	segmentPtr[sDim].y2 = y2;
        sDim++;
      }
    }
  }

  if (boundary_toggle == SKE_ON){
    for (i = 1; i <= ipd->icndim; i++) {
      if ((tail_ipo = ipd->icn[i].tail_ipo) && (head_ipo = ipd->icn[i].head_ipo)){
	segmentPtr[sDim].x1 = ipd->ipo[tail_ipo].x;
	segmentPtr[sDim].y1 = ipd->ipo[tail_ipo].y; 
	segmentPtr[sDim].x2 = ipd->ipo[head_ipo].x;
	segmentPtr[sDim].y2 = ipd->ipo[head_ipo].y;
	sDim++;
      }
    }
  }

  *segmentDim = sDim;
  *pointDim   = (boundary_toggle == SKE_ON)? vsk->pdim : 0;
}
#endif
