/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char sy_DrawReconId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Graphics.h"
#include "inipoints.xbm"

#ifndef NOX
#include <X11/Xatom.h>
static Atom proto_atom = None, delete_atom = None;

typedef struct{
   Window  winId;
   char    *tag;
   void   (*eventHandler)();
} SyLtEventAction;

static XEvent  LtXevent;

static GC      gc;
static Pixmap  icon;

static XArc     *arcPtr;
static int      arcDim;

static Display *display;
static char    *displayName = NULL;
static int     screenNr;

static void Sy_CatchEvent(XEvent *event, Window winId, unsigned int winNr, char *tag);
static void Sy_ExposeEvent(XExposeEvent *event, Window winId, unsigned int winNr);
static void Sy_Vertex2XArc(SySkelDAC *vsk, char *foreground, XArc *arcPtr, int *arcDim, int *winWidth, int *winHeight, int xZen, int yZen, float uscale, float thres, SyDisplayControl control);
static void Sy_Vertex2XArcRaw(SySkelDAC *vsk, char *foreground, XArc *arcPtr, int *arcDim, float thres, SyDisplayControl control);
#endif

int Sy_XReconstruct (SySkelDAC *vsk, char *fginfo, float thres, char *windowName, char *iconName, int xPos, int yPos, int xZen, int yZen, float uscale, char *userDisplay, SyDisplayControl control)
{
#ifndef NOX
  SyLtEventAction  eventBuff;
  int            winWidth, winHeight;
  int            borderWidth = SY_DEF_BORDERWIDTH;
  int            arc_mem, page;
  
  page         = getpagesize();
  arc_mem      = (((vsk->vdim + 1)*sizeof (XArc))/page+1)*page;

  arcPtr       = (XArc*) Sy_Malloc ((unsigned)arc_mem);

  if (control.raw_mode == SKE_ON){
    Sy_Vertex2XArcRaw (vsk, fginfo, arcPtr, &arcDim, thres, control); 
    winWidth  = control.window_dimX;
    winHeight = control.window_dimY;
  }
  else{
    Sy_Vertex2XArc (vsk, fginfo, arcPtr, &arcDim, 
		    &winWidth, &winHeight, xZen, yZen, uscale, thres, control); 
  }

  displayName  = userDisplay != NULL ? userDisplay : NULL;
  if ((display = XOpenDisplay (displayName)) == NULL) return (-1);  
  screenNr     = XDefaultScreen (display);

  eventBuff.winId = XCreateSimpleWindow (display, 
					 RootWindow (display, screenNr), 
					 xPos, yPos, winWidth, 
					 winHeight, borderWidth,
					 BlackPixel (display, screenNr),
					 WhitePixel (display, screenNr));
  
  gc   = XCreateGC (display, eventBuff.winId, 0, 0);
   
  XSetFillStyle (display, gc, FillSolid);
  
  icon = XCreateBitmapFromData (display, eventBuff.winId, ipo_icon_bits, 
                                ipo_icon_width, ipo_icon_height);
  XSetForeground (display, gc, BlackPixel (display, screenNr));
  XSetBackground (display, gc, WhitePixel (display, screenNr));
  XSetStandardProperties (display, eventBuff.winId, windowName, 
			  iconName , icon, NULL, 0, NULL);

  XSelectInput (display, eventBuff.winId, 
		ExposureMask | StructureNotifyMask);

  proto_atom = XInternAtom(display, "WM_PROTOCOLS", False);
  delete_atom = XInternAtom(display, "WM_DELETE_WINDOW", False);
  if ((proto_atom != None) && (delete_atom != None))
    XChangeProperty(display, eventBuff.winId, proto_atom, XA_ATOM, 32,
		    PropModeReplace, (unsigned char*) &delete_atom, 1);

  eventBuff.eventHandler = Sy_CatchEvent;
  XMapWindow (display, eventBuff.winId);

  while (LtXevent.type != Expose && LtXevent.xany.window != eventBuff.winId) 
    XNextEvent(display, &LtXevent);
  while (1){
    if (LtXevent.xany.window == eventBuff.winId)
      eventBuff.eventHandler (&LtXevent, LtXevent.xany.window, 1, NULL);
    XNextEvent (display, &LtXevent);
  }
#else
  Sy_Message("%s", "no X11 graphics support available\n");
#endif
}

#ifndef NOX
static void Sy_CatchEvent (XEvent *event, Window winId, 
			   unsigned int winNr, char *tag)
{
  XClientMessageEvent *me;
  switch (event->type){
  case Expose :
    Sy_ExposeEvent ((XExposeEvent*)event, winId, winNr); 
    break;
  case DestroyNotify:
    exit(2); break;
  case ClientMessage:
    me = (XClientMessageEvent*)event;
    if ((me->window == winId) && (me->data.l[0] == delete_atom)){
      exit(2);
    }
    break;
  default : break;
  }
  return;
}

static void Sy_ExposeEvent (XExposeEvent *event, Window winId, 
			    unsigned int winNr)
{
   XArc   *cur_point_block;
   int      count;
   for (cur_point_block = arcPtr, count = 0; count < arcDim/SY_XBLOCK;
        cur_point_block += SY_XBLOCK, count++){
     XFillArcs  (display, winId, gc, cur_point_block, SY_XBLOCK);
   };
   if (arcDim%SY_XBLOCK != 0)
     XFillArcs  (display, winId, gc, cur_point_block, arcDim%SY_XBLOCK);

   return;
}

static void Sy_Vertex2XArc (SySkelDAC *vsk, char *foreground, XArc *arcPtr, int *arcDim, int *winWidth, int *winHeight, int xZen, int yZen, float uscale, float thres, SyDisplayControl control)
{
  float    scaleX, scaleY, scaleXY;
  float    xOff,   yOff, 
           xShift, yShift;
  float    x1, y1, ext, radius;

  int      maxX, minX, maxY, minY;
  int      dimX, dimY; 
  int      effDimX, effDimY;

  int      tail_point, head_point;
  int      aDim, i, valid;

  maxX = minX = vsk->point[3].x;
  minY = vsk->point[3].y;
  maxY = vsk->point[vsk->pdim].y;

  for (i = 1; i <= vsk->pdim; i++){
     maxX = max(maxX, vsk->point[i].x);
     minX = min(minX, vsk->point[i].x);
  }

  dimX = maxX - minX;
  dimY = maxY - minY;

  if (uscale > 0){
    effDimX = max(min(dimX * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
    effDimY = max(min(dimY * uscale, SY_MAX_WINDIM), SY_MIN_WINDIM);
    xOff    = (xZen > 0) ? xZen : dimX/2 + minX;
    yOff    = (yZen > 0) ? yZen : dimY/2 + minY;
    scaleXY = uscale;
  }
  else{
    effDimX = max(min(dimX, SY_DEF_WINDIM), SY_MIN_WINDIM);
    effDimY = max(min(dimY, SY_DEF_WINDIM), SY_MIN_WINDIM);
    scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
    scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
    scaleXY = min(scaleX, scaleY);
    xOff    = dimX/2 + minX;
    yOff    = dimY/2 + minY;
  }

  xShift  = effDimX/2 + SY_MIN_BORDER;
  yShift  = effDimY/2 + SY_MIN_BORDER;

  if (control.display_mode == SKE_PLAIN) 
    control.display_mode = SKE_FOREGROUND;

  for (i = 1, aDim = 0; i <= vsk->vdim; i++){
    valid = ((foreground[i] == SY_FOREGRND) && (control.display_mode == SKE_FOREGROUND) ||
	     ((foreground[i] == SY_BACKGRND) && (control.display_mode == SKE_BACKGROUND)));
    if (valid && (vsk->regul[i].birth[2] >= thres)){
      x1     = vsk->vertex[i].xden/vsk->vertex[i].num;
      y1     = vsk->vertex[i].yden/vsk->vertex[i].num;
      radius = vsk->regul[i].radius;
      ext    = max(SY_ALV(x1),SY_ALV(y1)) + radius;
      if (ext < SY_PLANESIZE){
	arcPtr[aDim].x = scaleXY * (x1 -radius - xOff) + xShift;
	arcPtr[aDim].y = scaleXY * (y1 -radius - yOff) + yShift;
	arcPtr[aDim].width  = arcPtr[aDim].height = 2.0 * scaleXY * radius;
	arcPtr[aDim].angle1 = arcPtr[aDim].angle2 = 360*64;
	aDim++;
      }
    }
  }

  *winWidth   = effDimX + 2*SY_MIN_BORDER;
  *winHeight  = effDimY + 2*SY_MIN_BORDER;
  *arcDim      = aDim-1;
}

static void Sy_Vertex2XArcRaw (SySkelDAC *vsk, char *foreground, XArc *arcPtr, int *arcDim, float thres, SyDisplayControl control)
{
  float    x1, y1, ext, radius;
  int      tail_point, head_point;
  int      aDim, i, valid;

  if (control.display_mode == SKE_PLAIN) 
    control.display_mode = SKE_FOREGROUND;

  for (i = 1, aDim = 0; i <= vsk->vdim; i++){
    valid = ((foreground[i] == SY_FOREGRND) && (control.display_mode == SKE_FOREGROUND) ||
            ((foreground[i] == SY_BACKGRND) && (control.display_mode == SKE_BACKGROUND)));
    if (valid && (vsk->regul[i].birth[2] >= thres)){
      x1     = vsk->vertex[i].xden/vsk->vertex[i].num;
      y1     = vsk->vertex[i].yden/vsk->vertex[i].num;
      radius = vsk->regul[i].radius;
      ext    = max(SY_ALV(x1),SY_ALV(y1)) + radius;
      if (ext < SY_PLANESIZE){
	arcPtr[aDim].x = x1 - radius;
	arcPtr[aDim].y = y1 - radius;
	arcPtr[aDim].width  = arcPtr[aDim].height = 2.0 * radius;
	arcPtr[aDim].angle1 = arcPtr[aDim].angle2 = 360*64;
	aDim++;
      }
    }
  }
  *arcDim      = aDim-1;
}
#endif





