%{ 

/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Division of Applied Sciences,
 *          Harvard University, Cambridge, Massachusetts (rlo@hrl.harvard.edu)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1995 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/* yacc grammar rules for I/O routines */

#include <stdio.h>
#include "Sy_Lang.h"

extern SyAvlTree  *sy_yacc_tree;
extern SyAvlTree  *sy_func_tree;
extern SyIoObject *sy_global_obj_mem;

SyIoToken    sy_yacc_iotoken;

static SyIoObject   *tmp_obj_mem;
static SyDataObject *sy_yacc_fo;
static char* objnamestk[1024];
static stkptr = 0;

int   tmp_int;
float tmp_float;
char  strbuf[1024], tmp_str[1024];
int   is_data = 0;

static void call_loader(char *funcname);
static void call_object_loader(char *objname);
%}

%start entity

%union  {
	 int   ival;
	 float fval;
         char  chval;
	 char  strval[1024];
	}	

%token <ival>   INTEGER HEX OBJECT STOP ASSIGN ATERM 
%token <ival>   COMMENTSKIP COMMENTBEGIN COMMENTEND 
%token <ival>   MEMORY SHMEM FILEIO BEGINLOADSEQ 
%token <fval>   REAL
%token <chval>  SYMBOL
%token <strval> IDENT STRING 

%% /* beginning of rules section */

entity	:       objects 
        |       /* empty */
        |       error { printf ("%s\n", yytext); } ";" 
        ;

objects :       genobj
        |       objects genobj
	;

genobj  :       object
        |       assignment
        |       simpleval
        |       loadinstr
        ; 

object  :       OBJECT   {
#ifdef DEBUG               
                          printf("new object: "); 
#endif
                          }
 
                IDENT    {
#ifdef DEBUG               
                          printf("%s\n", tmp_str);
#endif 
		          Sy_AddIoObject(&sy_yacc_tree, tmp_str);
	                  objnamestk[stkptr++] = Sy_Strcpy(tmp_str);
                          sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
	        '{' 
                 valist 
                '}' { 
#ifdef DEBUG               
                      printf("exiting object\n");
	              printf ("Trying to call loader for %s\n",
	                      objnamestk[stkptr-1]);
#endif
 		      call_object_loader (objnamestk[stkptr-1]); 

		      stkptr--; free(objnamestk[stkptr]);
	              if (stkptr > 0){
	                (void)Sy_GetIoObject(&sy_yacc_tree, 
                                             objnamestk[stkptr-1]);
		      }
		      else{
			(void)Sy_GetIoObject(&sy_yacc_tree, "world");
		      }
		      sy_yacc_iotoken.yacc_tree = sy_yacc_tree;
                    }
        ; 

valist  :       /* empty */ 
        |       valist assignment
        |       valist simpleval
        |       valist loadinstr
        |       valist object
        ;

assignment :    IDENT ASSIGN INTEGER ATERM
                { 
#ifdef DEBUG
                  printf ("assign %d to %s\n", $3, $1);
#endif 
		  Sy_AddDataSameObject(&sy_yacc_tree, $1, SY_INT_T, &$3);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
           |    IDENT ASSIGN REAL    ATERM
                { 
#ifdef DEBUG
                  printf ("assign %f to %s\n", $3, $1); 
#endif
		  Sy_AddDataSameObject(&sy_yacc_tree, $1, SY_FLOAT_T, &$3);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
           |    IDENT ASSIGN HEX     ATERM
                { 
#ifdef DEBUG
                  printf ("assign %x to %s\n", $3, $1);
#endif
		  Sy_AddDataSameObject(&sy_yacc_tree, $1, SY_INT_T, &$3);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
           |    IDENT ASSIGN IDENT   ATERM
                { 
#ifdef DEBUG
                  printf ("assign %s to %s\n", $3, $1);
#endif 
		  Sy_AddDataSameObject(&sy_yacc_tree, $1, SY_STR_T, $3);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
           |    IDENT ASSIGN STRING  ATERM
	        { 
#ifdef DEBUG
                  printf ("assign %s to %s\n", $3, $1);
#endif 	
		  Sy_AddDataSameObject(&sy_yacc_tree, $1, SY_STR_T, $3);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
           ;

simpleval  :    INTEGER ATERM
           |    REAL    ATERM
           |    HEX     ATERM
           |    IDENT   ATERM
           |    STRING  ATERM
           ;

loadinstr  :    IDENT ASSIGN '(' MEMORY ')' HEX ATERM
	        { 
#ifdef DEBUG
                  printf ("assign from address %x to %s\n", $6, $1);
#endif
		  Sy_AddDataSameObject(&sy_yacc_tree, "memory", 
		  SY_INT_T, &$6);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
           |    IDENT ASSIGN '(' SHMEM  ')' HEX ATERM
		{ Sy_AddDataSameObject(&sy_yacc_tree, "shmem", 
		  SY_INT_T, &$6);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
           |    IDENT ASSIGN '(' FILEIO ')' STRING ATERM
                { Sy_AddDataSameObject(&sy_yacc_tree, "file", 
		  SY_STR_T, &$6);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;}
           |    IDENT ASSIGN BEGINLOADSEQ '(' IDENT ')'
	        {call_loader ($5);}
	   |    BEGINLOADSEQ '(' IDENT ')'
	        {call_loader ($3);}
           |    IDENT ASSIGN BEGINLOADSEQ 
	        {is_data = 1; yyclearin; 
#ifdef DEBUG
                 printf ("beginning default load seq\n");
#endif
                }
           ;
	
%% /* start of programs */

static void call_loader(char *funcname)
{
  int retval;

  is_data = 1; yyclearin; 
#ifdef DEBUG
  printf ("beginning load sequence, calling %s\n", funcname);
#endif
  tmp_obj_mem = sy_global_obj_mem;
  sy_yacc_fo = Sy_GetDataObject(&sy_func_tree, "callback", funcname);
  sy_global_obj_mem = tmp_obj_mem;
  if (sy_yacc_fo == NULL){
    fprintf (stderr, "Undefined loader function `%s'.\n", funcname);
    /* trying to skip over data segment */
    retval = Sy_SyncWithStream(sy_yacc_iotoken.in, SY_SYNC_COOKIE_BEGIN);
    if (retval == 0){
      fprintf (stderr, "Fatal error: could not synchronize with input data\n");
      exit(2);
    }
    retval = Sy_SyncWithStream(sy_yacc_iotoken.in, SY_SYNC_COOKIE_END);
    if (retval == 0){
      fprintf (stderr, "Fatal error: could not synchronize with input data\n");
      exit(2);
    }
  }
  else{
    sy_yacc_iotoken.funcname = sy_yacc_fo->name;
    retval = ((int(*)(SyIoToken*))
	      sy_yacc_fo->t.func_val)(&sy_yacc_iotoken);
    if (retval == 0){
      fprintf (stderr, "Fatal error: could not synchronize with input data\n");
      exit(2);
    }
  }
}

static void call_object_loader(char *objname)
{
  int retval, nchar;
  char *funcname = NULL;
  static char prefix[] = "Sy_Load_";
  int prefixlen, objnamelen;

  SyDataObject *dobj;

  dobj = Sy_GetDataObject (&sy_yacc_tree, objname, "type");
  if (dobj == NULL){
#ifdef DEBUG
    fprintf (stderr, "Undefined type of object `%s'.\n", objname);
#endif
    return;
  }

  /* create function name */
  prefixlen = strlen(prefix); objnamelen = strlen(dobj->t.s_val);
  nchar = prefixlen + objnamelen;
  funcname = (char*) Sy_Malloc (sizeof(char)*(nchar+1));
  strncpy(funcname,prefix,prefixlen);
  strncpy(funcname+prefixlen,dobj->t.s_val,objnamelen+1);

  tmp_obj_mem = sy_global_obj_mem;
  sy_yacc_fo  = Sy_GetDataObject(&sy_func_tree, "callback", funcname);
  sy_global_obj_mem = tmp_obj_mem;

#ifdef DEBUG
  fprintf (stderr, "calling loader %s\n", funcname);
#endif

  if (sy_yacc_fo == NULL){
#ifdef DEBUG
    fprintf (stderr, "Undefined loader for object `%s'.\n", objname);
#else
    ; /* do nothing */
#endif	
  }
  else{
    sy_yacc_iotoken.funcname = sy_yacc_fo->name;
    retval = ((int(*)(SyIoToken*))
	      sy_yacc_fo->t.func_val)(&sy_yacc_iotoken);
    if (retval == 0){
      fprintf (stderr, "Fatal error: could not load object `%s'\n", objname);
    }
  }
  Sy_Free(funcname);
}

#include "lex.yy.c"


