/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 *
 * Postscript file generation from SyIpo data
 *
 */

static char sy_Ipo2PSId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"
#include "Sy_Graphics.h"

typedef enum {SySlide, SyLetter, SyA4, SyDefault} SyPaperType;

/**********************************************************************
 * function Sy_Ipo2Postscript (ipd, xPos, yPos, xZen, yZen, uscale)
 *********************************************************************/
void Sy_Ipo2Postscript (SyIpoDAC *ipd, int xPos, int yPos, int xZen, int yZen, float uscale)
{
  float    scaleX, scaleY, scaleXY;
  float    xOff,   yOff, 
           xShift, yShift, x1, y1, x2, y2;

  int      maxX, minX, maxY, minY;
  int      dimX, dimY; 
  int      effDimX, effDimY, maxwindim;

  int      tail_ipo, head_ipo;
  int      sDim, i;
  int      bblx, bbly, bbux, bbuy;

  SyPaperType paper;
  float      paper_size_x, paper_size_y;

  switch (paper){
  case SySlide:
    paper_size_x = 7.0;
    paper_size_y = 10.5;
    break;
  case SyA4:
    paper_size_x = 8.26;
    paper_size_y = 11.69;
    break;
  case SyLetter:
    paper_size_x = 8.5;
    paper_size_y = 10.5;
    break;
  default:
    paper_size_x = 8.26;
    paper_size_y = 11.69;
    break;
  }

  fprintf (stdout, "%%!PS-Adobe-1.0\n");
  fprintf (stdout, "%%%%Title: Initiating Points\n");
  fprintf (stdout, "%%%%Creator: Sy_Ipo2Postscript\n");
  fprintf (stdout, "%%%%CreationDate:\n");
  fprintf (stdout, "%%%%For:\n");
  fprintf (stdout, "%%%%Pages: 1\n");

  maxwindim = SY_INCH2IPO (min(paper_size_x, paper_size_y));

  maxX = minX = ipd->ipo[3].x;
  minY = ipd->ipo[3].y;
  maxY = ipd->ipo[ipd->ipodim].y;

  for (i = 1; i <= ipd->ipodim; i++){
     maxX = max(maxX, ipd->ipo[i].x);
     minX = min(minX, ipd->ipo[i].x);
  }

  dimX = maxX - minX;
  dimY = maxY - minY;

  if (uscale > 0){
    effDimX = max(min(dimX * uscale, maxwindim), SY_MIN_WINDIM);
    effDimY = max(min(dimY * uscale, maxwindim), SY_MIN_WINDIM);
    xOff    = (xZen > 0) ? xZen : dimX/2 + minX;
    yOff    = (yZen > 0) ? yZen : dimY/2 + minY;
    scaleXY = uscale;
  }
  else{
    effDimX = max(min(dimX, maxwindim), SY_MIN_WINDIM);
    effDimY = max(min(dimY, maxwindim), SY_MIN_WINDIM);
    scaleX  = dimX > 0 ? (float )effDimX/(float )dimX : 1.0; 
    scaleY  = dimY > 0 ? (float )effDimY/(float )dimY : 1.0;
    scaleXY = min(scaleX, scaleY);
    xOff    = dimX/2 + minX;
    yOff    = dimY/2 + minY;
  }

  bblx = SY_MIN_BORDER;
  bbly = SY_MIN_BORDER;
  bbux = bblx + effDimX;
  bbuy = bbly + effDimY;

  fprintf (stdout, "%%%%BoundingBox: %d %d %d %d\n", bblx, bbly, bbux, bbuy);
  fprintf (stdout, "%%%%EndComments\n");

  fprintf (stdout, "/dot {0.3 0 360 arc fill} def\n");
  fprintf (stdout, "0.5 setlinewidth 2 setlinecap\n");

  xShift  = effDimX/2 + SY_MIN_BORDER;
  yShift  = effDimY/2 + SY_MIN_BORDER;

  for (i = 1; i <= ipd->ipodim; i++){
    x1 = scaleXY * ((float)ipd->ipo[i].x - xOff) + xShift;
    y1 = scaleXY * ((float)ipd->ipo[i].y - yOff) + yShift;
    fprintf (stdout, "%f %f dot\n", x1, y1);
  }

  for (i = 1, sDim = 0; i <= ipd->icndim; i++) {
    if ((tail_ipo = ipd->icn[i].tail_ipo) && (head_ipo = ipd->icn[i].head_ipo)){
      x1 = scaleXY * ((float)ipd->ipo[tail_ipo].x - xOff) + xShift;
      y1 = scaleXY * ((float)ipd->ipo[tail_ipo].y - yOff) + yShift; 
      x2 = scaleXY * ((float)ipd->ipo[head_ipo].x - xOff) + xShift;
      y2 = scaleXY * ((float)ipd->ipo[head_ipo].y - yOff) + yShift;
      sDim++;
    }
    fprintf (stdout, "newpath %f %f moveto %f %f lineto stroke closepath\n", 
	     x1, y1, x2, y2);
  }
  fprintf (stdout, "showpage\n");
}


