/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Division of Applied Sciences,
 *          Harvard University, Cambridge, Massachusetts. 
	    (rlo@hrl.harvard.edu)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1995 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

#include <stdio.h>
#include <math.h>
#include "Sy_InImage.h"
#ifdef HAS_GEN_LOADER
#include "image.h"
#include "options.h"
#else
#include "Sy_SunRaster.h"
#endif

char *ProgramName= "image loader";

#ifdef HAS_GEN_LOADER
void Sy_FreeInternalImage(Image *im);
#endif

#ifdef HAS_GEN_LOADER
SyImage_t *Sy_LoadImageFromFile (char *fname, SyImageType typ, 
				  char *formathint, int verbose)
{
  OptionSet    gopt;
  Option       typopt;
  Image        *image = NULL;
  SyImage_t    *rpic  = NULL;
  SyImBitMask  *ibm   = NULL;

  int i, x, y, r, g, b, linebytes, imval, iptr;
  SyBoolean is_gray_map = SY_FALSE;

  gopt.next    = NULL;
  gopt.options = NULL;

  if (formathint != NULL){
    gopt.options = &typopt;
    typopt.type = TYPE;
    typopt.info.type = formathint;
  }

  image = loadImage(&gopt, NULL, fname, verbose);
  if (!image) return NULL;

  rpic = (SyImage_t*) malloc (sizeof(SyImage_t));

/* #define IBAD    0  invalid image (used when freeing) */
/* #define IBITMAP 1  image is a bitmap */
/* #define IRGB    2  image is RGB */
/* #define ITRUE   3  image is true color */

  if (verbose){
/* fprintf (stderr, "height: %d,", image->height ); */
/* fprintf (stderr, "width: %d,", image->width ); */
   fprintf (stderr, "depth: %d,", image->depth );
/* fprintf (stderr, "type   = %d\n", image->type ); */
   fprintf (stderr, "pixlen: %d,", image->pixlen );
   fprintf (stderr, "rgb-map size: %d\n", image->rgb.size);
  }

#ifdef DEBUG
  for (i = 0; i < image->rgb.size; i++){
    printf ("<%d> %d %d %d \n", i, image->rgb.red[i] >> 8,
	                           image->rgb.green[i] >> 8,
                                   image->rgb.blue[i] >> 8);
  }
#endif

  rpic->height = image->height;
  rpic->width  = image->width;
  rpic->depth  = image->depth;

  if (image->pixlen > 3 || image->pixlen < 1){
    rpic->pixlen = -1;
  }
  else rpic->pixlen = image->pixlen;

  rpic->gamma    = image->gamma;
  rpic->rgb.size = image->rgb.size;

  if (image->rgb.size > 0){
    rpic->rgb.red   = (unsigned char*) malloc (sizeof(unsigned char)*
					     rpic->rgb.size);
    rpic->rgb.green = (unsigned char*) malloc (sizeof(unsigned char)*
					     rpic->rgb.size);
    rpic->rgb.blue  = (unsigned char*) malloc (sizeof(unsigned char)*
					     rpic->rgb.size);

    is_gray_map = SY_TRUE;

    for (i = 0; i < rpic->rgb.size; i++){
      r = (image->rgb.red[i]   + 128) >> 8;
      g = (image->rgb.green[i] + 128) >> 8;
      b = (image->rgb.blue[i]  + 128) >> 8;

      if (r > 255) r = 255;
      if (g > 255) g = 255;
      if (b > 255) b = 255;

      rpic->rgb.red  [i]   = r;
      rpic->rgb.green[i]   = g;
      rpic->rgb.blue [i]   = b;

      if (r != g || r != b) is_gray_map = SY_FALSE;
    }    
  }
  
  switch(image->type){

  case IBITMAP:
    ibm = NULL;
    if (image->rgb.size > 0){
      ibm = Sy_MakeImBitMask (image->width);
      for (y = 0; y < image->height; y++){
	for (x = 0; x < image->width; x++){
	  iptr  = y*ibm->linebytes + ibm->offset[x];
	  imval = image->data[iptr] & ( ibm->mask[x] ) ? 1 : 0;
	  if (image->rgb.red[imval] == 0){
	    image->data[iptr] = image->data[iptr] & ~( ibm->mask[x] );
	  }
	  else{
	    image->data[iptr] = image->data[iptr] | ( ibm->mask[x] ); 
	  }
	}
      }
    }
    
    switch(typ){
    case SY_ORIG:
    case SY_GREY_OR_BITMAP:
    case SY_BITMAP:
      rpic->type  = SY_BITMAP;
      rpic->data  = image->data; image->data = NULL;
      rpic->title = image->title; image->title = NULL;
      rpic->linebytes = ((rpic->width%8) == 0) ? (rpic->width >> 3) : 
	(rpic->width >> 3) + 1;
      break;
    case SY_GRAY:
    case SY_GREY:
      rpic->type  = SY_GREY;
      rpic->pixlen= 1;
      rpic->linebytes = rpic->width;
      rpic->title = image->title; image->title = NULL;
      rpic->data  = (unsigned char*) malloc (sizeof(unsigned char)*
					    image->width*image->height);

      if (ibm == NULL) ibm = Sy_MakeImBitMask (image->width);
      for (y = 0; y < image->height; y++){
	for (x = 0; x < image->width; x++){
	  iptr  = y*ibm->linebytes + ibm->offset[x];
	  imval = image->data[iptr] & ( ibm->mask[x] ) ? 1 : 0;
	  rpic->data[y*image->width + x] = imval ? 255 : 0;
	}
      }
      break;
    case SY_RGB:
    case SY_TRUECOLOR:
      fprintf (stderr, 
	       "Conversion from bitmap to RGB image not implemented.\n");
      Sy_FreeImage(rpic);
      rpic = NULL;
      break;
    } /* switch (typ) */
    Sy_FreeImBitMask(ibm);
    break;

  case IRGB:
    switch(typ){
    case SY_BITMAP:
      fprintf (stderr, 
	       "Conversion from RGB to bitmap image not implemented.\n");
      Sy_FreeImage(rpic);
      rpic = NULL;
      break;
    case SY_GRAY:
    case SY_GREY:
    case SY_GREY_OR_BITMAP:
      rpic->type  = SY_GREY;
      rpic->pixlen= 1;
      rpic->linebytes = rpic->width;
      rpic->title = image->title; image->title = NULL;
      rpic->data  = image->data; image->data = NULL;

      if (is_gray_map){
	for (i = 0; i < rpic->width*rpic->height; i++){
	  rpic->data[i] = rpic->rgb.red[rpic->data[i]];
	}
      } else {
	for (i = 0; i < rpic->width*rpic->height; i++){
	  rpic->data[i] = floor (
	    (float)(rpic->rgb.red[rpic->data  [i]] ) * 0.229 + 
	    (float)(rpic->rgb.green[rpic->data[i]] ) * 0.587 + 
	    (float)(rpic->rgb.blue[rpic->data [i]] ) * 0.114 + 0.5);
	}
      }
      break;
    case SY_ORIG:
    case SY_RGB:
      rpic->type  = SY_RGB;
      rpic->title = image->title; image->title = NULL;
      rpic->data  = image->data; image->data = NULL;
      break;
    case SY_TRUECOLOR:
      fprintf (stderr, 
	       "Conversion from RGB to truecolor image not implemented.\n");
      Sy_FreeImage(rpic);
      rpic = NULL;
      break;
    } /* switch (typ) */
    break;

  case ITRUE:
    fprintf (stderr, 
	     "Handling of truecolor image not yet implemented.\n");
    Sy_FreeImage(rpic);
    rpic = NULL;
    break;

  default:
    rpic = NULL;
  } /* switch (image->type) */

  Sy_FreeInternalImage(image);
  return rpic;
}
#else
SyImage_t *Sy_LoadImageFromFile (char *fname, SyImageType typ, 
				  char *formathint, int verbose)
{
  Pixrect        *pr;
  FILE           *inStream;
  colormap_t     colormap;
  SyImage_t      *rpic  = NULL;
  SyImBitMask    *ibm   = NULL;

  int i, x, y, linebytes, 
    imval, iptr, is_inpipe;

  colormap.type = RMT_NONE;

  is_inpipe  = *fname == '-';
  inStream   = is_inpipe ? stdin : fopen (fname,"r");

  if (inStream == NULL){
    Sy_SimpleError ("%s: error while opening input stream\n", 
		    "Sy_LoadImageFromFile");
    return NULL;
  }

  if ( (pr = pr_load (inStream, &colormap)) == 0 ){
    if (!is_inpipe) fclose (inStream); 
    return NULL;
  }
  else if (!is_inpipe) fclose (inStream);

  rpic = (SyImage_t*) malloc (sizeof(SyImage_t));
  rpic->rgb.size  = 0;
  rpic->rgb.red   = NULL;
  rpic->rgb.green = NULL;
  rpic->rgb.blue  = NULL;
  
  /* get image attributes */
  rpic->width      = pr->pr_size.x;
  rpic->height     = pr->pr_size.y;
  rpic->depth      = pr->pr_depth;
  rpic->linebytes  = mpr_linebytes (rpic->width, rpic->depth);
  rpic->data       = (char*) mpr_d (pr)->md_image;
  
#ifdef DEBUG
  if (verbose){
    printf ("height       = %d\n", rpic->height );
    printf ("width        = %d\n", rpic->width );
    printf ("depth        = %d\n", rpic->depth );
    printf ("linebytes    = %d\n", rpic->linebytes );
    printf ("rgb map:size = %d\n", rpic->rgb.size);
  }
#endif
  return rpic;
}
#endif

#ifdef HAS_GEN_LOADER
void Sy_FreeInternalImage(Image *im)
{
  if (im == NULL) return;

  if (im->title)     free(im->title); 
  if (im->data)      free(im->data);
  if (im->rgb.size > 0){
    if (im->rgb.red)   free(im->rgb.red); 
    if (im->rgb.green) free(im->rgb.green); 
    if (im->rgb.blue)  free(im->rgb.blue);
  }
  free(im);
}
#endif

void Sy_FreeImBitMask (SyImBitMask *ibm)
{
  if (ibm == NULL) return;
  if (ibm->offset) free(ibm->offset); 
  if (ibm->mask)   free(ibm->mask); 
  free(ibm);
}

SyImBitMask *Sy_MakeImBitMask (int width)
{
  int i;
  SyImBitMask *ibm;

  ibm = (SyImBitMask*) malloc (sizeof(SyImBitMask));
  ibm->offset = (int*) malloc (sizeof(int)*width);
  ibm->mask = (unsigned char*) malloc (sizeof(unsigned char)*width);

  ibm->linebytes = ((width%8) == 0) ? (width >> 3) : (width >> 3) + 1;
  for (i = 0; i < width; i++){
    ibm->offset[i] = i >> 3;
    ibm->mask[i]   = 128 >> (i%8);
  }
  return ibm;
}

void Sy_FreeImage(SyImage_t *im)
{
  if (im == NULL) return;

  if (im->title)     free(im->title); 
  if (im->data)      free(im->data);
  if (im->rgb.red)   free(im->rgb.red); 
  if (im->rgb.green) free(im->rgb.green); 
  if (im->rgb.blue)  free(im->rgb.blue);
  free (im);
}
