/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 *
 * all-purpose skeleton traverse
 *
 */

static char sy_SkelTraverseId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

#define SY_MEMINC 64

static int depth     = 0, 
           max_depth = 0, 
           nrec      = 0;

static int verbose_local;

static SyWalkBuff buff_0 = {0,0};

SyWalkState syWstate;

static SyWalktask* task;

void Sy_InitSkelTraverse (SySkelDAC *vsk, SyIpoDAC *ipd, int verbose, 
			  SyWalktask *wt)
{
  task                         = wt;
  syWstate.verbose             = verbose_local = verbose;
  syWstate.vsk                 = vsk;
  syWstate.ipd                 = ipd;
  syWstate.thres               = wt->thres;
  syWstate.group_thres         = wt->group_thres;
  syWstate.narr_thres          = wt->narr_thres;
  syWstate.search_thres        = wt->search_thres;
  syWstate.is_search_threshold = wt->is_search_threshold;
}  

/**********************************************************************
 * function 
 * void Sy_SkeletonTraverse ()
 * 
 * purpose: 
 *
 *
 * param: 
 *          
 *********************************************************************/

void Sy_SkeletonTraverse(SySkelDAC *vsk, SyIpoDAC *ipd, float thres, 
			 int verbose, SyWalktask *wt)
{
  void Sy_WalkSkeleton(SySkelDAC *vsk, SyIpoDAC *ipd, int *next_edge, 
		       int *current_head_vertex, int skel_id, float thres);
  void Sy_TraverseMessage(SySkelDAC *vsk, int current_head_vertex, 
			  int skel_id, int foreground);

  int i, skel_id = 0;
  int head_vertex, tail_vertex,
      head_edge, tail_edge,
      left_point, right_point;

  int next_edge, current_head_vertex;

  int head_flag, tail_flag, conn_flag, 
      tail_conn_flag, head_conn_flag,
      foreground;

  float search_thres;

  enum SyInitTravOper oper    = TRAV_NOP;
  enum SyTravDir      travdir = TRAV_VOID;

  vsk->label.e_ = Sy_AllocLabel (vsk->edim);
  vsk->label.v_ = Sy_AllocLabel (vsk->vdim);

#ifdef DEBUG
  sprintf (syStrBuf, "Skeleton Traversal: %s ", sy_SkelTraverseId);
  Sy_Message (SY_FMTNL, syStrBuf);
#endif

  task                  = wt;
  syWstate.verbose      = verbose_local = verbose;
  syWstate.vsk          = vsk;
  syWstate.ipd          = ipd;
  syWstate.thres        = thres;
  syWstate.group_thres  = wt->group_thres;
  
  syWstate.is_connected  = 0;
  syWstate.is_border     = vsk->is_border;
  syWstate.is_eresid     = vsk->is_eresid;

  search_thres = (wt->is_search_threshold) ? 
    syWstate.search_thres : thres;

  /* INIT call */
  if (task->user_function[(int)INIT] != NULL) 
    task->user_function[(int)INIT]();

  if (wt->is_loc_preset != SY_TRUE){

    for (i = 1, skel_id = 0; i <= vsk->edim; i++){
      
      if (vsk->label.e_[i] == 0){
	
	head_vertex = vsk->edge[i].head_vertex;
	tail_vertex = vsk->edge[i].tail_vertex; 
	tail_edge   = vsk->edge[i].tail_edge;
	head_edge   = vsk->edge[i].head_edge;
	left_point  = vsk->edge[i].left_point;
	right_point = vsk->edge[i].right_point;
	
	head_flag   = (vsk->regul[head_vertex].birth[2] > search_thres);
	tail_flag   = (vsk->regul[tail_vertex].birth[2] > search_thres);
	
	conn_flag   = Sy_IpoConnected (ipd, left_point, right_point);
	
	head_conn_flag = (head_edge != 0) ? 
	  !Sy_IsContinuation (ipd, vsk, i, TRAV_HEAD, wt): 0;
	
	tail_conn_flag = (tail_edge != 0) ? 
	  !Sy_IsContinuation (ipd, vsk, i, TRAV_TAIL, wt): 0;
	
	oper        = TRAV_NOP;
	travdir     = TRAV_VOID;
	
	switch (vsk->edge[i].rank){
	case 0:
	  if (tail_flag && head_flag){
	    travdir = TRAV_TAIL;
	    oper    = TRAV_IN_EDGE;
	  }
	  break;
	case 1:
	  if (tail_flag || head_flag){
	    if (tail_flag && (vsk->label.v_[tail_vertex] == 0) && 
		!tail_conn_flag){
	      travdir = TRAV_TAIL;
	      oper    = TRAV_IN_EDGE;
	      break;
	    }
	  }
	  break;
	case 2:
	  if (tail_flag ^ head_flag){
	    if (tail_flag && (vsk->label.v_[tail_vertex] == 0) && 
		!tail_conn_flag){
	      travdir = TRAV_TAIL;
	      oper    = TRAV_IN_EDGE;
	      break;
	    }
	    else if (head_flag && (vsk->label.v_[head_vertex] == 0) &&
		     !head_conn_flag){
	      travdir = TRAV_HEAD;
	      oper    = TRAV_IN_EDGE;
	      break;
	    }
	    break;
	  }
	  
	  if (tail_flag && head_flag){
	    if (conn_flag && (tail_edge != 0) && (head_edge != 0)){
	      if ((vsk->label.v_[tail_vertex] == 0) && !tail_conn_flag){
		travdir = TRAV_TAIL;
		oper    = TRAV_IN_EDGE;
		break;
	      } else if ((vsk->label.v_[head_vertex] == 0) && !head_conn_flag){
		travdir = TRAV_HEAD;
		oper    = TRAV_IN_EDGE;
		break;
	      }
	      break;
	    }
	    else if ((tail_edge == 0) || (head_edge == 0)){
	      travdir = (tail_edge == 0)? TRAV_HEAD : TRAV_TAIL;
	      if (travdir == TRAV_HEAD){
		oper    = (vsk->label.v_[head_vertex] == 0) ? 
		  TRAV_OUT_EDGE : TRAV_NOP;}
	      else{
		oper    = (vsk->label.v_[tail_vertex] == 0) ? 
		  TRAV_OUT_EDGE : TRAV_NOP;
	      }

	      break;
	    }
	  }
	} /* switch */
	
	if ((travdir == TRAV_TAIL) && tail_conn_flag) oper = TRAV_NOP;
	if ((travdir == TRAV_HEAD) && head_conn_flag) oper = TRAV_NOP;
	
	if (oper == TRAV_NOP) continue;
	
	switch (travdir){
	case TRAV_HEAD:
	  next_edge           = (oper == TRAV_IN_EDGE)? head_edge : i;
	  current_head_vertex = head_vertex;
	  break;
	case TRAV_TAIL:
	  next_edge           = (oper == TRAV_IN_EDGE)? tail_edge : i;
	  current_head_vertex = tail_vertex;
	  break;
	}
	
	skel_id++;
	vsk->label.e_[i]    = (oper == TRAV_IN_EDGE)? SY_BORDERMARK: skel_id;
	
	max_depth            = 0;
	syWstate.id                = skel_id;
	syWstate.start.head_vertex = current_head_vertex;
	syWstate.start.this_edge   = (oper == TRAV_IN_EDGE)? i : 0;
	syWstate.start.next_edge   = next_edge;
	syWstate.current           = buff_0;
	syWstate.foreground        = foreground = 
	  Sy_SkeletonType (vsk, ipd, current_head_vertex, next_edge);
	
	
	
#ifdef DEBUG
	sprintf (syStrBuf, "t.e %d, n.e %d, v %d ",
		 syWstate.start.this_edge, next_edge, current_head_vertex);
	Sy_Message (SY_FMTNL, syStrBuf);
	
	if (oper == TRAV_IN_EDGE)
	  Sy_Message (SY_FMTNL, "trav in edge ");
	else
	  Sy_Message (SY_FMTNL, "trav out edge ");
#endif
	
	Sy_TraverseMessage (vsk, current_head_vertex, skel_id, foreground);
	
	/* TRAVerse call */	  
	if (task->user_function[(int)TRAV] != NULL) task->user_function[(int)TRAV]();
	
	Sy_WalkSkeleton (vsk, ipd, &next_edge, &current_head_vertex, skel_id, thres);
	
	/* TERMination call */	  
	if (task->user_function[(int)TERM] != NULL) task->user_function[(int)TERM]();
	
      }
    }/* for */ 
  }
  else{
	skel_id++;
	vsk->label.e_[wt->start_edge] = SY_BORDERMARK;
	
	current_head_vertex        = wt->start_vertex;
	next_edge                  = wt->start_edge;

	max_depth                  = 0;
	syWstate.id                = skel_id;
	syWstate.start.head_vertex = wt->start_vertex;
	syWstate.start.this_edge   = wt->start_edge;
	syWstate.start.next_edge   = wt->start_edge;
	syWstate.current           = buff_0;
	syWstate.foreground        = foreground = 
	  Sy_SkeletonType (vsk, ipd, current_head_vertex, next_edge);
	
	Sy_TraverseMessage (vsk, current_head_vertex, skel_id, foreground);	
	/* TRAVerse call */	  
	if (task->user_function[(int)TRAV] != NULL) task->user_function[(int)TRAV]();
   
	Sy_WalkSkeleton (vsk, ipd, &next_edge, &current_head_vertex, skel_id, thres);
	
	/* TERMination call */	  
	if (task->user_function[(int)TERM] != NULL) task->user_function[(int)TERM](); 
  } /* else : explicit control of starting point */
  /* exit call */
  if (task->user_function[(int)FEXIT] != NULL) task->user_function[(int)FEXIT]();
}
   
void Sy_WalkSkeleton(SySkelDAC *vsk, SyIpoDAC *ipd, int *next_edge, 
		     int *current_head_vertex, int skel_id, float thres)
{
  void Sy_WalkSkeleton(SySkelDAC *vsk, SyIpoDAC *ipd, int *next_edge, 
		       int *current_head_vertex, int skel_id, float thres);

  int current_head_vertex_buff, 
      next_edge_buff, ref_edge;

  int head_flag, tail_flag, 
      conn_flag, conn[4], walking_flag = 1;

  int tail_vertex, head_vertex, 
      right_point, left_point;
  
  int is_border;

  ref_edge = *next_edge;

  /* WALK call */
  if (task->user_function[(int)WALK] != NULL){
    syWstate.last = syWstate.current;
    syWstate.current.head_vertex = *current_head_vertex;
    syWstate.current.next_edge   = *next_edge;
    task->user_function[(int)WALK]();
  }

  if (vsk->label.v_[*current_head_vertex] == 0){
    
    current_head_vertex_buff            = *current_head_vertex;
    next_edge_buff                      = *next_edge;
    vsk->label.v_[*current_head_vertex] = skel_id;
    
    while ((next_edge_buff != 0) && walking_flag){

      syWstate.current.head_vertex = current_head_vertex_buff;
      syWstate.current.next_edge   = next_edge_buff;

      head_vertex = vsk->edge[next_edge_buff].head_vertex;
      tail_vertex = vsk->edge[next_edge_buff].tail_vertex;
      left_point  = vsk->edge[next_edge_buff].left_point;
      right_point = vsk->edge[next_edge_buff].right_point;
  
      head_flag   = (vsk->regul[head_vertex].birth[2] > thres);
      tail_flag   = (vsk->regul[tail_vertex].birth[2] > thres);
    
      conn[0]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn) &&
		  (ipd->ipl[left_point].tail_icn != 0));

      conn[1]  = ((ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn) &&
		  (ipd->ipl[left_point].tail_icn != 0));
      
      conn[2]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn) &&
		  (ipd->ipl[left_point].head_icn != 0));

      conn[3]  = ((ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn) &&
		  (ipd->ipl[left_point].head_icn != 0));

      conn_flag = (conn[0] || conn[1] || conn[2] || conn[3]);

      is_border  = (vsk->edge[next_edge_buff].head_edge == 0) 
	        || (vsk->edge[next_edge_buff].tail_edge == 0);

      if ((head_flag && tail_flag) && (!conn_flag || is_border)){
	
	*current_head_vertex  = current_head_vertex_buff;
	*next_edge            = next_edge_buff;
	
	head_vertex = vsk->edge[*next_edge].head_vertex;
	tail_vertex = vsk->edge[*next_edge].tail_vertex;
	
	vsk->label.e_[*next_edge] = skel_id;
	
	if (*current_head_vertex == tail_vertex){
	  *current_head_vertex = head_vertex;
	  *next_edge           = vsk->edge[*next_edge].head_edge;
	}
	else{
	  *current_head_vertex = tail_vertex;
	  *next_edge           = vsk->edge[*next_edge].tail_edge;
	}

	if (++depth > max_depth) max_depth = depth; nrec++;
#ifdef DEBUG
	if (syIsVerbose){
	  sprintf (syStrBuf, "> depth:%6d, max.depth:%6d, nrec:%6d\r",
		   depth, max_depth, nrec);
	  Sy_Message (SY_FMT, syStrBuf);
	}
#endif
	Sy_WalkSkeleton (vsk, ipd, next_edge, current_head_vertex, 
			 skel_id, thres);
	
	syWstate.current.head_vertex = current_head_vertex_buff;
        syWstate.current.next_edge   = next_edge_buff;

	/* ROT call */
	if (task->user_function[(int)ROT] != NULL) 
	  task->user_function[(int)ROT]();
	
	depth--;
#ifdef DEBUG
	if (syIsVerbose){
	  sprintf (syStrBuf, "> depth:%6d\r", depth); 
	  Sy_Message (SY_FMT, syStrBuf);
	}
#endif
      }
      else{
	/* EOSK call */
	syWstate.is_connected = conn_flag;
	syWstate.is_border    = is_border;
	if (task->user_function[(int)EOSK] != NULL) 
	  task->user_function[(int)EOSK]();
      }

      do {
	if (current_head_vertex_buff == vsk->edge[next_edge_buff].tail_vertex){
	  next_edge_buff = vsk->edge[next_edge_buff].tail_edge;
	}
	else{
	  next_edge_buff = vsk->edge[next_edge_buff].head_edge;
	}
	walking_flag = (next_edge_buff != 0) ? 
	  (next_edge_buff != ref_edge): SY_FALSE;       
      } while (walking_flag && vsk->label.e_[next_edge_buff] != 0);

    }
  }
}


void Sy_SkelRotate (SySkelDAC *vsk, SyIpoDAC *ipd, int *next_edge, 
		    int *current_head_vertex, float thres)
{

  int current_head_vertex_buff, 
      next_edge_buff, ref_edge;

  int head_flag, tail_flag, 
      conn_flag, walking_flag = 1;

  int tail_vertex, head_vertex, 
      right_point, left_point;

  syWstate.start.head_vertex = *current_head_vertex;
  syWstate.start.next_edge   = *next_edge;
  syWstate.current           = syWstate.start;

  ref_edge             = *next_edge;
    
  current_head_vertex_buff  = *current_head_vertex;
  next_edge_buff            = *next_edge;

  if (task->user_function[(int)TRAV] != NULL) task->user_function[(int)TRAV]();

  while ((next_edge_buff != 0) && walking_flag){

    head_vertex = vsk->edge[next_edge_buff].head_vertex;
    tail_vertex = vsk->edge[next_edge_buff].tail_vertex;
    left_point  = vsk->edge[next_edge_buff].left_point;
    right_point = vsk->edge[next_edge_buff].right_point;
  
    head_flag   = (vsk->regul[head_vertex].birth[2] > thres);
    tail_flag   = (vsk->regul[tail_vertex].birth[2] > thres);
    
    conn_flag =
      ((   ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn)
       || (ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn)
       || (ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn)
       || (ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn));
    
    if ((head_flag && tail_flag) && !conn_flag){
	
      syWstate.current.next_edge  = next_edge_buff;
      syWstate.current.head_vertex = current_head_vertex_buff;

      /* ROT call */
      if (task->user_function[(int)ROT] != NULL) 
	task->user_function[(int)ROT]();
	
    }

    if (current_head_vertex_buff == vsk->edge[next_edge_buff].tail_vertex){
      next_edge_buff = vsk->edge[next_edge_buff].tail_edge;
    }
    else{
      next_edge_buff = vsk->edge[next_edge_buff].head_edge;
    }
    if (next_edge_buff != 0){
      walking_flag = (next_edge_buff != ref_edge);
    }
  }

  syWstate.current.head_vertex = current_head_vertex_buff;
  syWstate.current.next_edge  = next_edge_buff;
  if (task->user_function[(int)TERM] != NULL) task->user_function[(int)TERM]();
}


void Sy_SimplePolyScan (SySkelDAC *vsk, SyIpoDAC *ipd, int *next_edge, 
			int *current_head_vertex)
{

  int current_head_vertex_buff, 
      next_edge_buff, ref_edge;

  int walking_flag = 1;

  int tail_vertex, head_vertex;

  syWstate.start.head_vertex = *current_head_vertex;
  syWstate.start.next_edge   = *next_edge;
  syWstate.last              = syWstate.start;
  syWstate.current           = syWstate.start;

  ref_edge             = *next_edge;
    
  current_head_vertex_buff  = *current_head_vertex;
  next_edge_buff            = *next_edge;

  while ((next_edge_buff != 0) && walking_flag){

    /* SCAN call */
    if (task->user_function[(int)SCAN] != NULL) 
      task->user_function[(int)SCAN]();
	
    if (current_head_vertex_buff == vsk->edge[next_edge_buff].tail_vertex){
      current_head_vertex_buff = vsk->edge[next_edge_buff].head_vertex;
      next_edge_buff           = vsk->edge[next_edge_buff].head_edge;
    }
    else{
      current_head_vertex_buff = vsk->edge[next_edge_buff].tail_vertex;
      next_edge_buff           = vsk->edge[next_edge_buff].tail_edge;
    }

    syWstate.last = syWstate.current;
    syWstate.current.head_vertex = current_head_vertex_buff;
    syWstate.current.this_edge   = syWstate.last.next_edge;
    syWstate.current.next_edge   = next_edge_buff;

    if (next_edge_buff != 0){
      walking_flag = (next_edge_buff != ref_edge);
    }
  }
  if (task->user_function[(int)TERM] != NULL) 
    task->user_function[(int)TERM]();
}

void Sy_TraverseMessage (SySkelDAC *vsk, int current_head_vertex, 
			 int skel_id, int foreground)
{
  SyVertex vtmp;
  float hotx, hoty;

  vtmp = vsk->vertex[current_head_vertex];
  hotx = vtmp.xden/vtmp.num; 
  hoty = vtmp.yden/vtmp.num;
  
  sprintf(syStrBuf, "Traversing skeleton(%s) %3d at (%7.1f,%7.1f) ", 
	  syWstate.foreground ? "foreground" : "background", 
	  skel_id, hotx, hoty);
  Sy_Message (SY_FMTNL, syStrBuf);
}


SyBoolean Sy_IsContinuation (SyIpoDAC *ipd, SySkelDAC *vsk, 
			     SyIndex edge_index, enum SyTravDir dir,
			     SyWalktask *wt)
{
  SyIndex next_edge, ref_edge, ref_vertex,
          head_vertex, tail_vertex;
  SyBoolean tail_flag, head_flag;
  float thres;

  next_edge = (dir == TRAV_HEAD) ? vsk->edge[edge_index].head_edge :
                                   vsk->edge[edge_index].tail_edge;

  if (next_edge == 0) return (SY_FALSE);
  
  thres = (wt->is_search_threshold) ? wt->search_thres : wt->thres;

  ref_vertex = (dir == TRAV_HEAD)? vsk->edge[edge_index].head_vertex :
                                   vsk->edge[edge_index].tail_vertex;
  ref_edge   = edge_index;
  
  while(next_edge != ref_edge){

    head_vertex = vsk->edge[next_edge].head_vertex;
    tail_vertex = vsk->edge[next_edge].tail_vertex;     
    head_flag   = (vsk->regul[head_vertex].birth[2] > thres);
    tail_flag   = (vsk->regul[tail_vertex].birth[2] > thres);

    if ( tail_flag && head_flag ){
      
      if (!Sy_IpoConnected(ipd, vsk->edge[next_edge].left_point,
			   vsk->edge[next_edge].right_point) ){
	return(SY_TRUE);
      }
    }
    if (ref_vertex == vsk->edge[next_edge].head_vertex){
      next_edge = vsk->edge[next_edge].head_edge;}
    else{
      next_edge = vsk->edge[next_edge].tail_edge;}
    
  } /* while */
      
  return(SY_FALSE);
}


/***************************** new **********************************/

typedef struct _SyVertex2NodeInfo {
  int current_node, last_edge, last_branch;
} SyVertex2NodeInfo;


static SyWalktask wt;
static int *e_label = NULL;
static int *v_label = NULL;
static SyVertex2NodeInfo *v2node  = NULL;
static union SyGenericPoint *genpoint = NULL;
static SyVSkeletonRef *refbuf;
static SyVSkeletonBndryOffset *lb_buf, *rb_buf;
static float *radiusbuf;

static int bufdim = 0;

static SyBoolean is_true_node = SY_FALSE, is_disjoint = SY_FALSE,
                 is_terminating = SY_TRUE;

static double len_tot, sum_rad, sum_rad2, sum_strength, sum_strength2,
              max_rad, min_rad, max_strength, min_strength, strength0, 
              strength1;

static int branch_hierarchy = 0;

static SyControl *local_ctrl;


static void Sy_ScanTerm(void);

void Sy_InitVSkeleton (SyVSkeleton* vs)
{
  vs->id = vs->user_id = 0;
  vs->branchdim = vs->nodedim = 0;
  vs->maxbranchdim = vs->maxnodedim = 0;
  vs->is_foreground = 0;
  vs->branch = NULL;
  vs->node   = NULL;
}

void Sy_InitVSkeletonBranch (SyVSkeletonBranch *b)
{
  b->etype = SY_NILBRANCH;
  b->id = b->user_id = 0;
  b->hierarchy_level = 0;
  b->is_foreground = SY_FALSE;
  b->strength = b->max_strength = b->min_strength = b->mean_strength
    = b->var_strength = 0.0;
  b->tail_branch = b->head_branch = b->tail_node = b->head_node = 0;
  b->elementdim = b->maxelementdim = 0;

  b->axis        = NULL;
  b->ref         = NULL;
  b->radius      = NULL;
  b->left_bndry  = NULL;
  b->right_bndry = NULL;
}

void Sy_InitVSkeletonNode (SyVSkeletonNode *n)
{
  n->etype = SY_NILBRANCH;
  n->ntype = SY_NILNODE;
  n->id = n->user_id = 0;
  n->hierarchy_level = 0;
  n->is_foreground = SY_FALSE;
  n->strength = 0.0;
  n->rank = 0;
  n->maxrank = 0;
  n->inbranch = NULL;
}

void Sy_PolyScan (SySkelDAC *vsk, SyIpoDAC *ipd, int *next_edge, 
		  int *current_head_vertex, SyControl *ctrl)
{
  int current_head_vertex_buff, step_count = 0,
      next_edge_buff, ref_edge, last_edge = 0, rot_ref_edge;

  int head_flag, tail_flag, conn_flag, walking_flag = 1, is_skeleton, 
      rotate_flag = 0, endpoint_flag, node_flag, rot_count;

  int tail_vertex, head_vertex, 
      right_point, left_point;

  float thres;
  
  thres    = ctrl->threshold;
  ref_edge = *next_edge;

  syWstate.start.head_vertex = *current_head_vertex;
  syWstate.start.next_edge   = *next_edge;
  syWstate.last              = syWstate.start;
  syWstate.current           = syWstate.start;
    
  current_head_vertex_buff  = *current_head_vertex;
  next_edge_buff            = *next_edge;

  syWstate.is_endpoint = 
       (vsk->regul[current_head_vertex_buff].birth[2] > thres)
    && (vsk->regul[current_head_vertex_buff].birth[1] <= thres);

  syWstate.is_node = (vsk->regul[current_head_vertex_buff].birth[0] > thres);

  if (task->user_function[(int)TRAV] != NULL) 
    task->user_function[(int)TRAV]();

  syWstate.is_terminating = SY_FALSE;

  while ((next_edge_buff != 0) && walking_flag){

    rotate_flag = 1; rot_ref_edge = last_edge == 0 ? next_edge_buff:last_edge; 
    rot_count   = 0;

    do {       
      head_vertex = vsk->edge[next_edge_buff].head_vertex;
      tail_vertex = vsk->edge[next_edge_buff].tail_vertex;
      left_point  = vsk->edge[next_edge_buff].left_point;
      right_point = vsk->edge[next_edge_buff].right_point;
      
      head_flag   = (vsk->regul[head_vertex].birth[2] > thres);
      tail_flag   = (vsk->regul[tail_vertex].birth[2] > thres);
      
      conn_flag =
	((   ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].tail_icn)
	 || (ipd->ipl[left_point].tail_icn == ipd->ipl[right_point].head_icn)
	 || (ipd->ipl[left_point].head_icn == ipd->ipl[right_point].tail_icn)
	 || (ipd->ipl[left_point].head_icn == ipd->ipl[right_point].head_icn));
      
      is_skeleton = ((head_flag && tail_flag) && !conn_flag); 

      if (is_skeleton) {rotate_flag = SY_FALSE; break;}
      else {
	rot_count++;
	if (current_head_vertex_buff == vsk->edge[next_edge_buff].tail_vertex){
	  next_edge_buff = vsk->edge[next_edge_buff].tail_edge;
	}
	else{
	  next_edge_buff = vsk->edge[next_edge_buff].head_edge;
	}

	rotate_flag = next_edge_buff != 0 && next_edge_buff != rot_ref_edge;
      }
    } while (rotate_flag);
 
    if (rot_count > 0){
      syWstate.current.head_vertex = current_head_vertex_buff;
      syWstate.current.next_edge   = next_edge_buff;
      syWstate.current.this_edge   = syWstate.last.next_edge;
    }

    if (!is_skeleton){ /* no continuation here -> endpoint, e.g. in
                          exoskeletons at the terminating ray... */
      walking_flag = SY_FALSE;
      syWstate.is_endpoint = SY_TRUE;
      break;
    }

    syWstate.is_loop = (next_edge_buff == ref_edge) && (step_count > 0);
    if (syWstate.is_loop){
      walking_flag = SY_FALSE;
      break;
    }

    if (task->user_function[(int)SCAN] != NULL) 
      task->user_function[(int)SCAN]();
    
    last_edge = next_edge_buff;
    syWstate.last = syWstate.current;

    if (current_head_vertex_buff == vsk->edge[next_edge_buff].tail_vertex){
      current_head_vertex_buff = vsk->edge[next_edge_buff].head_vertex;
      next_edge_buff           = vsk->edge[next_edge_buff].head_edge;
    }
    else{
      current_head_vertex_buff = vsk->edge[next_edge_buff].tail_vertex;
      next_edge_buff           = vsk->edge[next_edge_buff].tail_edge;
    }
    step_count++;

    syWstate.current.head_vertex = current_head_vertex_buff;
    syWstate.current.next_edge   = next_edge_buff;
    syWstate.current.this_edge   = syWstate.last.next_edge;

    syWstate.is_endpoint = 
         (vsk->regul[current_head_vertex_buff].birth[2] > thres)
      && (vsk->regul[current_head_vertex_buff].birth[1] <= thres);

    syWstate.is_node = (vsk->regul[current_head_vertex_buff].birth[0] > thres);
    
    if (next_edge_buff != 0){
      walking_flag = 
	ctrl->stop_at_nodes ? (!syWstate.is_endpoint &&
			       !syWstate.is_node &&
			       !syWstate.is_loop) : !syWstate.is_loop;
    }
  } /* while */

  syWstate.is_terminating = SY_TRUE;
  if (task->user_function[(int)SCAN] != NULL) 
    task->user_function[(int)SCAN]();

  syWstate.is_invalid = !is_skeleton && !syWstate.is_endpoint 
                        && !syWstate.is_node;
  if (syWstate.is_invalid){
    fprintf (stderr, "Inconsistency in skeleton data detected...\n");
  }
	     
  if (task->user_function[(int)TERM] != NULL) 
    task->user_function[(int)TERM]();
}

static void Sy_LabelTrav(void)
{
  fprintf (stderr, "Traversing skeleton(%s) no %d\n", 
	   syWstate.foreground ? "foreground" : "background", 
	   syWstate.id);
}

static void Sy_ScanTrav(void)
{
  len_tot = sum_rad = sum_rad2 = sum_strength = sum_strength2 = 0.0;
  max_rad = min_rad = max_strength = min_strength = 0.0;
}

static void Sy_CompileTrav(void)
{

  int next_edge, skel_id, chv;
  
  next_edge = syWstate.current.next_edge;
  chv       = syWstate.current.head_vertex;
  skel_id    = syWstate.id;

  if (syIsVerbose){
    fprintf (stderr, "in SK=%d (%s): ", skel_id, syWstate.foreground ?
	     "fg":"bg");
 
    if (v2node[chv].current_node != 0){
      fprintf(stderr, "node(%s)=%d ->branches ", 
	      is_true_node ? syWstate.is_node ? "X" : "T" : "L", 
	      v2node[chv].current_node);
    }
  }
}

static void Sy_CompileTerm(void)
{
  
  SyVSkeletonSet  *vss;
  SyVSkeleton     *curskel;
  SyVSkeletonNode *cn;
 
  int last_edge, last_branch, next_edge, current_node, 
      current_branch, skel_id, chv;

  vss       = syWstate.vss; 
  next_edge = syWstate.current.next_edge;
  chv       = syWstate.current.head_vertex;
  skel_id    = syWstate.id;

  /* locate skeleton within skeleton set*/

  curskel        = &vss->skel[skel_id];
  last_branch    = v2node[chv].last_branch;
  current_node   = v2node[chv].current_node;
  current_branch = e_label[next_edge];

  if (last_branch != 0){
    if (curskel->branch[last_branch].tail_branch == 0){
      curskel->branch[last_branch].tail_branch = current_branch;
    }
    else{
      curskel->branch[last_branch].head_branch = current_branch;
    }
  }
  v2node[chv].last_branch  = current_branch;

  if (syWstate.is_loop && syIsVerbose){
    fprintf(stderr, "node(L)=%d ", current_node);
  }

  if (syIsVerbose){
    fprintf (stderr, "\n");
  }
}

static void Sy_CompileRot(void)
{
  SyVSkeletonSet    *vss;
  SyVSkeleton       *curskel;
  SyVSkeletonNode   *cn;
  SyVSkeletonBranch *cbr;
  SyRegul           *regul;
 
  int last_edge, last_branch, next_edge, next_edge_buf, current_node, 
      current_branch, skel_id, chv, current_head_vertex, tmp_hierarchy,
      i, old_dim;

  double xden, yden, num;

  vss       = syWstate.vss; 
  next_edge = syWstate.current.next_edge;
  chv       = syWstate.current.head_vertex;
  skel_id   = syWstate.id;
  regul     = local_ctrl->is_separate_saliency ? local_ctrl->regul : 
                                                 syWstate.vsk->regul;

  /* locate skeleton within skeleton set, allocate new
     space for skeleton if necessary */

  if (skel_id > vss->maxdim){
    old_dim      = vss->maxdim;
    vss->maxdim += SY_MEMINC;
    vss->dim     = skel_id;
    vss->skel    = (SyVSkeleton*) Sy_Realloc((char*)vss->skel, 
					     sizeof(SyVSkeleton)*(vss->maxdim+1));

    for (i = old_dim+1; i <= vss->maxdim; i++){
      Sy_InitVSkeleton(vss->skel+i);
      vss->skel[i].maxnodedim = SY_MEMINC;
      vss->skel[i].nodedim    = 0;
      vss->skel[i].node       = (SyVSkeletonNode*) 
	Sy_Malloc(sizeof(SyVSkeletonNode)*(vss->skel[i].maxnodedim+1));
      vss->skel[i].maxbranchdim = SY_MEMINC;
      vss->skel[i].branchdim    = 0;
      vss->skel[i].branch       = (SyVSkeletonBranch*) 
	Sy_Malloc(sizeof(SyVSkeletonBranch)*(vss->skel[i].maxnodedim+1));
    }
  }

  if (skel_id > vss->dim) vss->dim = skel_id;

  curskel                = &vss->skel[skel_id];
  curskel->id            = skel_id;
  curskel->is_foreground = syWstate.foreground;

  /* locate current node within current skeleton, 
     allocate new node if necessary */

  if (v2node[chv].current_node == 0){ /* new node */
    if (curskel->nodedim >= curskel->maxnodedim){
      curskel->maxnodedim += SY_MEMINC;
      curskel->node = (SyVSkeletonNode*) 
	Sy_Realloc ((char*)curskel->node, sizeof(SyVSkeletonNode)*
		    (curskel->maxnodedim+1));

      for (i = curskel->nodedim+1; i <= curskel->maxnodedim; i++){
	Sy_InitVSkeletonNode(curskel->node+i);
      }

    }
    curskel->nodedim++;
    v_label[chv] = skel_id;
    current_node = v2node[chv].current_node = curskel->nodedim;
    cn           = &curskel->node[current_node];
    Sy_InitVSkeletonNode(cn);

    cn->id       = current_node;
    cn->user_id  = 0;
    cn->rank     = 0;
    cn->maxrank  = 3;
    cn->inbranch = (int*) Sy_Malloc(sizeof(int)*(cn->maxrank+1));
    cn->ntype    = syWstate.is_node ? SY_INTERNAL : SY_END;
    cn->etype    = local_ctrl->branch_el_type;

    xden = syWstate.vsk->vertex[chv].xden;
    yden = syWstate.vsk->vertex[chv].yden;
    num  = syWstate.vsk->vertex[chv].num;
    
    switch (local_ctrl->branch_el_type){
    case SY_NILBRANCH:
    case SY_REAL:
      cn->loc.rp.x = xden/num; 
      cn->loc.rp.y = yden/num;
      break;
    case SY_INTEGER:
      cn->loc.np.x = floor(xden/num + 0.5);
      cn->loc.np.y = floor(yden/num + 0.5);
      break;
    case SY_RATIONAL:
      cn->loc.qp.xden = xden;
      cn->loc.qp.yden = yden;
      cn->loc.qp.xnum = cn->loc.qp.ynum = num;
      break;
    } /* switch */

    cn->radius          = syWstate.vsk->regul[chv].radius;
    cn->strength        = regul[chv].birth[0];
    cn->hierarchy_level = syWstate.vsk->is_hierarchy ?
                          syWstate.vsk->hier[next_edge] : 0;
    cn->is_foreground   = syWstate.foreground;

    if (syIsVerbose){
      fprintf(stderr, "node(%s)=%d ->branches ", 
	      is_true_node ? syWstate.is_node ? "X" : "T" : "L", current_node);
    }
  }
  else{ /* node was visited previously */
    current_node = v2node[chv].current_node;
    cn           = &curskel->node[current_node];

    if (cn->rank >= cn->maxrank){
      cn->maxrank += 3;
      cn->inbranch = (int*) Sy_Realloc((char*)cn->inbranch, 
				       sizeof(int)*(cn->maxrank+1));
    }

    /* skeleton pyramid information */
    if (syWstate.vsk->is_hierarchy){
      if ((tmp_hierarchy = syWstate.vsk->hier[next_edge]) > 0){
	if (cn->hierarchy_level == 0){
	  cn->hierarchy_level = tmp_hierarchy;}
	else {
	  if (cn->hierarchy_level > tmp_hierarchy) 
	    cn->hierarchy_level = tmp_hierarchy;
	}
      }
    } /* if is hierarchy */
  }
  
  last_branch = v2node[chv].last_branch;
  
  /* locate branch within skeleton, allocate branch structure
     if necessary */

  if (e_label[next_edge] == 0){ /* create new branch */
    if (curskel->branchdim >= curskel->maxbranchdim){
      curskel->maxbranchdim += SY_MEMINC;
      curskel->branch = (SyVSkeletonBranch*) 
	Sy_Realloc((char*)curskel->branch, sizeof(SyVSkeletonBranch)
		   *(curskel->maxbranchdim+1));

      for (i = curskel->branchdim+1; i <= curskel->maxbranchdim; i++){
	Sy_InitVSkeletonBranch(curskel->branch+i);
      }

    }
    
    curskel->branchdim++;
    current_branch     = curskel->branchdim;
    e_label[next_edge] = current_branch;
    cbr                = &curskel->branch[current_branch];

    Sy_InitVSkeletonBranch(cbr);

    cbr->tail_node       = current_node;
    cbr->tail_branch     = 0;
    cbr->id              = current_branch;
    cbr->user_id         = 0;
    cbr->skel_id         = skel_id;
    cbr->hierarchy_level = syWstate.vsk->is_hierarchy ? 
                           syWstate.vsk->hier[next_edge] : 0;
    cbr->is_foreground   = syWstate.foreground;
    cbr->has_boundary    = SY_TRUE;

    /* scan branch */
    current_head_vertex       = chv;
    next_edge_buf             = next_edge;
    local_ctrl->stop_at_nodes = SY_TRUE;

    wt.user_function[(int)TRAV] = (void(*)())Sy_ScanTrav;
    wt.user_function[(int)TERM] = (void(*)())Sy_ScanTerm; 
    
    syWstate.branch    = cbr;
    syWstate.branch_id = cbr->id = current_branch;

    if (syIsVerbose){
      fprintf(stderr, "B:%d ", current_branch);
    }

    is_disjoint = SY_FALSE;
    Sy_PolyScan (syWstate.vsk, syWstate.ipd, &next_edge_buf, 
		 &current_head_vertex, local_ctrl);

    cn          = &curskel->node[current_node];
    is_disjoint = SY_FALSE;

    wt.user_function[(int)TRAV] = (void(*)())Sy_CompileTrav;
    wt.user_function[(int)TERM] = (void(*)())Sy_CompileTerm; 
    /*******************************************************/

  }
  else { /* branch was already visited before */
    current_branch = e_label[next_edge];
    cbr = &curskel->branch[current_branch];
    cbr->head_node = current_node;    
    if (syIsVerbose){
      fprintf(stderr, "[B:%d] ", current_branch);
    }
  }

  cn->rank++;
  cn->inbranch[cn->rank] = current_branch; 

  if (syWstate.vsk->is_hierarchy){
    if (cn->hierarchy_level == 0){
      cn->hierarchy_level = cbr->hierarchy_level;
    }
    else if (cbr->hierarchy_level > 0 
	     && cn->hierarchy_level > cbr->hierarchy_level){
      cn->hierarchy_level = cbr->hierarchy_level;
    }
  }

  /* connect branch visited last to current branch */
  if (last_branch != 0){
    if (curskel->branch[last_branch].tail_branch == 0){
      curskel->branch[last_branch].tail_branch = current_branch;
    }
    else{
      curskel->branch[last_branch].head_branch = current_branch;
    }
  }
  
  v2node[chv].last_branch  = current_branch;
}

static void Sy_ScanTerm(void)
{
  SyVSkeletonSet    *vss;
  SyVSkeleton       *curskel;
  SyVSkeletonNode   *cn;
  SyVSkeletonBranch *cbr;
  SyRegul           *regul;

  int i;
  int last_edge, current_node, 
      current_branch, skel_id, chv;
  double xden, yden, num;

  vss        = syWstate.vss; 
  last_edge  = syWstate.current.this_edge;
  skel_id    = syWstate.id;

  curskel    = &vss->skel[skel_id];

  chv        = syWstate.current.head_vertex;
  cbr        = syWstate.branch;
  cbr->etype = SY_REAL;

  regul = local_ctrl->is_separate_saliency ? local_ctrl->regul : 
          syWstate.vsk->regul;
  
  if (syIsVerbose){
    fprintf (stderr, "(%d) ", bufdim);
  }

  if (syWstate.is_endpoint){ /* insert new term node */
    if (curskel->nodedim >= curskel->maxnodedim){
      curskel->maxnodedim += SY_MEMINC;
      curskel->node = (SyVSkeletonNode*) 
	Sy_Realloc ((char*)curskel->node, sizeof(SyVSkeletonNode)*
		    (curskel->maxnodedim+1));

      for (i = curskel->nodedim+1; i <= curskel->maxnodedim; i++){
	Sy_InitVSkeletonNode(curskel->node+i);
      }
    }

    curskel->nodedim++;
    v_label[chv] = skel_id;
    current_node = v2node[chv].current_node = curskel->nodedim;

    cn = &curskel->node[current_node];
    Sy_InitVSkeletonNode(cn);

    cn->id              = current_node;
    cn->user_id         = 0;
    cn->rank            = 1; 
    cn->maxrank         = 1;
    cn->inbranch        = (int*) Sy_Malloc(sizeof(int)*(cn->maxrank+1));
    cn->inbranch[1]     = cbr->id;
    cn->ntype           = SY_END;
    cn->radius          = syWstate.vsk->regul[chv].radius;
    cn->strength        = regul[chv].birth[0];
    cn->hierarchy_level = branch_hierarchy;
    cn->is_foreground   = syWstate.foreground;
    cn->etype           = local_ctrl->branch_el_type;
    cbr->head_branch    = 0;
    cbr->head_node      = current_node;

    /* compute x,y position */
  
    xden = syWstate.vsk->vertex[chv].xden;
    yden = syWstate.vsk->vertex[chv].yden;
    num  = syWstate.vsk->vertex[chv].num;

    switch (local_ctrl->branch_el_type){
    case SY_NILBRANCH:
    case SY_REAL:
      cn->loc.rp.x = xden/num; 
      cn->loc.rp.y = yden/num;
      break;
    case SY_INTEGER:
      cn->loc.np.x = floor(xden/num + 0.5);
      cn->loc.np.y = floor(yden/num + 0.5);
      break;
    case SY_RATIONAL:
      cn->loc.qp.xden = xden;
      cn->loc.qp.yden = yden;
      cn->loc.qp.xnum = cn->loc.qp.ynum = num;
      break;
    } /* switch */

    if (syIsVerbose){
      fprintf(stderr, "node(T)=%d ", current_node);
    }
  }

  cbr->axis = (union SyGenericPoint*) 
    Sy_Malloc (sizeof(union SyGenericPoint)*(bufdim+1));
  memcpy((char*)(cbr->axis+1), (char*)genpoint, 
	 bufdim*sizeof(union SyGenericPoint));

  cbr->ref = (SyVSkeletonRef*) Sy_Malloc(sizeof(SyVSkeletonRef)*(bufdim+1));
  memcpy((char*)(cbr->ref+1), (char*)refbuf, sizeof(SyVSkeletonRef)*bufdim);
  
  cbr->radius = (float*) Sy_Malloc(sizeof(float)*(bufdim+1));
  memcpy((char*)(cbr->radius+1), (char*)radiusbuf, sizeof(float)*bufdim);
  
  cbr->left_bndry = (SyVSkeletonBndryOffset*) 
                     Sy_Malloc(sizeof(SyVSkeletonBndryOffset)*(bufdim+1));

  memcpy((char*)(cbr->left_bndry+1), (char*)lb_buf, 
	 sizeof(SyVSkeletonBndryOffset)*bufdim);

  cbr->right_bndry = (SyVSkeletonBndryOffset*)
                      Sy_Malloc(sizeof(SyVSkeletonBndryOffset)*(bufdim+1));

  memcpy((char*)(cbr->right_bndry+1), (char*)rb_buf, 
	 sizeof(SyVSkeletonBndryOffset)*bufdim);
  
  cbr->maxelementdim = cbr->elementdim = bufdim;

  /* compute some branch characteristics */   
  cbr->length      = len_tot;
  cbr->max_radius  = max_rad;
  cbr->min_radius  = min_rad;
  cbr->mean_radius = bufdim > 0 ? sum_rad/bufdim : 0.0;
  cbr->var_radius  = bufdim > 1 ? 
    (sum_rad2-bufdim*SY_SQUARE(cbr->mean_radius))/(bufdim-1) : 0.0;

  
  if (is_disjoint){ 
    cbr->strength = cbr->max_strength = cbr->min_strength
                  = cbr->mean_strength = (len_tot-2.0)*cbr->mean_radius;
    cbr->var_strength = 0.0;
  }
  else if (bufdim <= 2){
    cbr->strength = max(strength0, strength1);
    cbr->max_strength = cbr->min_strength = cbr->mean_strength = 
      cbr->strength;
    cbr->var_strength = 0.0;
  }
  else {
    cbr->strength      = max_strength;
    cbr->max_strength  = max_strength;
    cbr->min_strength  = min_strength;
    
    cbr->mean_strength = sum_strength/(bufdim-2);
    cbr->var_strength  = bufdim <= 3 ? 0.0 : 
      (sum_strength2-(bufdim-2)*SY_SQUARE(cbr->mean_strength))/(bufdim-3);

  } /* if */
 
  cbr->hierarchy_level = branch_hierarchy;

  /* that's all folks */
  bufdim = 0; 
}

static void Sy_CompileScan(void)
{
  SyVSkeletonBranch *cbr;
  SyIpoDAC *ipd;
  SySkelDAC *vsk;
 
  SyVector rv, lv, tv;
  SyRegul *regul;

  double xden, yden, num, sal, mx, my;
  int chv, next_edge;
  int tmpint, tmp_hierarchy;

  static int left_point, right_point, 
             left_point_0, right_point_0;

  static double x0, y0, x1, y1;

  ipd = syWstate.ipd;
  vsk = syWstate.vsk;
  cbr = syWstate.branch;

  next_edge = syWstate.current.next_edge;
  regul = local_ctrl->is_separate_saliency ? local_ctrl->regul : vsk->regul;
  
  /* compute x,y position */
  chv  = syWstate.current.head_vertex;
  
  xden = vsk->vertex[chv].xden;
  yden = vsk->vertex[chv].yden;
  num  = vsk->vertex[chv].num;

  v_label[chv] = cbr->id;

  switch (local_ctrl->branch_el_type){
  case SY_NILBRANCH:
  case SY_REAL:
    genpoint[bufdim].rp.x = xden/num; 
    genpoint[bufdim].rp.y = yden/num;
    break;
  case SY_INTEGER:
    genpoint[bufdim].np.x = floor(xden/num+0.5);
    genpoint[bufdim].np.y = floor(yden/num+0.5);
    break;
  case SY_RATIONAL:
    genpoint[bufdim].qp.xden = xden;
    genpoint[bufdim].qp.yden = yden;
    genpoint[bufdim].qp.xnum = genpoint[bufdim].qp.ynum = num;
    break;
  } /* switch */

  x1 = xden/num; 
  y1 = yden/num;
  
  /* compute branch length */
  len_tot += bufdim > 0 ? SY_NORM((x1-x0),(y1-y0)) : 0.0;

#ifdef DEBUG
  fprintf (stderr, "(%4.2f,%4.2f) ", xden/num, yden/num); 
#endif 

  /* saliency info */
  sal = regul[chv].birth[2];

  switch (bufdim){
  case 0: 
    strength0 = regul[chv].birth[0];
    branch_hierarchy = vsk->is_hierarchy ? vsk->hier[next_edge] : 0;
    break;
  case 1:
    if (!syWstate.is_terminating){
      if (sal == SY_POS_INFTY) is_disjoint = SY_TRUE;
      max_strength   = sal; 
      min_strength   = sal;
      sum_strength  += sal;
      sum_strength2 += sal*sal;
    }
    else{
      strength1 = regul[chv].birth[0];
    }
    break;
  default:
    if (!syWstate.is_terminating){
      if (sal == SY_POS_INFTY) is_disjoint = SY_TRUE;
      if (max_strength < sal) max_strength = sal;
      if (min_strength > sal) min_strength = sal;
      sum_strength  += sal;
      sum_strength2 += sal*sal;
    }
    else{
      strength1 = regul[chv].birth[0];
    }
    break;
  } /* switch */

  /* compute radius */
  radiusbuf[bufdim] = vsk->regul[chv].radius;
  sum_rad  += radiusbuf[bufdim];
  sum_rad2 += SY_SQUARE(radiusbuf[bufdim]);

  if (bufdim){
    if (max_rad < radiusbuf[bufdim]) max_rad = radiusbuf[bufdim];
    if (min_rad > radiusbuf[bufdim]) min_rad = radiusbuf[bufdim];
  } 
  else {
    max_rad = radiusbuf[bufdim]; min_rad = radiusbuf[bufdim];
  }
  
  /* compute references */
  next_edge = syWstate.current.next_edge;

  if (!syWstate.is_endpoint && !syWstate.is_node && !syWstate.is_loop){
    e_label[next_edge] = cbr->id;

    /* skeleton pyramid information */
    if (vsk->is_hierarchy){
      if ((tmp_hierarchy = vsk->hier[next_edge]) > 0){
	if (branch_hierarchy == 0){
	  branch_hierarchy = tmp_hierarchy;}
	else {
	  if (branch_hierarchy > tmp_hierarchy) 
	    branch_hierarchy = tmp_hierarchy;
	}
      }
    } /* if is hierarchy */
  } /* if */

  left_point_0  = left_point;
  right_point_0 = right_point;

  left_point  = vsk->edge[next_edge].left_point;
  right_point = vsk->edge[next_edge].right_point;

  if (ipd->is_contour_index){
    refbuf[bufdim].left_contour  = ipd->contour_index[left_point].contour_id;
    refbuf[bufdim].right_contour = ipd->contour_index[right_point].contour_id;
    refbuf[bufdim].left_point    = ipd->contour_index[left_point].point_id;
    refbuf[bufdim].right_point   = ipd->contour_index[right_point].point_id;
  }
  else{
    refbuf[bufdim].left_contour  = ipd->rpe[left_point].label;
    refbuf[bufdim].right_contour = ipd->rpe[right_point].label;
    refbuf[bufdim].left_point    = left_point;
    refbuf[bufdim].right_point   = right_point;
  }

  if (bufdim > 0){ /* at least two axis points present */
    tv.dirx = x1 - x0; tv.diry = y1 - y0;
    mx = 0.5*(x1 + x0); my = 0.5*(y1 + y0);

    lv.dirx = ipd->ipo[left_point_0].x - mx;
    lv.diry = ipd->ipo[left_point_0].y - my;

    rv.dirx = ipd->ipo[right_point_0].x - mx;
    rv.diry = ipd->ipo[right_point_0].y - my;

    if (SY_DET(tv.dirx, tv.diry, lv.dirx, lv.diry) > 0.0){
      lb_buf[bufdim-1].xoff = lv.dirx;
      lb_buf[bufdim-1].yoff = lv.diry;
      rb_buf[bufdim-1].xoff = rv.dirx;
      rb_buf[bufdim-1].yoff = rv.diry;
    }
    else {/* exchange left and right */
      lb_buf[bufdim-1].xoff = rv.dirx;
      lb_buf[bufdim-1].yoff = rv.diry;
      rb_buf[bufdim-1].xoff = lv.dirx;
      rb_buf[bufdim-1].yoff = lv.diry;

      tmpint = refbuf[bufdim-1].left_point; 
      refbuf[bufdim-1].left_point  = refbuf[bufdim-1].right_point;
      refbuf[bufdim-1].right_point = tmpint;

      if (ipd->is_contour_index){
	tmpint = refbuf[bufdim-1].left_contour; 
	refbuf[bufdim-1].left_contour  = refbuf[bufdim-1].right_contour;
	refbuf[bufdim-1].right_contour = tmpint;
      }
    } /* else */

  } /* if bufdim > 0 */

  x0 = x1; y0 = y1;
  bufdim++;
}
 
SyVSkeletonSet *Sy_ExtractSkeletons (SySkelDAC *vsk, SyIpoDAC *ipd, 
				     SyControl *ctrl)
{
  SyBoolean head_flag, tail_flag, 
            conn_flag, is_border, 
            is_endpoint, is_node, 
            is_skeleton, is_foreground;
  
  int tail_vertex, head_vertex, 
      right_point, left_point;

  int i, next_edge, skel_id = 0;
  float thres;

  SyVSkeletonSet *vss;
  
  if (vsk == NULL) return NULL;
  
  local_ctrl = ctrl;
  thres      = ctrl->threshold;

  vss                = Sy_Malloc(sizeof(SyVSkeletonSet));
  vss->maxdim        = SY_MEMINC;
  vss->dim           = 0;
  vss->skel          = Sy_Malloc(sizeof(SyVSkeleton)*(vss->maxdim+1));
  vss->is_foreground = SY_TRUE;
  syWstate.vss       = vss;

  for (i = 1; i <= vss->maxdim; i++){
    Sy_InitVSkeleton(vss->skel+i);
    vss->skel[i].maxnodedim = SY_MEMINC;
    vss->skel[i].nodedim    = 0;
    vss->skel[i].node = (SyVSkeletonNode*) 
      Sy_Malloc(sizeof(SyVSkeletonNode)*(vss->skel[i].maxnodedim+1));
    vss->skel[i].maxbranchdim = SY_MEMINC;
    vss->skel[i].branchdim    = 0;
    vss->skel[i].branch = (SyVSkeletonBranch*) 
      Sy_Malloc(sizeof(SyVSkeletonBranch)*(vss->skel[i].maxnodedim+1));
  }

  /* allocate buffer for branch data */
  genpoint  = (union SyGenericPoint*) Sy_Malloc (sizeof(union SyGenericPoint)
						 *(vsk->vdim+1));
  refbuf    = (SyVSkeletonRef*) Sy_Malloc (sizeof(SyVSkeletonRef) *(vsk->vdim+1));
  radiusbuf = (float*) Sy_Malloc (sizeof(float) *(vsk->vdim+1));
  lb_buf    = (SyVSkeletonBndryOffset*) Sy_Malloc(sizeof(SyVSkeletonBndryOffset)*
						  (vsk->vdim+1));
  rb_buf    = (SyVSkeletonBndryOffset*) Sy_Malloc(sizeof(SyVSkeletonBndryOffset)*
						  (vsk->vdim+1));

  Sy_ResetWalktask (&wt);

  if (syIsVerbose){
    fprintf (stderr, "Compiling into attributed graph (Skeleton Set):\n");
    fprintf (stderr, "> PASS 1: Labeling...\n");
  }

  Sy_InitSkelTraverse (vsk, ipd, syIsVerbose, &wt);

  Sy_SkeletonTraverse (vsk, ipd, ctrl->threshold, syIsVerbose, &wt);
 
  wt.user_function[(int)INIT]  = (void(*)())NULL;
  wt.user_function[(int)TRAV]  = (void(*)())Sy_CompileTrav;
  wt.user_function[(int)TERM]  = (void(*)())Sy_CompileTerm; 
  wt.user_function[(int)ROT ]  = (void(*)())Sy_CompileRot; 
  wt.user_function[(int)FEXIT] = (void(*)())NULL;
  wt.user_function[(int)SCAN]  = (void(*)())Sy_CompileScan;
  wt.user_function[(int)WALK]  = (void(*)())NULL;
  
  if (syIsVerbose){
    fprintf(stderr, "> PASS 2: Compiling...\n");
  }

  /* allocate labels for vertices and edges 
     index array for vertices -> nodes
  */
  
  e_label = Sy_AllocInt (vsk->edim);
  v_label = Sy_AllocInt (vsk->vdim);
  v2node  = (SyVertex2NodeInfo*) 
    Sy_Malloc (sizeof(SyVertex2NodeInfo)*(vsk->vdim+1));
  memset((char*)v2node, 0, sizeof(SyVertex2NodeInfo)*(vsk->vdim+1));

  /* searching for nodes and endpoints of the skeleton */

  for (i = 1; i <= vsk->edim; i++){

    head_vertex = vsk->edge[i].head_vertex;
    tail_vertex = vsk->edge[i].tail_vertex; 

    if (v2node[head_vertex].current_node && 
	v2node[tail_vertex].current_node) continue;

    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;
    
    head_flag   = (vsk->regul[head_vertex].birth[2] > thres);
    tail_flag   = (vsk->regul[tail_vertex].birth[2] > thres);

    conn_flag   = Sy_IpoConnected (ipd, left_point, right_point);

    is_border   = (vsk->edge[i].head_edge == 0) || 
                  (vsk->edge[i].tail_edge == 0);

    is_skeleton = (head_flag && tail_flag && !conn_flag);

    if (!is_skeleton) {
      e_label[i] = -1;
      continue;
    }

    next_edge = i;

    is_endpoint = (vsk->regul[head_vertex].birth[2] > thres) &&
                  (vsk->regul[head_vertex].birth[1] <= thres);
    is_node     = (vsk->regul[head_vertex].birth[0] > thres);

    if ((is_node && v2node[head_vertex].current_node == 0) 
	|| (is_endpoint && v_label[head_vertex] == 0)){
      skel_id                     = vsk->label.e_[next_edge];
      syWstate.is_node           = is_node;
      syWstate.is_endpoint       = is_endpoint;
      syWstate.id                = skel_id;
      syWstate.start.head_vertex = head_vertex;
      syWstate.start.this_edge   = 0;
      syWstate.start.next_edge   = next_edge;
      syWstate.current           = syWstate.start;
      
      syWstate.foreground        = is_foreground = 
	Sy_SkeletonType (vsk, ipd, head_vertex, next_edge);
      
      /*  Sy_TraverseMessage (vsk, head_vertex, skel_id, is_foreground); */

      is_true_node = SY_TRUE;
      Sy_SkelRotate (vsk, ipd, &next_edge, &head_vertex, thres);
    }
    else{
      is_endpoint = (vsk->regul[tail_vertex].birth[2] > thres) &&
	            (vsk->regul[tail_vertex].birth[1] <= thres);
      is_node     = (vsk->regul[tail_vertex].birth[0] > thres);
      
      if ((is_node && v2node[tail_vertex].current_node == 0)
	  || (is_endpoint && v_label[tail_vertex] == 0)){
	skel_id                     = vsk->label.e_[next_edge];
	syWstate.is_node           = is_node;
	syWstate.is_endpoint       = is_endpoint;
	syWstate.id                = skel_id;
	syWstate.start.head_vertex = tail_vertex;
	syWstate.start.this_edge   = 0;
	syWstate.start.next_edge   = next_edge;
	syWstate.current           = syWstate.start;
	
	syWstate.foreground        = is_foreground = 
	  Sy_SkeletonType (vsk, ipd, tail_vertex, next_edge);
	
	/* Sy_TraverseMessage (vsk, tail_vertex, skel_id, is_foreground); */

	is_true_node = SY_TRUE;
	Sy_SkelRotate (vsk, ipd, &next_edge, &tail_vertex, thres);
      }
    }
  } /* for all edges */

  /* scanning for remaining loops */

  for (i = 1; i <= vsk->edim; i++){
    if (e_label[i]) continue;

    head_vertex = vsk->edge[i].head_vertex;
    tail_vertex = vsk->edge[i].tail_vertex; 
    left_point  = vsk->edge[i].left_point;
    right_point = vsk->edge[i].right_point;
    
    head_flag   = (vsk->regul[head_vertex].birth[2] > thres);
    tail_flag   = (vsk->regul[tail_vertex].birth[2] > thres);

    conn_flag   = Sy_IpoConnected (ipd, left_point, right_point);

    is_border  = (vsk->edge[i].head_edge == 0) || 
                 (vsk->edge[i].tail_edge == 0);

    is_skeleton = (head_flag && tail_flag && !conn_flag);

    if (!is_skeleton) {
      e_label[i] = -1;
      continue;
    }

    next_edge = i; 

    skel_id = vsk->label.e_[next_edge];  

    syWstate.is_node           = SY_FALSE;
    syWstate.is_endpoint       = SY_FALSE;
    syWstate.id                = skel_id;
    syWstate.start.head_vertex = tail_vertex;
    syWstate.start.this_edge   = 0;
    syWstate.start.next_edge   = next_edge;
    syWstate.current           = syWstate.start;
    
    syWstate.foreground        = is_foreground = 
      Sy_SkeletonType (vsk, ipd, tail_vertex, next_edge);
    
    is_true_node = SY_FALSE;
    Sy_SkelRotate (vsk, ipd, &next_edge, &tail_vertex, thres);

  } /* for all edges */

  Sy_Free(e_label);
  Sy_Free(v_label);
  Sy_Free(v2node);

  Sy_Free(genpoint);
  Sy_Free(refbuf);
  Sy_Free(radiusbuf);
  Sy_Free(lb_buf);
  Sy_Free(rb_buf);

  Sy_Free(vsk->label.e_); vsk->label.e_ = NULL;
  Sy_Free(vsk->label.v_); vsk->label.v_ = NULL;

  return vss;
}





