/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * potential function on TRIAN data 
 *
 */

static char sy_TrianPotiId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

static int depth     = 0, 
           max_depth = 0, 
           nrec      = 0;

static int verbose_local;

static void Sy_TermPot (SyTriDAC *tri, SyTriPot *pot, int next_edge, int current_head_point);

static SyTriPotState wp;

static FILE *fp;

SyTriPot* Sy_Tripot (SyTriDAC *tri)
{
  void Sy_TripotWalk(SyTriDAC *tri, SyTriPot *pot, int *this_edge, int *next_edge, int *current_head_point, int trid);

  SyTriPot* pot;

  SyIntVector vec, ref_vec;

  int i, trid = 0;
  int head_edge, tail_edge, 
      head_point, tail_point;

  int this_edge, next_edge, current_head_point;

  sprintf (syStrBuf, "SyTriPot %s  ", sy_TrianPotiId);
  Sy_Message (SY_FMTNL, syStrBuf);

  fp = fopen("tripot.out", "w");

  pot = Sy_AllocTriPot (tri->edim);

  for (i = 1; i <= tri->edim; i++){
    if (pot->edge[i].a_label == 0){
      if (tri->edge[i].tail_edge == 0){

	wp.start.next_edge  = next_edge = i;
	wp.start.this_edge  = this_edge = 0;
	wp.start.head_point = current_head_point = tri->edge[i].tail_point;
	wp.start.tail_val   = 0.0;
	wp.current          = wp.start;
	wp.id               = pot->edge[i].a_label = ++trid;
	wp.is_rotating      = 1;

	sprintf (syStrBuf, "starting contour %d ", trid);
	Sy_Message (SY_FMTNL, syStrBuf);
	fprintf (fp, "#starting contour %d\n", trid);
	Sy_TripotWalk (tri, pot, &this_edge, &next_edge, &current_head_point, 
			 wp.id);

	if (wp.is_rotating == 1){
	  Sy_TermPot (tri, pot, wp.start.next_edge, wp.start.head_point);
	  pot->len[trid] = wp.current.tail_val;
	}	
	sprintf (syStrBuf, "length: %f ", pot->len[trid]);
	Sy_Message (SY_FMTNL, syStrBuf);
	fprintf(fp, "\n\n");	  
      }
      else if (tri->edge[i].head_edge == 0){
	wp.start.next_edge  = next_edge = i;
	wp.start.this_edge  = this_edge = 0;
	wp.start.head_point = current_head_point = tri->edge[i].head_point;
	wp.start.tail_val   = 0.0;
	wp.current          = wp.start;
	wp.id               = pot->edge[i].a_label = ++trid;
	wp.is_rotating      = 1;

	sprintf (syStrBuf, "starting contour %d ", trid);
	Sy_Message (SY_FMTNL, syStrBuf);
	fprintf (fp, "#starting contour %d\n", trid);
	Sy_TripotWalk (tri, pot, &this_edge, &next_edge, &current_head_point, 
			 wp.id);

	if (wp.is_rotating == 1){
	  Sy_TermPot (tri, pot, wp.start.next_edge, wp.start.head_point);
	  pot->len[trid] = wp.current.tail_val;
	}
	sprintf (syStrBuf, "length: %f ", pot->len[trid]);
	Sy_Message (SY_FMTNL, syStrBuf);
	fprintf(fp, "\n\n");
      }
    }
  }

  for (i = 1; i <= tri->edim; i++){
    if (pot->edge[i].a_label == 0){
      wp.start.next_edge  = next_edge = tri->edge[i].head_edge;
      wp.start.this_edge  = this_edge = i;
      wp.start.head_point = current_head_point = tri->edge[i].head_point;
      wp.current          = wp.start;
      wp.id               = pot->edge[i].a_label = ++trid;
      wp.is_rotating      = 1;
      
      sprintf (syStrBuf, "starting contour %d ", trid);
      Sy_Message (SY_FMTNL, syStrBuf);
      fprintf (fp, "#starting contour %d\n", trid);
      Sy_TripotWalk (tri, pot, &this_edge, &next_edge, &current_head_point, 
		       wp.id);

      if (wp.is_rotating == 1) pot->len[trid] = wp.current.tail_val;

      sprintf (syStrBuf, "length: %f ", pot->len[trid]);
      Sy_Message (SY_FMTNL, syStrBuf);
      fprintf(fp, "\n\n");
    }
    else if (pot->edge[i].b_label == 0){

      tail_point = tri->edge[i].tail_point;
      head_point = tri->edge[i].head_point;

      ref_vec  = pot->edge[i].a_vec;
      vec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
      vec.diry = tri->point[head_point].y - tri->point[tail_point].y;

      if ((vec.dirx * ref_vec.dirx + vec.diry * ref_vec.diry) < 0){
	wp.start.next_edge = next_edge = tri->edge[i].head_edge;
	wp.start.head_point= current_head_point = head_point;
      }
      else{
	wp.start.next_edge = next_edge = tri->edge[i].tail_edge;
	wp.start.head_point= current_head_point = tail_point;
      }
      
      wp.start.this_edge = this_edge = i;
      wp.current         = wp.start;
      wp.id              = pot->edge[i].b_label = ++trid;
      wp.is_rotating     = 1;

      sprintf (syStrBuf, "starting contour %d ", trid);
      Sy_Message (SY_FMTNL, syStrBuf);
      fprintf (fp, "#starting contour %d\n", trid);
      Sy_TripotWalk (tri, pot, &this_edge, &next_edge, &current_head_point, 
		       wp.id);

      if (wp.is_rotating == 1) pot->len[trid] = wp.current.tail_val;
      sprintf (syStrBuf, "length: %f ", pot->len[trid]);
      Sy_Message (SY_FMTNL, syStrBuf);
      fprintf(fp, "\n\n");
    }
  }

  close(fp);
  return (pot);
}
   
void Sy_TripotWalk (SyTriDAC *tri, SyTriPot *pot, int *this_edge, int *next_edge, int *current_head_point, int trid)
{
  void Sy_TripotWalk(SyTriDAC *tri, SyTriPot *pot, int *this_edge, int *next_edge, int *current_head_point, int trid);

  SyIntVector vec;
  float   vecnorm;

  int current_head_point_buff, 
      next_edge_buff,
      this_edge_par,
      last_head_point_buff,
      last_edge_buff, ref_edge;

  int walking_flag = 1;

  int tail_point, head_point,
      tail_edge, head_edge;

  last_head_point_buff = wp.current.head_point;
  last_edge_buff       = wp.current.next_edge;
  ref_edge             = *this_edge;

  wp.last               = wp.current;
  wp.current.this_edge  = wp.current.next_edge;
  wp.current.head_point = *current_head_point;
  wp.current.next_edge  = *next_edge;
  
  if (*this_edge != 0){
    tail_point = tri->edge[*this_edge].tail_point;
    head_point = tri->edge[*this_edge].head_point;
    
    if (head_point == *current_head_point){
      vec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
      vec.diry = tri->point[head_point].y - tri->point[tail_point].y;
    }
    else{
      vec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
      vec.diry = tri->point[tail_point].y - tri->point[head_point].y;
    }
    
    vecnorm = sqrt((double)(vec.dirx*vec.dirx + vec.diry*vec.diry));
    sprintf (syStrBuf, "FORWARD, edge %d, vecnorm: %f ", *this_edge, vecnorm); 
    Sy_Message (SY_FMTNL, syStrBuf);
    
    /* *** */
    fprintf(fp, "%d %d\n", tri->point[*current_head_point].x, 
	    tri->point[*current_head_point].y);

    if (pot->edge[*this_edge].a_label == 0){
      pot->edge[*this_edge].a_label    = trid;
      pot->edge[*this_edge].a_vec      = vec;
      pot->edge[*this_edge].a_tail_val = wp.current.tail_val;
      pot->edge[*this_edge].a_head_val = (wp.current.tail_val += vecnorm);
    }
    else{
      pot->edge[*this_edge].b_label    = trid;
      pot->edge[*this_edge].b_vec      = vec;
      pot->edge[*this_edge].b_tail_val = wp.current.tail_val;
      pot->edge[*this_edge].b_head_val = (wp.current.tail_val += vecnorm);
    }
  } /* *this_edge != 0 */

  current_head_point_buff = *current_head_point;
  next_edge_buff          = *next_edge;
      
  while ((next_edge_buff != 0) && walking_flag){

    wp.current.head_point = current_head_point_buff;
    wp.current.next_edge  = next_edge_buff;
	
    *current_head_point = current_head_point_buff;
    *next_edge          = next_edge_buff;

    if (wp.is_rotating){

      head_point = tri->edge[*next_edge].head_point;
      tail_point = tri->edge[*next_edge].tail_point;
      
      if (*current_head_point == tail_point){
	*current_head_point = head_point;
	*next_edge          = tri->edge[*next_edge].head_edge;
      }
      else{
	*current_head_point = tail_point;
	*next_edge          = tri->edge[*next_edge].tail_edge;
      }

      if (wp.start.this_edge != 0){
	if (wp.start.this_edge == *next_edge) wp.is_rotating = 0;
      }
      else{ 
	if (wp.start.next_edge == *next_edge) wp.is_rotating = 0;
      }

      if (++depth > max_depth) max_depth = depth; nrec++;
      if (syIsVerbose){
	sprintf (syStrBuf, "> depth:%6d, max.depth:%6d, nrec:%6d\r",
		 depth, max_depth, nrec);
	Sy_Message (SY_FMT, syStrBuf);
      }

      this_edge_par = next_edge_buff;
      Sy_TripotWalk (tri, pot, &this_edge_par, next_edge, current_head_point, 
		       trid);
      
      wp.current.head_point = current_head_point_buff;
      wp.current.next_edge  = next_edge_buff;
      wp.last.next_edge     = last_edge_buff;
      wp.last.head_point    = last_head_point_buff;

      if (wp.is_rotating && (pot->edge[next_edge_buff].b_label == 0)){
	pot->edge[next_edge_buff].b_label    = trid;
	pot->edge[next_edge_buff].b_vec.dirx = vec.dirx = 
	  -1.0 * pot->edge[next_edge_buff].a_vec.dirx;
	pot->edge[next_edge_buff].b_vec.diry = vec.diry = 
	  -1.0 * pot->edge[next_edge_buff].a_vec.diry;
    
	vecnorm = sqrt((double)(vec.dirx*vec.dirx + vec.diry*vec.diry));
	sprintf (syStrBuf, "BACK, edge %d, vecnorm: %f ", 
		 next_edge_buff, vecnorm);    
	Sy_Message (SY_FMTNL, syStrBuf);
	head_point = tri->edge[next_edge_buff].head_point;
	
	pot->edge[next_edge_buff].b_tail_val = wp.current.tail_val;
	pot->edge[next_edge_buff].b_head_val = (wp.current.tail_val += vecnorm);
      }
      
      depth--;
      if (syIsVerbose){
	sprintf (syStrBuf, "> depth:%6d\r", depth); 
	Sy_Message (SY_FMTNL, syStrBuf);
      }
      
      if (wp.is_rotating){
	if (current_head_point_buff == tri->edge[next_edge_buff].tail_point){
	  next_edge_buff = tri->edge[next_edge_buff].tail_edge;
	}
	else{
	  next_edge_buff = tri->edge[next_edge_buff].head_edge;
	}
	if (next_edge_buff != 0){
	  walking_flag = (next_edge_buff != ref_edge);
	}
      }
      else{
	walking_flag = 0;
      }

    } /* if(wp.start.this_edge != *next_edge) */
    else {
      wp.is_rotating = walking_flag = 0;

      if ((tri->edge[*next_edge].tail_edge == 0) ||
	  (tri->edge[*next_edge].head_edge == 0)){

	tail_point = tri->edge[*next_edge].tail_point;
	head_point = tri->edge[*next_edge].head_point;

	if (head_point == current_head_point_buff){
	  vec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
	  vec.diry = tri->point[tail_point].y - tri->point[head_point].y;
	}
	else{
	  vec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
	  vec.diry = tri->point[head_point].y - tri->point[tail_point].y;
	}
	
	vecnorm = sqrt((double)(vec.dirx*vec.dirx + vec.diry*vec.diry));
	sprintf (syStrBuf, "TERMINATE, edge %d, vecnorm: %f ", 
		 *next_edge, vecnorm); 
	Sy_Message (SY_FMTNL, syStrBuf);

	pot->edge[*next_edge].b_label    = trid;
	pot->edge[*next_edge].b_vec      = vec;
	pot->edge[*next_edge].b_head_val = wp.current.tail_val;
	pot->edge[*next_edge].b_tail_val = (wp.current.tail_val += vecnorm);
      }
      pot->len[trid] = wp.current.tail_val;
    }
  } /* while */
}

static void Sy_TermPot (SyTriDAC *tri, SyTriPot *pot, int next_edge, int current_head_point)
{ 
  int tail_point, head_point;
  SyIntVector vec;
  float   vecnorm;

  tail_point = tri->edge[next_edge].tail_point;
  head_point = tri->edge[next_edge].head_point;
  
  if (head_point == current_head_point){
    vec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
    vec.diry = tri->point[head_point].y - tri->point[tail_point].y;
  }
  else{
    vec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
    vec.diry = tri->point[tail_point].y - tri->point[head_point].y;
  }
  
  vecnorm = sqrt((double)(vec.dirx*vec.dirx + vec.diry*vec.diry));
  sprintf (syStrBuf, "TERM (SINGLE), edge %d, vecnorm: %f ", 
	   next_edge, vecnorm); 
  Sy_Message (SY_FMTNL, syStrBuf);

  pot->edge[next_edge].b_label    = wp.id;
  pot->edge[next_edge].b_vec      = vec;
  pot->edge[next_edge].b_head_val = wp.current.tail_val;
  pot->edge[next_edge].b_tail_val = (wp.current.tail_val += vecnorm);
}

SyTriPot* Sy_AllocTriPot (int nr_elem)
{
  SyTriPot* pot;

  int page, pot_mem, len_mem, nr_obj = 512;

  pot = (SyTriPot*) Sy_Malloc (sizeof(SyTriPot));

  page       = getpagesize();
  pot_mem    = ((++nr_elem* sizeof(SyTriPotElement))/page+1)*page;
  pot->edge = (SyTriPotElement*) Sy_Malloc ((unsigned)pot_mem);
  (void) memset ((char*)pot->edge, 0, pot_mem);

  len_mem   = (nr_obj+1)*sizeof(SY_TRIAN_OBJ_TAB);
  pot->len  = (SY_TRIAN_OBJ_TAB*) Sy_Malloc((unsigned)len_mem);
  (void) memset ((char*)pot->len, 0, len_mem);

  pot->obj_count = 0;
  pot->obj_dim   = nr_obj;
  pot->count     = 0;
  pot->edim      = nr_elem;

  return (pot);
}









