/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

static char sy_TrianRegularId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"

extern enum SyResidualCompMode     residualCompMode;
extern enum SyResidualBorderMode   residualBorderMode;

extern Boolean is_residual_fix,
               is_maxfix;

static is_verbose = 1;

static void Sy_GetTriDist(SySkelDAC *vsk, SyTriDAC *tri, SyTriPot
			  *pot, int left_point, int right_point, 
			  int ref_vertex, float *ret_dist, int *is_path); 

static void Sy_GetSector(SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot, 
			 int ref_point, int ref_vertex, int *first_edge, 
			 int *second_edge, int o_point, int *is_connected); 

static void Sy_TriCorresEdge(SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot, 
			     int left_point, int right_point, int tail_vertex, 
			     int head_vertex, SyVector *tvec, SyVector *hvec, 
			     int rank, float *ret_dist, int *is_path, 
			     int *is_tcorr, int *is_hcorr, int *is_mindev); 

static void Sy_IsCorrection(SySkelDAC *vsk, int tail_vertex, int head_vertex,
			    SyVector lovec, SyVector hivec, SyVector rvec0, 
			    int *is_corr);

static void Sy_Insert(SySkelDAC *vsk, int vindex, float dev, float rad); 

void Sy_CorrResidEdge(SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		      SyVertex tvertex, SyVertex hvertex, SyVector tvec0, 
		      SyVector hvec0, int left_point, int right_point, 
		      int ref_point, int *tcorr, int *hcorr, float *dist, 
		      int is_mindev);

void  Sy_CorrResidRay(SySkelDAC *vsk, SyIpoDAC *ipd, SyRegpot *pot, 
		      SyVertex tvertex, SyVertex hvertex, SyVector tvec0, 
		      int left_point, int right_point, int ref_point, 
		      int *tcorr, float *dist, int *is_mindev);

void Sy_ComplexRegularize (SySkelDAC *vsk, SyTriDAC *tri, 
			   void (*regul_method) (/* ??? */))
{
  SyTriPot* pot;
/*
  if (residualBorderMode == RES_BORDER){
    Sy_SkelBorder_with_tri (vsk, tri, is_verbose);
  }
  else{
    vsk->is_border = 0;
  }
*/
  vsk->is_border = 0;

  sprintf (syStrBuf, "Regularizer (Triangulation)%s ", sy_TrianRegularId);
  Sy_Message (SY_FMTNL, syStrBuf);

  vsk->regul = Sy_AllocRegul (vsk->vdim);

  if (residualCompMode == RES_FULL){ 
    vsk->eresid    = Sy_AllocEresid (vsk->edim);
    vsk->erdim     = vsk->edim;
    vsk->is_eresid = 1;
  }
  else{ 
    vsk->eresid    = (SyEresid*) NULL;
    vsk->erdim     = 0;
    vsk->is_eresid = 0;
  }

  Sy_Message (SY_FMTNL, "Contour Tracer: ");

  pot = Sy_Tripot (tri);
  Sy_MakeTriPointref (tri);

  Sy_Message (SY_FMTNL, " Residual Computation:");

  (*regul_method) (vsk, tri, pot);
/*
  if (is_residual_fix) Sy_Infix_with_tri  (vsk, tri);
  if (is_maxfix)       Sy_Maxfix_with_tri (vsk, tri);
*/
}

void Sy_ComplexCirResidual (SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot)
{ 
  int    i, corrcount = 0; 
  SyPoint  lpoint, rpoint; 
  SyVertex tvertex, hvertex; 
  float  dist, dev, tdev, hdev;
  double trad, hrad; 
  double tper, hper; 
  SyVector tvec[2], hvec[2]; 
  int    left_point, right_point; 
  int    tail_vertex, head_vertex; 
  double targ, harg; 
  int    is_tcorr, is_hcorr, is_path, is_mindev; 

  Sy_Message (SY_FMTNL, "Method: Circularity Residual ");

  vsk->prune_method = SKE_CIR;

  vsk->point[0].x = vsk->point[0].y = 0;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){

    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;

      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;
      hvec[1].dirx = rpoint.x - hvertex.xden/hvertex.num;
      hvec[1].diry = rpoint.y - hvertex.yden/hvertex.num;

      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      Sy_TriCorresEdge (vsk, tri, pot, left_point, right_point, 
			    tail_vertex, head_vertex, tvec, hvec, 2,
			    &dist, &is_path, &is_tcorr, &is_hcorr, &is_mindev);

      if (!is_path){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
        if (is_tcorr) corrcount++;
	if (is_hcorr) corrcount++;

	targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	       (trad*trad);

	harg = (hvec[0].dirx*hvec[1].dirx + hvec[0].diry*hvec[1].diry)/
               (hrad*hrad);

        if (targ >  1.0) targ =  1.0;
        if (targ < -1.0) targ = -1.0;
        if (harg >  1.0) harg =  1.0;
        if (harg < -1.0) harg = -1.0;

	if (is_mindev){
	  tdev = SY_FRAC_2_PI * (SY_PI_2*dist - trad*acos(targ));
	  hdev = SY_FRAC_2_PI * (SY_PI_2*dist - hrad*acos(harg));
	  dev  = min(tdev, hdev);

	  Sy_Insert (vsk, tail_vertex, tdev = dev, trad);
	  Sy_Insert (vsk, head_vertex, hdev = dev, hrad);
	}
	else{
	  tper = trad * ((is_tcorr) ? (2*SY_PI - acos(targ)) : acos(targ));
	  hper = hrad * ((is_hcorr) ? (2*SY_PI - acos(harg)) : acos(harg));
	  
	  Sy_Insert (vsk, tail_vertex, tdev = SY_FRAC_2_PI *(SY_PI_2*dist-tper), trad);
	  Sy_Insert (vsk, head_vertex, hdev = SY_FRAC_2_PI *(SY_PI_2*dist-hper), hrad);
	}

	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = tdev;
	  vsk->eresid[i].head_residual = hdev;
	  vsk->eresid[i].tail_corr     = is_tcorr;
	  vsk->eresid[i].head_corr     = is_hcorr;
	  vsk->eresid[i].dist          = dist;
	}
      }
      break;
    case 1:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      head_vertex = vsk->edge[i].head_vertex;
      tail_vertex = vsk->edge[i].tail_vertex;
      
      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      
      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));

      Sy_TriCorresEdge (vsk, tri, pot, left_point, right_point, 
			    tail_vertex, head_vertex, tvec, hvec, 1,
			    &dist, &is_path, &is_tcorr, &is_hcorr, &is_mindev);

      if (!is_path){
	
	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	  }
	
	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
      }
      else{	
	if (is_tcorr) corrcount++;

	targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	       (trad*trad);
	
	if (targ >  1.0) targ =  1.0;
	if (targ < -1.0) targ = -1.0;

	tper = trad * ((is_tcorr) ? (2*SY_PI - acos(targ)) : acos(targ));
	dev  = SY_FRAC_2_PI * (SY_PI_2*dist-tper);
	
	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dev;
	  vsk->eresid[i].head_residual = SY_NEG_INFTY;
	  vsk->eresid[i].tail_corr     = is_tcorr;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	}
	
	Sy_Insert (vsk, tail_vertex, dev, trad);
	break;
      }
    }
  }
  sprintf (syStrBuf, "%d corrections ", corrcount);
  Sy_Message (SY_FMTNL, syStrBuf);
}

static void Sy_Insert(SySkelDAC *vsk, int vindex, float dev, float rad);

void Sy_ComplexKleinResidual (SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot)
{ 
  int    i, is_path; 
  SyPoint  lpoint, rpoint; 
  SyVertex tvertex, hvertex; 
  double trad, hrad;
  float  dist, dev; 
  double tper, hper; 
  SyVector tvec[2], hvec[2]; 
  int    left_point, right_point; 
  int    tail_vertex, head_vertex; 
  double targ, harg; 

  Sy_Message (SY_FMTNL, "Method: Klein's Residual ");

  vsk->prune_method = SKE_KLE;
  vsk->point[0].x = vsk->point[0].y = 0;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){
    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;
      
      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;
      hvec[1].dirx = rpoint.x - hvertex.xden/hvertex.num;
      hvec[1].diry = rpoint.y - hvertex.yden/hvertex.num;
      
      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));
      
      Sy_GetTriDist (vsk, tri, pot, left_point, right_point, head_vertex,
			 &dist, &is_path);
      
      if (!is_path){
	
 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
	targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	       (trad*trad);

	harg = (hvec[0].dirx*hvec[1].dirx + hvec[0].diry*hvec[1].diry)/
	        (hrad*hrad);

        if (targ >  1.0) targ =  1.0;
        if (targ < -1.0) targ = -1.0;
        if (harg >  1.0) harg =  1.0;
        if (harg < -1.0) harg = -1.0;

	tper = trad * acos(targ);
	hper = hrad * acos(harg);
	dev = min(dist-tper, dist-hper);

	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dev;
	  vsk->eresid[i].head_residual = dev;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	}

	Sy_Insert (vsk, tail_vertex, dev, trad);
	Sy_Insert (vsk, head_vertex, dev, hrad);
      }
      break;
    case 1:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      head_vertex = vsk->edge[i].head_vertex;
      tail_vertex = vsk->edge[i].tail_vertex;
      
      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      
      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      
      Sy_GetTriDist (vsk, tri, pot, left_point, right_point, head_vertex,
			 &dist, &is_path);

      if (!is_path){
	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
      }
      else{	
	targ = (tvec[0].dirx*tvec[1].dirx + tvec[0].diry*tvec[1].diry)/
	       (trad*trad);

	if (targ >  1.0) targ =  1.0;
	if (targ < -1.0) targ = -1.0;

	tper = trad * acos(targ);
	dev  = dist-tper;

	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dev;
	  vsk->eresid[i].head_residual = SY_NEG_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	}
	Sy_Insert (vsk, tail_vertex, dev, trad);
	break;
      }
    }
  }
}
 
void Sy_ComplexChordResidual (SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot)
{
  int    i, is_path;
  SyPoint  lpoint, rpoint;
  SyVertex tvertex, hvertex;
  double trad, hrad; 
  float  dist, dev;
  double tsec, hsec;
  SyVector tvec[2], hvec[2];
  int    left_point, right_point;
  int    tail_vertex, head_vertex;

  Sy_Message (SY_FMTNL, "Method: Chord Residual ");
  vsk->prune_method = SKE_CHO;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){
    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;

      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;
      hvec[1].dirx = rpoint.x - hvertex.xden/hvertex.num;
      hvec[1].diry = rpoint.y - hvertex.yden/hvertex.num;

      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      Sy_GetTriDist (vsk, tri, pot, left_point, right_point, head_vertex,
			 &dist, &is_path);

      if (!is_path){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
	tsec = sqrt(SY_SQUARE(tvec[0].dirx-tvec[1].dirx) + 
                    SY_SQUARE(tvec[0].diry-tvec[1].diry));
	hsec = sqrt(SY_SQUARE(hvec[0].dirx-hvec[1].dirx) + 
                    SY_SQUARE(hvec[0].diry-hvec[1].diry));
	
	dev  = min(dist-tsec, dist-hsec);
	
	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dist-tsec;
	  vsk->eresid[i].head_residual = dist-hsec;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	}

	Sy_Insert (vsk, tail_vertex, dev, trad);
	Sy_Insert (vsk, head_vertex, dev, hrad);
      }
      break;
    case 1:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      
      lpoint  = vsk->point[left_point];
      rpoint  = vsk->point[right_point];
      tvertex = vsk->vertex[tail_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      tvec[1].dirx = rpoint.x - tvertex.xden/tvertex.num;
      tvec[1].diry = rpoint.y - tvertex.yden/tvertex.num;
      
      Sy_GetTriDist (vsk, tri, pot, left_point, right_point, head_vertex,
		       &dist, &is_path);
      
      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));

      if (!is_path){
	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
      }
      else{	
	tsec = sqrt(SY_SQUARE(tvec[0].dirx-tvec[1].dirx) + 
                    SY_SQUARE(tvec[0].diry-tvec[1].diry));
	
	dev  = dist-tsec;
	
	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dev;
	  vsk->eresid[i].head_residual = SY_NEG_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	}
	Sy_Insert (vsk, tail_vertex, dev, trad);
	break;
      }
    }
  }
}

void Sy_ComplexPotResidual (SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot)
{
  int    i, is_path;
  SyPoint  lpoint;
  SyVertex tvertex, hvertex;
  double trad, hrad;
  float  dist;
  SyVector tvec[2], hvec[2];
  int    left_point, right_point;
  int    tail_vertex, head_vertex;

  Sy_Message (SY_FMTNL, "Method: Potentials  ");

  vsk->prune_method = SKE_POT;

  for (i = 1; i <= vsk->vdim; i++){
    vsk->regul[i].birth[0] = SY_NEG_INFTY;
    vsk->regul[i].birth[1] = SY_NEG_INFTY;
    vsk->regul[i].birth[2] = SY_NEG_INFTY;
  }

  for (i = 1; i <= vsk->edim; i++){
    switch(vsk->edge[i].rank){
    case 2:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      head_vertex = vsk->edge[i].head_vertex;

      lpoint  = vsk->point[left_point];
      tvertex = vsk->vertex[tail_vertex];
      hvertex = vsk->vertex[head_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      hvec[0].dirx = lpoint.x - hvertex.xden/hvertex.num;
      hvec[0].diry = lpoint.y - hvertex.yden/hvertex.num;

      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      hrad = sqrt(SY_SQUARE(hvec[0].dirx) + SY_SQUARE(hvec[0].diry));

      Sy_GetTriDist (vsk, tri, pot, left_point, right_point, head_vertex,
			 &dist, &is_path);

      if (!is_path){

 	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = SY_POS_INFTY;
	  vsk->eresid[i].head_residual = SY_POS_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = 0.0;
	}

	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
	Sy_Insert(vsk, head_vertex, SY_POS_INFTY, hrad);
      }
      else{	
	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dist;
	  vsk->eresid[i].head_residual = dist;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	}

	Sy_Insert (vsk, tail_vertex, dist, trad);
	Sy_Insert (vsk, head_vertex, dist, hrad);
      }
      break;
    case 1:
      left_point  = vsk->edge[i].left_point;
      right_point = vsk->edge[i].right_point;
      tail_vertex = vsk->edge[i].tail_vertex;
      
      lpoint  = vsk->point[left_point];
      tvertex = vsk->vertex[tail_vertex];
      
      tvec[0].dirx = lpoint.x - tvertex.xden/tvertex.num;
      tvec[0].diry = lpoint.y - tvertex.yden/tvertex.num;
      
      trad = sqrt(SY_SQUARE(tvec[0].dirx) + SY_SQUARE(tvec[0].diry));
      
      Sy_GetTriDist (vsk, tri, pot, left_point, right_point, head_vertex,
			 &dist, &is_path);

      
      if (!is_path){
	Sy_Insert(vsk, tail_vertex, SY_POS_INFTY, trad);
      }
      else{	
	if (residualCompMode == RES_FULL){
	  vsk->eresid[i].tail_residual = dist;
	  vsk->eresid[i].head_residual = SY_NEG_INFTY;
	  vsk->eresid[i].tail_corr     = 0;
	  vsk->eresid[i].head_corr     = 0;
	  vsk->eresid[i].dist          = dist;
	}
	Sy_Insert (vsk, tail_vertex, dist, trad);
	break;
      }
    }
  }
}

static void Sy_Insert (SySkelDAC *vsk, int vindex, float dev, float rad)
{
  float*  birth;

  vsk->regul[vindex].radius = rad;
  birth = vsk->regul[vindex].birth;

  if (birth[2] < dev){
    birth[0] = birth[1];
    birth[1] = birth[2];
    birth[2] = dev;
    return;
  }
  if (birth[1] < dev){
    birth[0] = birth[1];
    birth[1] = dev;
    return;
  }
  if (birth[0] < dev){
    birth[0] = dev;
  }
}

static void Sy_GetTriDist(SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot, 
			  int left_point, int right_point, int ref_vertex, 
			  float *ret_dist, int *is_path)
{
  int    ref_point, a_edge, b_edge; 
  float  left_potval, right_potval;
  float  dist;
  int    left_label, right_label,
         head_point, tail_point;
  int    is_connected;

  SyIntVector tvec;

  if ((tri->pref[left_point] == 0 ) || 
      (tri->pref[right_point] == 0)){
    *is_path  = 0;
    *ret_dist = SY_POS_INFTY;
    return;
  }

  Sy_GetSector (vsk, tri, pot, ref_point = left_point, ref_vertex,
		   &a_edge, &b_edge, right_point, &is_connected);

  if (is_connected){
    *is_path  = 1;
    *ret_dist = pot->edge[a_edge].a_head_val - pot->edge[a_edge].a_tail_val; 
    return;
  }

  head_point = tri->edge[a_edge].head_point;
  tail_point = tri->edge[a_edge].tail_point;

  if (head_point == ref_point){
    tvec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
    tvec.diry = tri->point[head_point].y - tri->point[tail_point].y;
  }
  else{
    tvec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
    tvec.diry = tri->point[tail_point].y - tri->point[head_point].y;
  }    

  if ((pot->edge[a_edge].a_vec.dirx * tvec.dirx +
      pot->edge[a_edge].a_vec.diry * tvec.diry) > 0){
    left_potval = pot->edge[a_edge].a_head_val;
    left_label  = pot->edge[a_edge].a_label;
  }
  else{
    left_potval = pot->edge[a_edge].b_head_val;
    left_label  = pot->edge[a_edge].b_label;
  }
  
  Sy_GetSector (vsk, tri, pot, ref_point = right_point, ref_vertex,
		   &a_edge, &b_edge, left_point, &is_connected);

  if (is_connected){
    *is_path  = 1;
    *ret_dist = pot->edge[a_edge].a_head_val - pot->edge[a_edge].a_tail_val; 
    return;
  }

  head_point = tri->edge[a_edge].head_point;
  tail_point = tri->edge[a_edge].tail_point;

  if (head_point == ref_point){
    tvec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
    tvec.diry = tri->point[head_point].y - tri->point[tail_point].y;
  }
  else{
    tvec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
    tvec.diry = tri->point[tail_point].y - tri->point[head_point].y;
  }    

  if ((pot->edge[a_edge].a_vec.dirx * tvec.dirx +
      pot->edge[a_edge].a_vec.diry * tvec.diry) > 0){
    right_potval = pot->edge[a_edge].a_head_val;
    right_label  = pot->edge[a_edge].a_label;
  }
  else{
    right_potval = pot->edge[a_edge].b_head_val;
    right_label  = pot->edge[a_edge].b_label;
  }
  
  if (left_label != right_label){
    *ret_dist = SY_POS_INFTY;
    *is_path  = 0;
    return;
  }
  else{    
    dist  = SY_ALV(left_potval - right_potval);
    dist  = min(dist, pot->len[left_label] - dist);
    *ret_dist = SY_ALV(dist);
    *is_path  = 1;
    return;
  }
}


static void Sy_GetSector (SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot, 
			  int ref_point, int ref_vertex, int *first_edge, 
			  int *second_edge, int o_point, int *is_connected)
{
  SyVector   hvec, a_tvec, b_tvec, rvec; 
  SyDblPoint p, q;
  SyVertex   *v, vv;
  int      ori;
  double   lvalue, rvalue;

  int    aindex, bindex, next_edge, ref_edge,
         head_point, tail_point;

  int    is_in_a_hplane   = 0, 
         is_in_b_hplane   = 0,
         is_in_ref_hplane = 0;

  *is_connected = 0;

  /* v = &(vsk->vertex[ref_vertex]); */

  p.x = tri->point[ref_point].x;
  p.y = tri->point[ref_point].y;
  q.x = tri->point[o_point].x;
  q.y = tri->point[o_point].y;

  vv.xden = p.x + q.x;
  vv.yden = p.y + q.y; 
  vv.num  = 2.0;

  v = &vv;

  ref_edge = aindex = tri->pref[ref_point];

  next_edge = bindex = (tri->edge[ref_edge].tail_point == ref_point) ? 
    tri->edge[ref_edge].tail_edge : tri->edge[ref_edge].head_edge;

  if (next_edge == 0){
    *first_edge = *second_edge = ref_edge;
    return;
  }

  if (tri->edge[ref_edge].tail_point == ref_point ){
    head_point = tri->edge[ref_edge].head_point;
    rvec.dirx = tri->point[head_point].x - p.x;
    rvec.diry = tri->point[head_point].y - p.y;
    }
  else{
    tail_point = tri->edge[ref_edge].tail_point;
    rvec.dirx = tri->point[tail_point].x - p.x;
    rvec.diry = tri->point[tail_point].y - p.y;
  }
  
  a_tvec    = rvec;

  hvec.dirx = rvec.diry;
  hvec.diry = -1.0 * rvec.dirx;
 
  lvalue = hvec.dirx*(v->xden - p.x * v->num);
  rvalue = hvec.diry*(p.y * v->num - v->yden);

  if (lvalue == rvalue){
    *first_edge = *second_edge = ref_edge;
    *is_connected = 1;
    return;
  }

  is_in_a_hplane = is_in_ref_hplane = (hvec.dirx*(v->xden - p.x * v->num))>=
    (hvec.diry*(p.y * v->num - v->yden));
  
  while ((next_edge != ref_edge) && (next_edge != 0)){

    if (tri->edge[next_edge].tail_point == ref_point ){
      head_point = tri->edge[next_edge].head_point;
      b_tvec.dirx = tri->point[head_point].x - p.x;
      b_tvec.diry = tri->point[head_point].y - p.y;
    }
    else{
      tail_point = tri->edge[next_edge].tail_point;
      b_tvec.dirx = tri->point[tail_point].x - p.x;
      b_tvec.diry = tri->point[tail_point].y - p.y;
    }

    hvec.dirx = b_tvec.diry;
    hvec.diry = -1.0 * b_tvec.dirx;

    lvalue = hvec.dirx*(v->xden - p.x * v->num);
    rvalue = hvec.diry*(p.y * v->num - v->yden);

    if (lvalue == rvalue){
      *first_edge = *second_edge = next_edge;
      *is_connected = 1;
      return;
    }

    is_in_b_hplane = (hvec.dirx*(v->xden - p.x * v->num))>=
                     (hvec.diry*(p.y * v->num - v->yden));

    ori = SY_DET(a_tvec.dirx, a_tvec.diry, b_tvec.dirx, b_tvec.diry);

    if (ori <= 0){
      if (is_in_a_hplane && !is_in_b_hplane){
	*first_edge  = aindex;
	*second_edge = bindex;
	return;
      }
      else{
	is_in_a_hplane = is_in_b_hplane;
	aindex         = bindex;
	a_tvec         = b_tvec;
      }
    }
    else{
      if (is_in_a_hplane || !is_in_b_hplane){
	*first_edge  = aindex;
	*second_edge = bindex;
	return;
      }
      else{
	is_in_a_hplane = is_in_b_hplane;
	aindex         = bindex;
	a_tvec         = b_tvec;
      }
    }      
      
    if (ref_point == tri->edge[next_edge].tail_point){
      next_edge = bindex = tri->edge[next_edge].tail_edge;
    }
    else{
      next_edge = bindex = tri->edge[next_edge].head_edge;
    }
  }

  ori = SY_DET(a_tvec.dirx, a_tvec.diry, rvec.dirx, rvec.diry);

  if (ori <= 0){
    if (is_in_a_hplane && !is_in_ref_hplane){
      *first_edge  = aindex;
      *second_edge = bindex;
      return;
      }
    else{
      *first_edge = -1; *second_edge = -2;
      Sy_SimpleError (SY_FMTNL, "Internal Error in Sy_GetSector");
      return;
    }
  }
  else{
    if (is_in_a_hplane || !is_in_ref_hplane){
      *first_edge  = aindex;
      *second_edge = bindex;
      return;
    }
    else{
      *first_edge = -1; *second_edge = -2;
      Sy_SimpleError (SY_FMTNL, "Internal Error in Sy_GetSector");
      return;
    }
  }      
}

static void Sy_TriCorresEdge (SySkelDAC *vsk, SyTriDAC *tri, SyTriPot *pot, 
			      int left_point, int right_point, 
			      int tail_vertex, int head_vertex, 
			      SyVector *tvec, SyVector *hvec, int rank, 
			      float *ret_dist, int *is_path, int *is_tcorr, 
			      int *is_hcorr, int *is_mindev)
{
  SyVector lovec, hivec;
  int    ref_point, 
         left_a_edge, left_b_edge,
         right_a_edge, right_b_edge; 
  float  left_potval,  right_potval, 
         l_aux_potval, r_aux_potval;
  float  dist, lodist;
  int    left_label, right_label,
         head_point, tail_point;
  int    is_connected;

  int    is_r_tcorr, is_r_hcorr, 
         is_l_tcorr, is_l_hcorr;

  int    l_mask = 0, r_mask = 0;

  SyIntVector l_tvec, r_tvec, laux_tvec, raux_tvec;

  *is_mindev = 0;

  if ((tri->pref[left_point] == 0 ) || 
      (tri->pref[right_point] == 0)){
    *is_path  = 0;
    *ret_dist = SY_POS_INFTY;
    return;
  }

  Sy_GetSector (vsk, tri, pot, ref_point = left_point, head_vertex,
		   &left_a_edge, &left_b_edge, right_point, &is_connected);

  if (is_connected){
    *is_path  = 1;
    *ret_dist = pot->edge[left_a_edge].a_head_val - 
                pot->edge[left_a_edge].a_tail_val; 
    *is_tcorr = *is_hcorr = 0;
    return;
  }

  head_point = tri->edge[left_b_edge].head_point;
  tail_point = tri->edge[left_b_edge].tail_point;

  if (head_point == ref_point){
    laux_tvec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
    laux_tvec.diry = tri->point[tail_point].y - tri->point[head_point].y;
  }
  else{
    laux_tvec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
    laux_tvec.diry = tri->point[head_point].y - tri->point[tail_point].y;
  }    

  head_point = tri->edge[left_a_edge].head_point;
  tail_point = tri->edge[left_a_edge].tail_point;

  if (head_point == ref_point){
    l_tvec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
    l_tvec.diry = tri->point[head_point].y - tri->point[tail_point].y;
  }
  else{
    l_tvec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
    l_tvec.diry = tri->point[tail_point].y - tri->point[head_point].y;
  }    

  if ((pot->edge[left_a_edge].a_vec.dirx * l_tvec.dirx +
      pot->edge[left_a_edge].a_vec.diry * l_tvec.diry) > 0){
    left_potval  = pot->edge[left_a_edge].a_head_val;
    l_aux_potval = pot->edge[left_a_edge].a_tail_val;
    left_label   = pot->edge[left_a_edge].a_label;
  }
  else{
    left_potval  = pot->edge[left_a_edge].b_head_val;
    l_aux_potval = pot->edge[left_a_edge].b_tail_val;
    left_label   = pot->edge[left_a_edge].b_label;
  }
  
  Sy_GetSector (vsk, tri, pot, ref_point = right_point, head_vertex,
		   &right_a_edge, &right_b_edge, left_point, &is_connected);

  if (is_connected){
    *is_path  = 1;
    *ret_dist = pot->edge[right_a_edge].a_head_val - 
                pot->edge[right_a_edge].a_tail_val;
    *is_tcorr    = *is_hcorr = 0; 
    return;
  }

  head_point = tri->edge[right_b_edge].head_point;
  tail_point = tri->edge[right_b_edge].tail_point;

  if (head_point == ref_point){
    raux_tvec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
    raux_tvec.diry = tri->point[tail_point].y - tri->point[head_point].y;
  }
  else{
    raux_tvec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
    raux_tvec.diry = tri->point[head_point].y - tri->point[tail_point].y;
  }    

  head_point = tri->edge[right_a_edge].head_point;
  tail_point = tri->edge[right_a_edge].tail_point;

  if (head_point == ref_point){
    r_tvec.dirx = tri->point[head_point].x - tri->point[tail_point].x;
    r_tvec.diry = tri->point[head_point].y - tri->point[tail_point].y;
  }
  else{
    r_tvec.dirx = tri->point[tail_point].x - tri->point[head_point].x;
    r_tvec.diry = tri->point[tail_point].y - tri->point[head_point].y;
  }    

  if ((pot->edge[right_a_edge].a_vec.dirx * r_tvec.dirx +
      pot->edge[right_a_edge].a_vec.diry * r_tvec.diry) > 0){
    right_potval = pot->edge[right_a_edge].a_head_val;
    r_aux_potval = pot->edge[right_a_edge].a_tail_val;
    right_label  = pot->edge[right_a_edge].a_label;
  }
  else{
    right_potval = pot->edge[right_a_edge].b_head_val;
    r_aux_potval = pot->edge[right_a_edge].b_tail_val;
    right_label  = pot->edge[right_a_edge].b_label;
  }
  
  if (left_label != right_label){
    *ret_dist = SY_POS_INFTY;
    *is_path  = 0;
    *is_tcorr = *is_hcorr = 0;
    return;
  }
  else{ /* left_label == right_label */    
    dist  = SY_ALV(left_potval - right_potval);
    dist  = min(dist, pot->len[left_label] - dist);
    *ret_dist = dist = SY_ALV(dist);
    *is_path  = 1;
    
    if (left_a_edge == left_b_edge){
      l_mask = 1;
    }
    else{
      lodist = SY_ALV(l_aux_potval - right_potval);
      lodist = min(lodist, pot->len[left_label] - lodist);
      lodist = SY_ALV(lodist);
      if (lodist < dist){
	lovec.dirx = -1.0 * l_tvec.dirx;
	lovec.diry = -1.0 * l_tvec.diry;
	hivec.dirx = laux_tvec.dirx;
        hivec.diry = laux_tvec.diry;
      }
      else{
	hivec.dirx = -1.0 * l_tvec.dirx;
	hivec.diry = -1.0 * l_tvec.diry;
	lovec.dirx = laux_tvec.dirx;
        lovec.diry = laux_tvec.diry;
      }
      Sy_IsCorrection (vsk, tail_vertex, head_vertex, lovec, hivec, 
		       tvec[0], &is_l_tcorr);
      if (rank > 1){
	Sy_IsCorrection (vsk, tail_vertex, head_vertex, lovec, hivec, 
			 hvec[0], &is_l_hcorr);
      }
    }

    if (right_a_edge == right_b_edge){
      r_mask = 1;
    }
    else{
      lodist = SY_ALV(r_aux_potval - left_potval);
      lodist = min(lodist, pot->len[left_label] - lodist);
      lodist = SY_ALV(lodist);
      if (lodist < dist){
	lovec.dirx = -1.0 * r_tvec.dirx;
	lovec.diry = -1.0 * r_tvec.diry;
	hivec.dirx = raux_tvec.dirx;
        hivec.diry = raux_tvec.diry;
      }
      else{
	hivec.dirx = -1.0 * r_tvec.dirx;
	hivec.diry = -1.0 * r_tvec.diry;
	lovec.dirx = raux_tvec.dirx;
        lovec.diry = raux_tvec.diry;
      }
      Sy_IsCorrection (vsk, tail_vertex, head_vertex, lovec, hivec, 
		       tvec[1], &is_r_tcorr);
      if (rank > 1){
	Sy_IsCorrection (vsk, tail_vertex, head_vertex, lovec, hivec, 
			 hvec[1], &is_r_hcorr);
      }
    }

    if (l_mask && r_mask){
      *is_mindev = 1;
      *is_tcorr  = *is_hcorr = 0;
      return;
    }
    if (l_mask){
      *is_mindev = 0;
      *is_tcorr  = is_r_tcorr; 
      *is_hcorr  = is_r_hcorr; 
      return;
    }
    if (r_mask){
      *is_mindev = 0;
      *is_tcorr  = is_l_tcorr; 
      *is_hcorr  = is_l_hcorr; 
      return;
    }
    
    *is_tcorr = is_l_tcorr && is_r_tcorr;
    *is_hcorr = is_l_hcorr && is_r_hcorr;
    return;
  }
}


static void Sy_IsCorrection (SySkelDAC *vsk, int tail_vertex, 
			     int head_vertex, SyVector lovec, 
			     SyVector hivec, SyVector rvec0, int *is_corr)
{
  SyVector  evec, flow, normal;
  SyVertex  *tvertex, *hvertex;
  double  lonorm, hinorm,
          loproj, hiproj, refproj;

  tvertex = &vsk->vertex[tail_vertex];
  hvertex = &vsk->vertex[head_vertex];

  evec.dirx = hvertex->xden/hvertex->num - tvertex->xden/tvertex->num;
  evec.diry = hvertex->yden/hvertex->num - tvertex->yden/tvertex->num;

  lonorm = sqrt(SY_SQUARE(lovec.dirx)+SY_SQUARE(lovec.diry));
  hinorm = sqrt(SY_SQUARE(hivec.dirx)+SY_SQUARE(hivec.diry));

  if ((evec.dirx * rvec0.dirx + evec.diry * rvec0.diry) < 0){
    flow.dirx = - evec.dirx; 
    flow.diry = - evec.diry;
  } else flow = evec;

  normal.dirx = rvec0.diry; normal.diry = - rvec0.dirx;

  if ((flow.dirx * normal.dirx + flow.diry * normal.diry) < 0){
    normal.dirx = - normal.dirx; normal.diry = - normal.diry;
  }
  loproj = lovec.dirx * normal.dirx + lovec.diry * normal.diry;
  hiproj = hivec.dirx * normal.dirx + hivec.diry * normal.diry;

 if (SY_SGN(loproj) != SY_SGN(hiproj)){
    *is_corr = loproj < hiproj;
  }
  else{
    refproj = loproj;
    loproj  = (lovec.dirx * rvec0.dirx + lovec.diry * rvec0.diry)/lonorm;
    hiproj  = (hivec.dirx * rvec0.dirx + hivec.diry * rvec0.diry)/hinorm;
    *is_corr  = (refproj > 0) ? loproj > hiproj : loproj < hiproj;
  }
}




