
/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Swiss Federal Institute of Technology ETH
 *          Zurich, Switzerland (rlo@vision.ethz.ch)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1993 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/*
 * purpose: utilities
 *
 */

static char sy_UtilId[] = "$Id$"; 

#include "Sy_Skeleton.h"
#include "Sy_Intrinsic.h"


#if defined(mips) || defined(ultrix)
char* strdup (s)
char* s;
{
  char *retval;
  retval = (char*) Sy_Malloc (sizeof(char)*(strlen(s)+1));
  strcpy (retval, s);
  return (retval);
}
#endif

void Sy_Usage (char *progName, char **usageStrings)
{
  char** strPtr;

  strPtr = usageStrings;

  fprintf (stderr, "@ usage: %s %s\n", progName, *strPtr);

  for (strPtr++; *strPtr; strPtr++){
    fprintf (stderr, "%s\n", *strPtr);
  }
}

char* Sy_BaseName (char *str)
{
  char *p, *q;
  p = strrchr (str, (int)'/');
  if ((p == NULL) || (*(p+1) == '\0')) return(str);
  q = strdup (p+1);
  if (q == NULL) return(str);
  return (q);
}

char* Sy_RawName (char *str)
{
  char *p, *q, *ex, *tmp;

  tmp = strdup (str);
  if (tmp == NULL) return (str); /* error handling */

  p   = strrchr (tmp, (int)'/'); 
  if (p == NULL){
    ex  = strchr  (tmp, (int)'.');
    if ((ex != NULL) && (tmp != ex)) *ex = '\0';
    q = strdup (tmp);
  }
  else{
    ex = strchr (p, (int)'.');
    if ((ex != NULL) && (*(p+1) != '.')) *ex = '\0';
    q = strdup (p+1);
  }
  if (q == NULL) return(tmp); /* error handling */
  free(tmp);
  return (q);
}

char* Sy_CreateSimpleFilename (char *str, char *ext)
{
  char* Sy_RawName(char *str);
  char  *tmpch, *out;
  tmpch = Sy_RawName(str);
  out   = (char*) Sy_Malloc((strlen(str)+strlen(ext)+4)*sizeof(char));
  out   = (char*) strcpy (out, tmpch);
  out   = strcat (out, ext);
  free(tmpch);
  return(out);
}

void Sy_PrintUsage (char *progName, char **usageStrings)
{
  char** strPtr;

  strPtr = usageStrings;

  fprintf (stderr, "@ usage: %s %s\n", progName, *strPtr);

  for (strPtr++; *strPtr; strPtr++){
    fprintf (stderr, "%s\n", *strPtr);
  }
}

extern Boolean syIsVerbose;

void Sy_Message (char *format, char *arg)
{ 
  if (format == (char*)NULL || arg == (char*)NULL) return;
  if (syIsVerbose){fprintf (stderr, format, arg);
		   fflush(stderr);}
}

void Sy_Warning (char *format, char *arg)
{ 
  if (format == (char*)NULL || arg == (char*)NULL) return;
  if (syIsVerbose) fprintf (stderr, "\nWARNING: ");
  if (syIsVerbose){fprintf (stderr, format, arg);
		   fflush(stderr);}
}

void Sy_SimpleError (char *format, char *arg)
{ 
  if (format == (char*)NULL || arg == (char*)NULL) return;
  fprintf (stderr, "\nERROR: ");
  fprintf (stderr, format, arg); fflush(stderr);
  exit(2);
}

void Sy_Error (char *module, char *function, char *format, char *arg)
{
  if (module == (char*)NULL || function == (char*) NULL || 
      format == (char*)NULL || arg == (char*)NULL) return;
  fprintf (stderr, "\nERROR in module %s, function %s:\n", module, function);
  fprintf (stderr, format, arg); fflush(stderr);
  exit(2);
}
      
void Sy_DestroyVoronoi (SyVoroDAC *vd)
{
  if (vd == NULL) return;

  if (vd->point) {
    free (vd->point); vd->point = NULL; 
  }
  if (vd->edge){
    free (vd->edge); vd->edge = NULL;
  }
  if (vd->vertex) {
    free (vd->vertex); vd->vertex = NULL; 
  }

  if (vd->label.p_ != NULL){
    free (vd->label.p_); vd->label.p_ = NULL;
  }
  if (vd->label.e_ != NULL){
    free (vd->label.e_); vd->label.e_ = NULL;
  }
  if (vd->label.v_ != NULL){
    free (vd->label.v_); vd->label.v_ = NULL;
  }
  
  vd->pdim = vd->edim = vd->vdim = 0;

  Sy_InitVoronoi(vd);
}

void Sy_DestroySkeleton (SySkelDAC *vsk)
{
  if (vsk == NULL) return;

  if (vsk->point) {
    free (vsk->point); vsk->point = NULL; 
  }
  if (vsk->edge) {
    free (vsk->edge); vsk->edge = NULL;
  }
  if (vsk->vertex) {
    free (vsk->vertex); vsk->vertex = NULL;
  }
  if (vsk->regul){ 
    free (vsk->regul); vsk->regul = NULL;
  }
  if (vsk->regul_aux){ 
    free (vsk->regul_aux); vsk->regul_aux = NULL;
  }

  if (vsk->hier && vsk->is_hierarchy){ 
    free (vsk->hier); vsk->hier = NULL;
    vsk->is_hierarchy = SY_FALSE;
  }
  if (vsk->eresid && vsk->is_eresid){ 
    free (vsk->eresid); vsk->eresid = NULL;
    vsk->is_eresid = SY_FALSE;
  }

  if (vsk->label.p_ != NULL){
    free (vsk->label.p_); vsk->label.p_ = NULL;
  }
  if (vsk->label.e_ != NULL){
    free (vsk->label.e_); vsk->label.e_ = NULL;
  }
  if (vsk->label.v_ != NULL){
    free (vsk->label.v_); vsk->label.v_ = NULL;
  }

  vsk->pdim = vsk->edim = vsk->vdim = 0;
  
  Sy_InitSkeleton(vsk);
}

void Sy_DestroyIpo (SyIpoDAC *ipd)
{
  if (ipd == NULL) return;

  if (ipd->ipo  != NULL){ 
    free (ipd->ipo); ipd->ipo = NULL;
  }
  if (ipd->icn  != NULL){
    free (ipd->icn); ipd->icn = NULL;
  }
  if (ipd->ipl  != NULL){ 
    free (ipd->ipl); ipd->ipl = NULL;
  }
  if (ipd->info  != NULL){ 
    free (ipd->info); ipd->info = NULL;
  }

  if (ipd->rpe  != NULL){ 
    free (ipd->rpe); ipd->rpe = NULL;
  }

  if (ipd->label.p_ != NULL){
    free (ipd->label.p_); ipd->label.p_ = NULL;
  }
  if (ipd->label.c_ != NULL){
    free (ipd->label.c_); ipd->label.c_ = NULL;
  }
  if (ipd->contour_index != NULL){
    free (ipd->contour_index); ipd->contour_index = NULL;
    ipd->is_contour_index = SY_FALSE;
  }
  
  ipd->ipodim = ipd->icndim = 0;

  Sy_InitIpo(ipd);
}

void Sy_DestroyTrian (SyTriDAC *tri)
{
  if (tri == NULL) return;

  if (tri->point     != NULL){
    free (tri->point); tri->point = NULL;
  }
  if (tri->edge      != NULL){ 
    free (tri->edge); tri->edge = NULL;
  }
  if (tri->regul     != NULL){ 
    free (tri->regul); tri->regul = NULL;
  }

  if (tri->label.p_ != NULL){
    free (tri->label.p_); tri->label.p_ = NULL;
  }
  if (tri->label.e_ != NULL){
    free (tri->label.e_); tri->label.e_ = NULL;
  }
  if (tri->label.v_ != NULL){
    free (tri->label.v_); tri->label.v_ = NULL;
  }
  if (tri->label.v_ != NULL){
    free (tri->label.v_); tri->label.v_ = NULL;
  }

  if (tri->eref     != NULL){ 
    free (tri->eref); tri->eref = NULL;
  }
  if (tri->pref     != NULL){ 
    free (tri->pref); tri->pref = NULL;
  }

  tri->pdim = tri->edim = 0;

  Sy_InitTrian(tri);
}

void Sy_NormVector (SyVector *v)
{
  double norm;
  
  norm = SY_NORM(v->dirx, v->diry);
  if (norm < 1.0e-50) return;

  norm = 1.0/norm;
  v->dirx *= norm; v->diry *= norm;
  return;
}

void Sy_InitIpo (SyIpoDAC *ipd)
{
  if (ipd == NULL) return;

  ipd->ipo      = NULL;
  ipd->icn      = NULL;
  ipd->ipl      = NULL;
  ipd->fipo     = NULL;
  ipd->info     = NULL;
  ipd->rpe      = NULL;
  ipd->label.p_ = NULL;
  ipd->label.c_ = NULL;

  ipd->contour_index    = NULL;
  ipd->is_contour_index = SY_FALSE;

  ipd->contourdim = ipd->maxcontourdim = 0;
  ipd->ipodim     = ipd->icndim        = 0;
  ipd->max_x = ipd->min_x = ipd->min_y = ipd->max_y = 0;
}

void Sy_InitVoronoi (SyVoroDAC *vd)
{
  if (vd == NULL) return;

  vd->point    = NULL; 
  vd->edge     = NULL;
  vd->vertex   = NULL; 
  vd->label.p_ = NULL;
  vd->label.e_ = NULL;
  vd->label.v_ = NULL;
  
  vd->pdim = vd->edim = vd->vdim = 0;
}

void Sy_InitSkeleton (SySkelDAC *vsk)
{
  if (vsk == NULL) return;

  vsk->point        = NULL; 
  vsk->edge         = NULL;
  vsk->vertex       = NULL;
  vsk->regul        = NULL;
  vsk->regul_aux    = NULL;
  vsk->hier         = NULL;
  vsk->eresid       = NULL;
  vsk->label.p_     = NULL;
  vsk->label.e_     = NULL;
  vsk->label.v_     = NULL;

  vsk->is_hierarchy       = SY_FALSE;
  vsk->is_eresid          = SY_FALSE;
  vsk->is_border          = SY_FALSE;
  vsk->is_smcir           = SY_FALSE;
  vsk->is_nfurcation_info = SY_FALSE;
 
  vsk->pdim = vsk->edim = vsk->vdim = 0;
}

void Sy_InitTrian (SyTriDAC *tri)
{
  if (tri == NULL) return;

  tri->point    = NULL;
  tri->edge     = NULL;
  tri->regul    = NULL;
  tri->label.p_ = NULL;
  tri->label.e_ = NULL;
  tri->label.v_ = NULL;
  tri->eref     = NULL;
  tri->pref     = NULL;

  tri->is_hardlink = NULL;

  tri->pdim     = tri->edim    = 0;
  tri->regdim   = tri->erefdim = 0;
  
  tri->is_linkinfo = SY_FALSE;
  tri->is_pointref = SY_FALSE;
}

void Sy_DestroyMScaleVertex(SyMScaleVertex *mv)
{
  if (mv == NULL) return;
  Sy_Free(mv->sc);
  Sy_Free(mv->incident);
}

void Sy_DestroyMScaleEdge(SyMScaleEdge *me)
{
  if (me == NULL) return;
  Sy_Free(me->sc);
}

void Sy_DestroyMScaleSkelDAC (SyMScaleSkelDAC *mvsk)
{
  int i;
  if (mvsk == NULL) return;

  if (mvsk->point) {
    free (mvsk->point); mvsk->point = NULL; 
  }
  if (mvsk->edge) {
    for (i = 0; i <= mvsk->edim; i++){
      Sy_DestroyMScaleEdge (mvsk->edge + i);
    }
    free (mvsk->edge); mvsk->edge = NULL;
  }
  if (mvsk->vertex) {
    for (i = 0; i <= mvsk->vdim; i++){
      Sy_DestroyMScaleVertex (mvsk->vertex + i);
    }
    free (mvsk->vertex); mvsk->vertex = NULL;
  }
  if (mvsk->sc){free (mvsk->sc); mvsk->sc = NULL;}

  if (mvsk->edge_select){ 
    free (mvsk->edge_select); mvsk->edge_select = NULL;
  }

  if (mvsk->vertex_select){ 
    free (mvsk->vertex_select); mvsk->vertex_select = NULL;
  }

  if (mvsk->label != NULL){
    free (mvsk->label); mvsk->label = NULL;
  }

  mvsk->pdim = mvsk->edim = mvsk->vdim = 0;  
}

void Sy_DestroySkeletonSet (SyVSkeletonSet* vss)
{
  SyVSkeletonSet    *rv;
  SyVSkeleton       *rs, *vs;
  SyVSkeletonBranch *rb, *b;
  int i,k;
  int eff_skeldim, eff_branchdim, eff_nodedim;

  if (vss == NULL) return;
  
  eff_skeldim = max(vss->dim, vss->maxdim);

  for (i = 1; i <= eff_skeldim; i++){
    vs = vss->skel + i;
    
    eff_nodedim = max(vs->nodedim, vs->maxnodedim);
    for (k = 1; k <= eff_nodedim; k++){
      if (vs->node[k].inbranch && vs->node[k].maxrank > 0) 
	free (vs->node[k].inbranch);
    }
      
    eff_branchdim = max(vs->branchdim, vs->maxbranchdim);
    for (k = 1; k <= eff_branchdim; k++){
      b  = vs->branch + k;
      if (b->axis)        free (b->axis);
      if (b->radius)      free (b->radius);
      if (b->ref)         free (b->ref);
      if (b->left_bndry)  free (b->left_bndry);
      if (b->right_bndry) free (b->right_bndry);
    }

    if (vs->node)   free (vs->node);
    if (vs->branch) free (vs->branch);

  } /* for all skeletons */
  
  if (vss->skel) free (vss->skel);

}

SyVSkeletonSet* Sy_CopyVSkeletonSet (SyVSkeletonSet* vss)
{
  SyVSkeletonSet    *rv;
  SyVSkeleton       *rs, *vs;
  SyVSkeletonBranch *rb, *b;
  int i,k;
  int nodebytes, branchbytes, skelbytes, bytes;

  rv = (SyVSkeletonSet*) malloc (sizeof(SyVSkeletonSet));
  memcpy (rv, vss, sizeof(SyVSkeletonSet));

  skelbytes = sizeof(SyVSkeleton)*(vss->maxdim + 1);
  rv->skel = (SyVSkeleton*) malloc (skelbytes);
  memcpy (rv->skel, vss->skel, skelbytes);
  
  for (i = 1; i <= vss->dim; i++){
    rs = rv->skel + i;
    vs = vss->skel + i;
    
    branchbytes = sizeof(SyVSkeletonBranch) * (vs->maxbranchdim + 1);
    rs->branch = (SyVSkeletonBranch*) malloc(branchbytes);

    nodebytes = sizeof(SyVSkeletonNode) * (vs->maxnodedim + 1);
    rs->node = (SyVSkeletonNode*) malloc(nodebytes);

    memcpy(rs->branch, vs->branch, branchbytes);
    memcpy(rs->node, vs->node, nodebytes);
    
    for (k = 1; k <= vs->nodedim; k++){
      bytes = sizeof(int)*(rs->node[k].maxrank+1);
      rs->node[k].inbranch = (int*) malloc (bytes);
      memcpy (rs->node[k].inbranch, vs->node[k].inbranch, bytes);
    }
      
    for (k = 1; k <= vs->branchdim; k++){
      rb = rs->branch + k;
      b  = vs->branch + k;
      bytes  = sizeof(union SyGenericPoint) * (b->maxelementdim + 1);
      rb->axis = (union SyGenericPoint*) malloc (bytes);
      memcpy (rb->axis, b->axis, bytes);
      
      bytes = sizeof(float) * (b->maxelementdim + 1);
      rb->radius = (float*) malloc (bytes);
      memcpy (rb->radius, b->radius, bytes);

      bytes = sizeof(SyVSkeletonRef) * (b->maxelementdim + 1);
      rb->ref = (SyVSkeletonRef*) malloc (bytes);
      memcpy (rb->ref, b->ref, bytes);

      bytes = sizeof(SyVSkeletonBndryOffset) * (b->maxelementdim + 1);
      rb->left_bndry = (SyVSkeletonBndryOffset*) malloc (bytes);
      memcpy (rb->left_bndry, b->left_bndry, bytes);

      bytes = sizeof(SyVSkeletonBndryOffset) * (b->maxelementdim + 1);
      rb->right_bndry = (SyVSkeletonBndryOffset*) malloc (bytes);
      memcpy (rb->right_bndry, b->right_bndry, bytes);
    }

  } /* for all skeletons */

  return rv;
}
