extern char *malloc(), *realloc();

# line 1 "Sy_IoParser.yacc"
 

/*
 *  $Id$
 * 
 *  Author: Robert L. Ogniewicz, Division of Applied Sciences,
 *          Harvard University, Cambridge, Massachusetts (rlo@hrl.harvard.edu)
 */

/* Copyright Notice
 * ================
 * Copyright (C) 1995 by Robert L. Ogniewicz
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
 */

/* yacc grammar rules for I/O routines */

#include <stdio.h>
#include "Sy_Lang.h"

extern SyAvlTree  *sy_yacc_tree;
extern SyAvlTree  *sy_func_tree;
extern SyIoObject *sy_global_obj_mem;

SyIoToken    sy_yacc_iotoken;

static SyIoObject   *tmp_obj_mem;
static SyDataObject *sy_yacc_fo;
static char* objnamestk[1024];
static stkptr = 0;

int   tmp_int;
float tmp_float;
char  strbuf[1024], tmp_str[1024];
int   is_data = 0;

static void call_loader(char *funcname);
static void call_object_loader(char *objname);

# line 56 "Sy_IoParser.yacc"
typedef union   {
	 int   ival;
	 float fval;
         char  chval;
	 char  strval[1024];
	} YYSTYPE;
# define INTEGER 257
# define HEX 258
# define OBJECT 259
# define STOP 260
# define ASSIGN 261
# define ATERM 262
# define COMMENTSKIP 263
# define COMMENTBEGIN 264
# define COMMENTEND 265
# define MEMORY 266
# define SHMEM 267
# define FILEIO 268
# define BEGINLOADSEQ 269
# define REAL 270
# define SYMBOL 271
# define IDENT 272
# define STRING 273
#define yyclearin yychar = -1
#define yyerrok yyerrflag = 0
extern int yychar;
extern int yyerrflag;
#ifndef YYMAXDEPTH
#define YYMAXDEPTH 150
#endif
YYSTYPE yylval, yyval;
# define YYERRCODE 256

# line 201 "Sy_IoParser.yacc"
 /* start of programs */

static void call_loader(char *funcname)
{
  int retval;

  is_data = 1; yyclearin; 
#ifdef DEBUG
  printf ("beginning load sequence, calling %s\n", funcname);
#endif
  tmp_obj_mem = sy_global_obj_mem;
  sy_yacc_fo = Sy_GetDataObject(&sy_func_tree, "callback", funcname);
  sy_global_obj_mem = tmp_obj_mem;
  if (sy_yacc_fo == NULL){
    fprintf (stderr, "Undefined loader function `%s'.\n", funcname);
    /* trying to skip over data segment */
    retval = Sy_SyncWithStream(sy_yacc_iotoken.in, SY_SYNC_COOKIE_BEGIN);
    if (retval == 0){
      fprintf (stderr, "Fatal error: could not synchronize with input data\n");
      exit(2);
    }
    retval = Sy_SyncWithStream(sy_yacc_iotoken.in, SY_SYNC_COOKIE_END);
    if (retval == 0){
      fprintf (stderr, "Fatal error: could not synchronize with input data\n");
      exit(2);
    }
  }
  else{
    sy_yacc_iotoken.funcname = sy_yacc_fo->name;
    retval = ((int(*)(SyIoToken*))
	      sy_yacc_fo->t.func_val)(&sy_yacc_iotoken);
    if (retval == 0){
      fprintf (stderr, "Fatal error: could not synchronize with input data\n");
      exit(2);
    }
  }
}

static void call_object_loader(char *objname)
{
  int retval, nchar;
  char *funcname = NULL;
  static char prefix[] = "Sy_Load_";
  int prefixlen, objnamelen;

  SyDataObject *dobj;

  dobj = Sy_GetDataObject (&sy_yacc_tree, objname, "type");
  if (dobj == NULL){
#ifdef DEBUG
    fprintf (stderr, "Undefined type of object `%s'.\n", objname);
#endif
    return;
  }

  /* create function name */
  prefixlen = strlen(prefix); objnamelen = strlen(dobj->t.s_val);
  nchar = prefixlen + objnamelen;
  funcname = (char*) Sy_Malloc (sizeof(char)*(nchar+1));
  strncpy(funcname,prefix,prefixlen);
  strncpy(funcname+prefixlen,dobj->t.s_val,objnamelen+1);

  tmp_obj_mem = sy_global_obj_mem;
  sy_yacc_fo  = Sy_GetDataObject(&sy_func_tree, "callback", funcname);
  sy_global_obj_mem = tmp_obj_mem;

#ifdef DEBUG
  fprintf (stderr, "calling loader %s\n", funcname);
#endif

  if (sy_yacc_fo == NULL){
#ifdef DEBUG
    fprintf (stderr, "Undefined loader for object `%s'.\n", objname);
#else
    ; /* do nothing */
#endif	
  }
  else{
    sy_yacc_iotoken.funcname = sy_yacc_fo->name;
    retval = ((int(*)(SyIoToken*))
	      sy_yacc_fo->t.func_val)(&sy_yacc_iotoken);
    if (retval == 0){
      fprintf (stderr, "Fatal error: could not load object `%s'\n", objname);
    }
  }
  Sy_Free(funcname);
}

#include "lex.yy.c"


int yyexca[] ={
-1, 0,
	0, 2,
	-2, 0,
-1, 1,
	0, -1,
	-2, 0,
	};
# define YYNPROD 35
# define YYLAST 239
int yyact[]={

    57,     3,    11,    13,     9,    33,    55,    11,    13,     9,
    51,    35,    27,    64,    15,    12,    63,    10,    14,    15,
    12,    62,    10,    14,    42,    43,    44,    19,    20,    41,
    40,    39,    38,    37,    24,    23,    22,    21,    54,    53,
    47,    26,     8,     7,     6,     5,    56,    50,    49,    48,
    46,    45,    25,     4,    52,    36,    16,    18,    17,     2,
     1,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,    60,    59,    58,    61,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,    11,    13,     9,     0,     0,     0,     0,     0,
     0,     0,     0,     0,    15,    12,     0,    10,    14,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
     0,     0,    28,    30,     0,     0,     0,     0,     0,     0,
     0,     0,     0,     0,    34,    29,     0,    31,    32 };
int yypact[]={

  -255, -1000,  -250, -1000, -1000, -1000, -1000, -1000, -1000, -1000,
  -234,  -225,  -226,  -227,  -228,    12, -1000,   -18,  -260,   -35,
 -1000, -1000, -1000, -1000, -1000,  -261, -1000, -1000,  -229,  -230,
  -231,  -232,  -233,  -242,    11,     9,   -83, -1000, -1000, -1000,
 -1000, -1000,     8,     7,     6,  -262, -1000, -1000,  -219,  -220,
  -267,     5,  -125,  -241,  -246,  -249, -1000, -1000, -1000, -1000,
 -1000, -1000, -1000, -1000, -1000 };
int yypgo[]={

     0,    60,    59,    58,    53,    45,    44,    43,    42,    57,
    55,    54 };
int yyr1[]={

     0,     1,     1,     3,     1,     2,     2,     4,     4,     4,
     4,     9,    10,     5,    11,    11,    11,    11,    11,     6,
     6,     6,     6,     6,     7,     7,     7,     7,     7,     8,
     8,     8,     8,     8,     8 };
int yyr2[]={

     0,     2,     0,     1,     6,     2,     4,     2,     2,     2,
     2,     1,     1,    15,     0,     4,     4,     4,     4,     9,
     9,     9,     9,     9,     4,     4,     4,     4,     4,    15,
    15,    15,    13,     9,     7 };
int yychk[]={

 -1000,    -1,    -2,   256,    -4,    -5,    -6,    -7,    -8,   259,
   272,   257,   270,   258,   273,   269,    -4,    -3,    -9,   261,
   262,   262,   262,   262,   262,    40,    59,   272,   257,   270,
   258,   272,   273,    40,   269,   272,   -10,   262,   262,   262,
   262,   262,   266,   267,   268,    40,    41,   123,    41,    41,
    41,   272,   -11,   258,   258,   273,    41,   125,    -6,    -7,
    -8,    -5,   262,   262,   262 };
int yydef[]={

    -2,    -2,     1,     3,     5,     7,     8,     9,    10,    11,
     0,     0,     0,     0,     0,     0,     6,     0,     0,     0,
    27,    24,    25,    26,    28,     0,     4,    12,     0,     0,
     0,     0,     0,     0,    34,     0,     0,    19,    20,    21,
    22,    23,     0,     0,     0,     0,    33,    14,     0,     0,
     0,     0,     0,     0,     0,     0,    32,    13,    15,    16,
    17,    18,    29,    30,    31 };
typedef struct { char *t_name; int t_val; } yytoktype;
#ifndef YYDEBUG
#	define YYDEBUG	0	/* don't allow debugging */
#endif

#if YYDEBUG

yytoktype yytoks[] =
{
	"INTEGER",	257,
	"HEX",	258,
	"OBJECT",	259,
	"STOP",	260,
	"ASSIGN",	261,
	"ATERM",	262,
	"COMMENTSKIP",	263,
	"COMMENTBEGIN",	264,
	"COMMENTEND",	265,
	"MEMORY",	266,
	"SHMEM",	267,
	"FILEIO",	268,
	"BEGINLOADSEQ",	269,
	"REAL",	270,
	"SYMBOL",	271,
	"IDENT",	272,
	"STRING",	273,
	"-unknown-",	-1	/* ends search */
};

char * yyreds[] =
{
	"-no such reduction-",
	"entity : objects",
	"entity : /* empty */",
	"entity : error",
	"entity : error ';'",
	"objects : genobj",
	"objects : objects genobj",
	"genobj : object",
	"genobj : assignment",
	"genobj : simpleval",
	"genobj : loadinstr",
	"object : OBJECT",
	"object : OBJECT IDENT",
	"object : OBJECT IDENT '{' valist '}'",
	"valist : /* empty */",
	"valist : valist assignment",
	"valist : valist simpleval",
	"valist : valist loadinstr",
	"valist : valist object",
	"assignment : IDENT ASSIGN INTEGER ATERM",
	"assignment : IDENT ASSIGN REAL ATERM",
	"assignment : IDENT ASSIGN HEX ATERM",
	"assignment : IDENT ASSIGN IDENT ATERM",
	"assignment : IDENT ASSIGN STRING ATERM",
	"simpleval : INTEGER ATERM",
	"simpleval : REAL ATERM",
	"simpleval : HEX ATERM",
	"simpleval : IDENT ATERM",
	"simpleval : STRING ATERM",
	"loadinstr : IDENT ASSIGN '(' MEMORY ')' HEX ATERM",
	"loadinstr : IDENT ASSIGN '(' SHMEM ')' HEX ATERM",
	"loadinstr : IDENT ASSIGN '(' FILEIO ')' STRING ATERM",
	"loadinstr : IDENT ASSIGN BEGINLOADSEQ '(' IDENT ')'",
	"loadinstr : BEGINLOADSEQ '(' IDENT ')'",
	"loadinstr : IDENT ASSIGN BEGINLOADSEQ",
};
#endif /* YYDEBUG */
#line 1 "/usr/lib/yaccpar"
/*	@(#)yaccpar 1.10 89/04/04 SMI; from S5R3 1.10	*/

/*
** Skeleton parser driver for yacc output
*/

/*
** yacc user known macros and defines
*/
#define YYERROR		goto yyerrlab
#define YYACCEPT	{ free(yys); free(yyv); return(0); }
#define YYABORT		{ free(yys); free(yyv); return(1); }
#define YYBACKUP( newtoken, newvalue )\
{\
	if ( yychar >= 0 || ( yyr2[ yytmp ] >> 1 ) != 1 )\
	{\
		yyerror( "syntax error - cannot backup" );\
		goto yyerrlab;\
	}\
	yychar = newtoken;\
	yystate = *yyps;\
	yylval = newvalue;\
	goto yynewstate;\
}
#define YYRECOVERING()	(!!yyerrflag)
#ifndef YYDEBUG
#	define YYDEBUG	1	/* make debugging available */
#endif

/*
** user known globals
*/
int yydebug;			/* set to 1 to get debugging */

/*
** driver internal defines
*/
#define YYFLAG		(-1000)

/*
** static variables used by the parser
*/
static YYSTYPE *yyv;			/* value stack */
static int *yys;			/* state stack */

static YYSTYPE *yypv;			/* top of value stack */
static int *yyps;			/* top of state stack */

static int yystate;			/* current state */
static int yytmp;			/* extra var (lasts between blocks) */

int yynerrs;			/* number of errors */

int yyerrflag;			/* error recovery flag */
int yychar;			/* current input token number */


/*
** yyparse - return 0 if worked, 1 if syntax error not recovered from
*/
int
yyparse()
{
	register YYSTYPE *yypvt;	/* top of value stack for $vars */
	unsigned yymaxdepth = YYMAXDEPTH;

	/*
	** Initialize externals - yyparse may be called more than once
	*/
	yyv = (YYSTYPE*)malloc(yymaxdepth*sizeof(YYSTYPE));
	yys = (int*)malloc(yymaxdepth*sizeof(int));
	if (!yyv || !yys)
	{
		yyerror( "out of memory" );
		return(1);
	}
	yypv = &yyv[-1];
	yyps = &yys[-1];
	yystate = 0;
	yytmp = 0;
	yynerrs = 0;
	yyerrflag = 0;
	yychar = -1;

	goto yystack;
	{
		register YYSTYPE *yy_pv;	/* top of value stack */
		register int *yy_ps;		/* top of state stack */
		register int yy_state;		/* current state */
		register int  yy_n;		/* internal state number info */

		/*
		** get globals into registers.
		** branch to here only if YYBACKUP was called.
		*/
	yynewstate:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;
		goto yy_newstate;

		/*
		** get globals into registers.
		** either we just started, or we just finished a reduction
		*/
	yystack:
		yy_pv = yypv;
		yy_ps = yyps;
		yy_state = yystate;

		/*
		** top of for (;;) loop while no reductions done
		*/
	yy_stack:
		/*
		** put a state and value onto the stacks
		*/
#if YYDEBUG
		/*
		** if debugging, look up token value in list of value vs.
		** name pairs.  0 and negative (-1) are special values.
		** Note: linear search is used since time is not a real
		** consideration while debugging.
		*/
		if ( yydebug )
		{
			register int yy_i;

			(void)printf( "State %d, token ", yy_state );
			if ( yychar == 0 )
				(void)printf( "end-of-file\n" );
			else if ( yychar < 0 )
				(void)printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				(void)printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ++yy_ps >= &yys[ yymaxdepth ] )	/* room on stack? */
		{
			/*
			** reallocate and recover.  Note that pointers
			** have to be reset, or bad things will happen
			*/
			int yyps_index = (yy_ps - yys);
			int yypv_index = (yy_pv - yyv);
			int yypvt_index = (yypvt - yyv);
			yymaxdepth += YYMAXDEPTH;
			yyv = (YYSTYPE*)realloc((char*)yyv,
				yymaxdepth * sizeof(YYSTYPE));
			yys = (int*)realloc((char*)yys,
				yymaxdepth * sizeof(int));
			if (!yyv || !yys)
			{
				yyerror( "yacc stack overflow" );
				return(1);
			}
			yy_ps = yys + yyps_index;
			yy_pv = yyv + yypv_index;
			yypvt = yyv + yypvt_index;
		}
		*yy_ps = yy_state;
		*++yy_pv = yyval;

		/*
		** we have a new state - find out what to do
		*/
	yy_newstate:
		if ( ( yy_n = yypact[ yy_state ] ) <= YYFLAG )
			goto yydefault;		/* simple state */
#if YYDEBUG
		/*
		** if debugging, need to mark whether new token grabbed
		*/
		yytmp = yychar < 0;
#endif
		if ( ( yychar < 0 ) && ( ( yychar = yylex() ) < 0 ) )
			yychar = 0;		/* reached EOF */
#if YYDEBUG
		if ( yydebug && yytmp )
		{
			register int yy_i;

			(void)printf( "Received token " );
			if ( yychar == 0 )
				(void)printf( "end-of-file\n" );
			else if ( yychar < 0 )
				(void)printf( "-none-\n" );
			else
			{
				for ( yy_i = 0; yytoks[yy_i].t_val >= 0;
					yy_i++ )
				{
					if ( yytoks[yy_i].t_val == yychar )
						break;
				}
				(void)printf( "%s\n", yytoks[yy_i].t_name );
			}
		}
#endif /* YYDEBUG */
		if ( ( ( yy_n += yychar ) < 0 ) || ( yy_n >= YYLAST ) )
			goto yydefault;
		if ( yychk[ yy_n = yyact[ yy_n ] ] == yychar )	/*valid shift*/
		{
			yychar = -1;
			yyval = yylval;
			yy_state = yy_n;
			if ( yyerrflag > 0 )
				yyerrflag--;
			goto yy_stack;
		}

	yydefault:
		if ( ( yy_n = yydef[ yy_state ] ) == -2 )
		{
#if YYDEBUG
			yytmp = yychar < 0;
#endif
			if ( ( yychar < 0 ) && ( ( yychar = yylex() ) < 0 ) )
				yychar = 0;		/* reached EOF */
#if YYDEBUG
			if ( yydebug && yytmp )
			{
				register int yy_i;

				(void)printf( "Received token " );
				if ( yychar == 0 )
					(void)printf( "end-of-file\n" );
				else if ( yychar < 0 )
					(void)printf( "-none-\n" );
				else
				{
					for ( yy_i = 0;
						yytoks[yy_i].t_val >= 0;
						yy_i++ )
					{
						if ( yytoks[yy_i].t_val
							== yychar )
						{
							break;
						}
					}
					(void)printf( "%s\n", yytoks[yy_i].t_name );
				}
			}
#endif /* YYDEBUG */
			/*
			** look through exception table
			*/
			{
				register int *yyxi = yyexca;

				while ( ( *yyxi != -1 ) ||
					( yyxi[1] != yy_state ) )
				{
					yyxi += 2;
				}
				while ( ( *(yyxi += 2) >= 0 ) &&
					( *yyxi != yychar ) )
					;
				if ( ( yy_n = yyxi[1] ) < 0 )
					YYACCEPT;
			}
		}

		/*
		** check for syntax error
		*/
		if ( yy_n == 0 )	/* have an error */
		{
			/* no worry about speed here! */
			switch ( yyerrflag )
			{
			case 0:		/* new error */
				yyerror( "syntax error" );
				goto skip_init;
			yyerrlab:
				/*
				** get globals into registers.
				** we have a user generated syntax type error
				*/
				yy_pv = yypv;
				yy_ps = yyps;
				yy_state = yystate;
				yynerrs++;
			skip_init:
			case 1:
			case 2:		/* incompletely recovered error */
					/* try again... */
				yyerrflag = 3;
				/*
				** find state where "error" is a legal
				** shift action
				*/
				while ( yy_ps >= yys )
				{
					yy_n = yypact[ *yy_ps ] + YYERRCODE;
					if ( yy_n >= 0 && yy_n < YYLAST &&
						yychk[yyact[yy_n]] == YYERRCODE)					{
						/*
						** simulate shift of "error"
						*/
						yy_state = yyact[ yy_n ];
						goto yy_stack;
					}
					/*
					** current state has no shift on
					** "error", pop stack
					*/
#if YYDEBUG
#	define _POP_ "Error recovery pops state %d, uncovers state %d\n"
					if ( yydebug )
						(void)printf( _POP_, *yy_ps,
							yy_ps[-1] );
#	undef _POP_
#endif
					yy_ps--;
					yy_pv--;
				}
				/*
				** there is no state on stack with "error" as
				** a valid shift.  give up.
				*/
				YYABORT;
			case 3:		/* no shift yet; eat a token */
#if YYDEBUG
				/*
				** if debugging, look up token in list of
				** pairs.  0 and negative shouldn't occur,
				** but since timing doesn't matter when
				** debugging, it doesn't hurt to leave the
				** tests here.
				*/
				if ( yydebug )
				{
					register int yy_i;

					(void)printf( "Error recovery discards " );
					if ( yychar == 0 )
						(void)printf( "token end-of-file\n" );
					else if ( yychar < 0 )
						(void)printf( "token -none-\n" );
					else
					{
						for ( yy_i = 0;
							yytoks[yy_i].t_val >= 0;
							yy_i++ )
						{
							if ( yytoks[yy_i].t_val
								== yychar )
							{
								break;
							}
						}
						(void)printf( "token %s\n",
							yytoks[yy_i].t_name );
					}
				}
#endif /* YYDEBUG */
				if ( yychar == 0 )	/* reached EOF. quit */
					YYABORT;
				yychar = -1;
				goto yy_newstate;
			}
		}/* end if ( yy_n == 0 ) */
		/*
		** reduction by production yy_n
		** put stack tops, etc. so things right after switch
		*/
#if YYDEBUG
		/*
		** if debugging, print the string that is the user's
		** specification of the reduction which is just about
		** to be done.
		*/
		if ( yydebug )
			(void)printf( "Reduce by (%d) \"%s\"\n",
				yy_n, yyreds[ yy_n ] );
#endif
		yytmp = yy_n;			/* value to switch over */
		yypvt = yy_pv;			/* $vars top of value stack */
		/*
		** Look in goto table for next state
		** Sorry about using yy_state here as temporary
		** register variable, but why not, if it works...
		** If yyr2[ yy_n ] doesn't have the low order bit
		** set, then there is no action to be done for
		** this reduction.  So, no saving & unsaving of
		** registers done.  The only difference between the
		** code just after the if and the body of the if is
		** the goto yy_stack in the body.  This way the test
		** can be made before the choice of what to do is needed.
		*/
		{
			/* length of production doubled with extra bit */
			register int yy_len = yyr2[ yy_n ];

			if ( !( yy_len & 01 ) )
			{
				yy_len >>= 1;
				yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
				yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
					*( yy_ps -= yy_len ) + 1;
				if ( yy_state >= YYLAST ||
					yychk[ yy_state =
					yyact[ yy_state ] ] != -yy_n )
				{
					yy_state = yyact[ yypgo[ yy_n ] ];
				}
				goto yy_stack;
			}
			yy_len >>= 1;
			yyval = ( yy_pv -= yy_len )[1];	/* $$ = $1 */
			yy_state = yypgo[ yy_n = yyr1[ yy_n ] ] +
				*( yy_ps -= yy_len ) + 1;
			if ( yy_state >= YYLAST ||
				yychk[ yy_state = yyact[ yy_state ] ] != -yy_n )
			{
				yy_state = yyact[ yypgo[ yy_n ] ];
			}
		}
					/* save until reenter driver code */
		yystate = yy_state;
		yyps = yy_ps;
		yypv = yy_pv;
	}
	/*
	** code supplied by user is placed in this switch
	*/
	switch( yytmp )
	{
		
case 3:
# line 74 "Sy_IoParser.yacc"
{ printf ("%s\n", yytext); } break;
case 11:
# line 87 "Sy_IoParser.yacc"
{
#ifdef DEBUG               
                          printf("new object: "); 
#endif
                          } break;
case 12:
# line 93 "Sy_IoParser.yacc"
{
#ifdef DEBUG               
                          printf("%s\n", tmp_str);
#endif 
		          Sy_AddIoObject(&sy_yacc_tree, tmp_str);
	                  objnamestk[stkptr++] = Sy_Strcpy(tmp_str);
                          sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 13:
# line 102 "Sy_IoParser.yacc"
{ 
#ifdef DEBUG               
                      printf("exiting object\n");
	              printf ("Trying to call loader for %s\n",
	                      objnamestk[stkptr-1]);
#endif
 		      call_object_loader (objnamestk[stkptr-1]); 

		      stkptr--; free(objnamestk[stkptr]);
	              if (stkptr > 0){
	                (void)Sy_GetIoObject(&sy_yacc_tree, 
                                             objnamestk[stkptr-1]);
		      }
		      else{
			(void)Sy_GetIoObject(&sy_yacc_tree, "world");
		      }
		      sy_yacc_iotoken.yacc_tree = sy_yacc_tree;
                    } break;
case 19:
# line 130 "Sy_IoParser.yacc"
{ 
#ifdef DEBUG
                  printf ("assign %d to %s\n", yypvt[-1].ival, yypvt[-3].strval);
#endif 
		  Sy_AddDataSameObject(&sy_yacc_tree, yypvt[-3].strval, SY_INT_T, &yypvt[-1].ival);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 20:
# line 137 "Sy_IoParser.yacc"
{ 
#ifdef DEBUG
                  printf ("assign %f to %s\n", yypvt[-1].fval, yypvt[-3].strval); 
#endif
		  Sy_AddDataSameObject(&sy_yacc_tree, yypvt[-3].strval, SY_FLOAT_T, &yypvt[-1].fval);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 21:
# line 144 "Sy_IoParser.yacc"
{ 
#ifdef DEBUG
                  printf ("assign %x to %s\n", yypvt[-1].ival, yypvt[-3].strval);
#endif
		  Sy_AddDataSameObject(&sy_yacc_tree, yypvt[-3].strval, SY_INT_T, &yypvt[-1].ival);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 22:
# line 151 "Sy_IoParser.yacc"
{ 
#ifdef DEBUG
                  printf ("assign %s to %s\n", yypvt[-1].strval, yypvt[-3].strval);
#endif 
		  Sy_AddDataSameObject(&sy_yacc_tree, yypvt[-3].strval, SY_STR_T, yypvt[-1].strval);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 23:
# line 158 "Sy_IoParser.yacc"
{ 
#ifdef DEBUG
                  printf ("assign %s to %s\n", yypvt[-1].strval, yypvt[-3].strval);
#endif 	
		  Sy_AddDataSameObject(&sy_yacc_tree, yypvt[-3].strval, SY_STR_T, yypvt[-1].strval);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 29:
# line 174 "Sy_IoParser.yacc"
{ 
#ifdef DEBUG
                  printf ("assign from address %x to %s\n", yypvt[-1].ival, yypvt[-6].strval);
#endif
		  Sy_AddDataSameObject(&sy_yacc_tree, "memory", 
		  SY_INT_T, &yypvt[-1].ival);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 30:
# line 182 "Sy_IoParser.yacc"
{ Sy_AddDataSameObject(&sy_yacc_tree, "shmem", 
		  SY_INT_T, &yypvt[-1].ival);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 31:
# line 186 "Sy_IoParser.yacc"
{ Sy_AddDataSameObject(&sy_yacc_tree, "file", 
		  SY_STR_T, &yypvt[-1].strval);
                  sy_yacc_iotoken.yacc_tree = sy_yacc_tree;} break;
case 32:
# line 190 "Sy_IoParser.yacc"
{call_loader (yypvt[-1].strval);} break;
case 33:
# line 192 "Sy_IoParser.yacc"
{call_loader (yypvt[-1].strval);} break;
case 34:
# line 194 "Sy_IoParser.yacc"
{is_data = 1; yyclearin; 
#ifdef DEBUG
                 printf ("beginning default load seq\n");
#endif
                } break;
	}
	goto yystack;		/* reset registers in driver code */
}
