.\" @(#)contours.5 1.0 95/06/25
'\"macro stdmacro
.nr X
.TH contours 5 "25 June 1995"
.\" Copyright 1995 Robert L. Ogniewicz
.SH NAME
contours \- a file format for contours
.SH DESCRIPTION
.IX "contours" "" "\fLcontours\fP \(em contour file format"
The contour file format can be used to feed boundary data obtained
from other programs than  
.I ctran
into the Voronoi tessellation routine
.I voro
or into the visualization program
.I shosk. It has been designed to provide a flexible user
interface to .I voro.
 The description
of a single contour consists of a header (carrying general
information, which can be added to easily) plus a data section.
It is possible to place comments /* text */ everywhere outside
the data section. Assignments to variables appear in the form
.IP
.B "var = statement;" 
.LP
where statement can be an integer (myvar = 1;), a real value (myvar =
3.14), a string enclosed in double quotes (myvar = "hello world !"),
or simply a single identifier (basically, a string without blanks,
starting with a non-numeric character) (myvar = true;). Also
the order of assignments within the header is irrelevant. 
 
.SH EXAMPLE: FILE FORMAT
Below a listing (excerpt) of a single contour as obtained from
a rectangle. This output can be obtained by running
.IP
.B "ctran -invert perf_rect_tiny.ras -contour output.dat"
.LP
Contents of output.dat:
.nf
/* Skeleton Project File*/          /* comments */

object contour_1 {                  /* contour 1 */
type = contour_unord;               /* must appear in the header */
format = ascii;                     /* currently only ASCII supported */
contour_id = 1;                     /* unique ID number */
closed = true;                      /* if variable closed not set, closed
                                       contours are assumed. Otherwise
                                       set closed = false; Not all pruning
                                       operations are currently supported 
                                     */ 
                                           
pointdim = 82;                      /* number of contour points */
x_extent = 0; y_extent = 0;         /* currently not used       */


/* 
The two lines below start a loading routine for a following data
segment. The next two lines must not be modified.  The character
sequence #2718sync314beg# is a `magic cookie' used for synchronization
between parser of header language and generic data loading
routines. Accordingly, #2718sync314end# is used to accomplish
synchronization after data input.  
All data between beginloadseq (Sy_GenLoader) and #2718sync314beg# 
is discarded. 
Data simply consists of a sequence of x y coordinates, one or
more points (x,y) per line: (2,3) is the first point, (3,3) is
the second one.
*/

beginloadseq (Sy_GenLoader)
#2718sync314beg#

2 3 3 3 4 3 5 3 6 3 7 3 8 3 9 3 10 3 11 3 12 3 13 3 14 3 15 3 16 3 17
3 18 3 19 3 20 3 21 3 22 3 23 3 24 3 25 3 26 3 27 3 28 3 29 3 30 3 31
3 31 4 31 5 31 6 31 7 31 8 31 9 31 10 31 11 31 12 31 13 31 14 31 15 30
15 29 15 28 15 27 15 26 15 25 15 24 15 23 15 22 15 21 15 20 15 19 15
18 15 17 15 16 15 15 15 14 15 13 15 12 15 11 15 10 15 9 15 8 15 7 15 6
15 5 15 4 15 3 15 2 15 2 14 2 13 2 12 2 11 2 10 2 9 2 8 2 7 2 6 2 5 2 4

#2718sync314end#
endloadseq;                             /* terminate loading sequence */
}                                       /* end contour 1              */
.fi
.LP
More contours can be incorporated by simple concatenation.
Simply add
.nf

object contour_2 {
type = contour_unord; format = ascii;
contour_id = 2;
closed = true;
pointdim = 124;
x_extent = 0; y_extent = 0;
beginloadseq (Sy_GenLoader)
#2718sync314beg#

 \.... data ....

#2718sync314end#
endloadseq;
} 
.fi
.SH SEE ALSO
.BR ctran (1),
.BR voro (1),
.BR prune (1),
.BR skeleton (5)
.SH AUTHOR
.nf
Robert L. Ogniewicz
E-Mail: rlo@hrl.harvard.edu
.fi

This work has been carried out at the
.I "Communication Technology Laboratory," 
Swiss Federal Institute of
Technology ETH, Zurich, Switzerland, and the
.I "Division of Applied Sciences," 
Harvard University, Cambridge, USA.
.SH COPYRIGHT
Copyright (C) 1995 Robert L. Ogniewicz 
.SH BUGS
Undoubtedly many...
.\".Ee
