#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/resource.h>
#include <sys/param.h>
#include "solex.h"

void SKE_init_timer(void);
void SKE_show_timer(void);

main(int argc, char** argv)
{
  int useconds = 1000000,
      tablesize, pagesize,
      i;

  long hostid;
  double s, mul;

  char hostname[1024];

  SKE_init_timer();
  
  printf("running usleep (%d microseconds)\n", useconds);
  usleep(useconds);

  printf("running gethostid: "); fflush(stdout);
  hostid = gethostid();
  printf("%x\n", hostid);
  
  printf("running getdtablesize: "); fflush(stdout);
  tablesize = getdtablesize();
  printf("%d\n", tablesize);

  printf("running getpagesize: "); fflush(stdout);
  pagesize = getpagesize();
  printf("%d\n", pagesize);

  printf("running gethostname: "); fflush(stdout);
  gethostname(hostname,sizeof(hostname));
  printf("%s\n", hostname);

#define ITER 1000000
  s = 1.0; mul = 1.0 + 1.0/(double)ITER;
  for (i = 1; i <= ITER; i++) s *= mul;
  printf ("e approx. %20.18lf\n", s);

  SKE_show_timer(); 

}

#define MICRO 1000000.0
static struct rusage rus0, rus1;
static int    wbuff0, wbuff1;
static struct tms buffer;         

typedef struct{
  float wtime;
  float utime;
  float stime;
} SkeRusdelta;

char wt[] = "wallclock:";
char ut[] = "user time:";
char st[] = "system time:";
char pr[] = "page reclaims:";
char pf[] = "page faults:";
char sw[] = "swaps:";
char ic[] = "input count:";
char oc[] = "output count:";
char vs[] = "vol. context swi.:";
char is[] = "invol. cont. swi.:";

void SKE_init_timer()
{
  getrusage(RUSAGE_SELF,&rus0);
  wbuff0 = times(&buffer);
}

void SKE_show_timer()
{
  float utime0, utime1, stime0, stime1, wall0, wall1;

  getrusage(RUSAGE_SELF,&rus1);
  wbuff1 = times(&buffer);

  wall0  = (float) wbuff0 / (float) HZ;
  wall1  = (float) wbuff1 / (float) HZ;

  utime0 = MICRO *(float) rus0.ru_utime.tv_sec + (float) rus0.ru_utime.tv_usec;
  utime1 = MICRO *(float) rus1.ru_utime.tv_sec + (float) rus1.ru_utime.tv_usec;
  stime0 = MICRO *(float) rus0.ru_stime.tv_sec + (float) rus0.ru_stime.tv_usec;
  stime1 = MICRO *(float) rus1.ru_stime.tv_sec + (float) rus1.ru_stime.tv_usec;

  fprintf(stderr, "\n*** resource utilization ***");
  fprintf(stderr, "****************************\n");
  fprintf(stderr, "%-18s %8.3f\n", wt, wall1-wall0);
  fprintf(stderr, "%-18s %8.3f", ut, (utime1-utime0)/MICRO);
  fprintf(stderr, " %-18s %8d\n", sw, rus1.ru_nswap   - rus0.ru_nswap);
  fprintf(stderr, "%-18s %8.3f", st, (stime1-stime0)/MICRO);
  fprintf(stderr, " %-18s %8d\n", ic, rus1.ru_inblock - rus0.ru_inblock);
  fprintf(stderr, "%-18s %8d", pr, rus1.ru_minflt  - rus0.ru_minflt);
  fprintf(stderr, " %-18s %8d\n", oc, rus1.ru_oublock - rus0.ru_oublock);
  fprintf(stderr, "%-18s %8d", pf, rus1.ru_majflt  - rus0.ru_majflt);  
  fprintf(stderr, " %-18s %8d\n", vs, rus1.ru_nvcsw   - rus0.ru_nvcsw);
  fprintf(stderr, "%-27s", "");
  fprintf(stderr, " %-18s %8d\n", is, rus1.ru_nivcsw  - rus0.ru_nivcsw);
  fprintf(stderr, "****************************");
  fprintf(stderr, "****************************\n");
}

SkeRusdelta SKE_get_timer(display_flag)
int display_flag;
{
  SkeRusdelta delta;
  float utime0, utime1, stime0, stime1, wall0, wall1;

  getrusage(RUSAGE_SELF,&rus1);
  wbuff1 = times(&buffer);

  wall0  = (float) wbuff0 / (float) HZ;
  wall1  = (float) wbuff1 / (float) HZ;

  utime0 = MICRO *(float) rus0.ru_utime.tv_sec + (float) rus0.ru_utime.tv_usec;
  utime1 = MICRO *(float) rus1.ru_utime.tv_sec + (float) rus1.ru_utime.tv_usec;
  stime0 = MICRO *(float) rus0.ru_stime.tv_sec + (float) rus0.ru_stime.tv_usec;
  stime1 = MICRO *(float) rus1.ru_stime.tv_sec + (float) rus1.ru_stime.tv_usec;

  delta.utime = (utime1-utime0)/MICRO;
  delta.stime = (stime1-stime0)/MICRO;
  delta.wtime = wall1-wall0;

  if (display_flag){
    fprintf(stderr, "\n*** resource utilization ***");
    fprintf(stderr, "****************************\n");
    fprintf(stderr, "%-18s %8.3f\n", wt, wall1-wall0);
    fprintf(stderr, "%-18s %8.3f", ut, (utime1-utime0)/MICRO);
    fprintf(stderr, " %-18s %8d\n", sw, rus1.ru_nswap   - rus0.ru_nswap);
    fprintf(stderr, "%-18s %8.3f", st, (stime1-stime0)/MICRO);
    fprintf(stderr, " %-18s %8d\n", ic, rus1.ru_inblock - rus0.ru_inblock);
    fprintf(stderr, "%-18s %8d", pr, rus1.ru_minflt  - rus0.ru_minflt);
    fprintf(stderr, " %-18s %8d\n", oc, rus1.ru_oublock - rus0.ru_oublock);
    fprintf(stderr, "%-18s %8d", pf, rus1.ru_majflt  - rus0.ru_majflt);  
    fprintf(stderr, " %-18s %8d\n", vs, rus1.ru_nvcsw   - rus0.ru_nvcsw);
    fprintf(stderr, "%-27s", "");
    fprintf(stderr, " %-18s %8d\n", is, rus1.ru_nivcsw  - rus0.ru_nivcsw);
    fprintf(stderr, "****************************");
    fprintf(stderr, "****************************\n");
  }
  
  return (delta);
}


